#include "LogWindow.h"
#include "Log.h"
#include <QtGui/QVBoxLayout>
#include <QtGui/QHBoxLayout>
#include <QtGui/QLabel>
#include <QtGui/QPushButton>
#include <QtGui/QScrollBar>
#include <QtGui/QKeyEvent>
#include <QtCore/QEventLoop>
#include "Colors.h"

#define TEXT_DIMS  100, 24

#define DEFAULT_LEVEL   2
#define MAX_LEVEL      10

Log::Window::Window(EntryList&  entries,
                    QWidget*    parent)
   : parent(parent), entries(entries)
{
   setWindowTitle(tr("NUU Log"));

   QVBoxLayout*  mainLayout = new QVBoxLayout(this);

   textEdit = new HTextEdit(TEXT_DIMS, this);
   textEdit->setReadOnly(true);
   textEdit->setLineWrapMode(QTextEdit::NoWrap);
   textEdit->setWordWrapMode(QTextOption::NoWrap);
   textEdit->setTabChangesFocus(true);
   mainLayout->addWidget(textEdit);

   QHBoxLayout*  controlLayout = new QHBoxLayout();
   QLabel*       levelLabel    = new QLabel(tr("Detail level"), this);
                 levelSpinBox  = new QSpinBox(this);
   levelSpinBox->setRange(0, MAX_LEVEL);
   levelSpinBox->setValue(DEFAULT_LEVEL);
   levelSpinBox->setToolTip(tr("Threshold beyond which more detailed log"
                               " messages are omitted.\n"
                               "Higher levels show more detail.\n"
                               "The valid range is 0 .. %1.\n"
                               "The default is %2.\n"
                               "Higher levels contain debug information.\n"
                               "This corresponds to yum's debuglevel (-d)"
                               " option.").arg(MAX_LEVEL)
                                          .arg(DEFAULT_LEVEL));
   connect(levelSpinBox, SIGNAL(valueChanged(int)), 
           this,         SLOT(levelChanged(int)));
   QPushButton*  closeButton  = new QPushButton(tr("&Close"), this);
   closeButton->setToolTip(tr("Close the Log Window"));
   controlLayout->addWidget(levelLabel);
   controlLayout->addWidget(levelSpinBox);
   controlLayout->addStretch(1);
   controlLayout->addWidget(closeButton);
   connect(closeButton, SIGNAL(clicked()),
           this,        SLOT(hide()));
   mainLayout->addLayout(controlLayout);

   setLayout(mainLayout);   
}

void Log::Window::setVisible(bool visible)
{
   bool wasVisible = isVisible();

   if (visible && !wasVisible) {
      textEdit->clear();
      shownCount = 0;
      
      appended();
      
      entries.lock();
      connect(&entries, SIGNAL(appended()), 
              this,     SLOT(appended()));
      entries.unlock();
      
      QScrollBar*  vbar = textEdit->verticalScrollBar();
      vbar->setValue(vbar->maximum());
   }

   QWidget::setVisible(visible);

   if (!visible && wasVisible) {
      entries.lock();
      disconnect(&entries, SIGNAL(appended()), 
                 this,     SLOT(appended()));
      entries.unlock();

      shownCount = 0;
      textEdit->clear();
   }
}

void Log::Window::appended()
{
   entries.lock();
   bool  newEntries = (entries.count() > shownCount);
   if (newEntries) {
      unsigned int  level = levelSpinBox->value();
      int           i     = 0;
      for (i = shownCount; i < entries.count(); i++) {
         Entry&   entry = entries[i];
         if (entry.value <= level) {
            QString  text  = entry.text;
            htmlProtect(text);
            switch (entry.kind) {
            case NORMAL:
               if (entry.value >= 10) {
                  text.prepend(QObject::tr("<font color=\"%1\">")
                                  .arg(Colors::minor()));
                  text.append(QObject::tr("</font>"));
               } else {
                  text.prepend(QObject::tr("<font>"));
                  text.append(QObject::tr("</font>"));
               }
               break;
            case STDOUT:
               text.prepend(QObject::tr("<font color=\"%1\">")
                               .arg(Colors::ok()));
               text.append(QObject::tr("</font>"));
               break;
            case ACTION:
               text.prepend(QObject::tr("<h2><font color=\"%1\">")
                               .arg(Colors::action()));
               text.append(QObject::tr("</font></h2>"));
               break;
            case WARNING:
               text.prepend(QObject::tr("<font color=\"%1\">")
                               .arg(Colors::warning()));
               text.append(QObject::tr("</font>"));
               break;
            case ERROR:
            default:
               text.prepend(QObject::tr("<font color=\"%1\">")
                               .arg(Colors::error()));
               text.append(QObject::tr("</font>"));
               break;
            }
            textEdit->appendCompact(text);
         }
      }
      shownCount = entries.count();
   }
   entries.unlock();
   if (newEntries) {
      QEventLoop  eventLoop;
      eventLoop.processEvents(QEventLoop::ExcludeUserInputEvents);
   }
}

void  Log::Window::htmlProtect(QString&  str)
{
   str.replace(QString("<"),  QString("&lt;"));
   str.replace(QString(">"),  QString("&gt;"));
   str.replace(QString(" "),  QString("&#x0020;"));
   str.replace(QString("\t"), QString("&#x0009;"));
#if 0
   str.replace(QString("&"),  QString("&amp;"));
   str.replace(QString("%"),  QString("&percnt;"));
   str.replace(QString("\""), QString("&quot;"));
   str.replace(QString("'"),  QString("&apos;"));
#endif
}

void Log::Window::keyPressEvent(QKeyEvent*  event)
{
   if (event->key() == Qt::Key_Escape) {
      event->accept();
      hide();
   } else {
      QWidget::keyPressEvent(event);
   }
}


void Log::Window::levelChanged(int  level)
{
   shownCount = 0;
   textEdit->clear();
   appended();
}
