#include "MainWindow.h"
#include <QtGui/QMenuBar>
#include <QtGui/QLabel>
#include <QtGui/QLineEdit>
#include "PackageTable.h"
#include "PackageTableModel.h"
#include "PackageTableView.h"
#include <QtCore/QObject>
#include <QtGui/QHBoxLayout>
#include <QtGui/QVBoxLayout>
#include <QtGui/QMessageBox>
#include "HProgressBar.h"
#include <QtCore/QTimer>
#include "Share.h"
#include "RepoAuthChecker.h"
#include <QtCore/QEventLoop>
#include <QtGui/QApplication>
#include "Execute.h"
#include "stringmacros.h"
#include <sys/utsname.h>

MainWindow::MainWindow(QWidget*  parent)
   : QMainWindow(parent),
     updateAllIcon(Share::absolute("icons/update.png")),
     checkIcon(Share::absolute("icons/check.png")),
     optRepoEditor(false),
     optLog(false),
     repoEditorDialog(NULL),
     logWindow(NULL)
{
   struct utsname  utsbuf;
   uname(&utsbuf);

   setWindowTitle(tr("NUU - %1").arg(utsbuf.nodename));

   createMenuBar();
   createMainArea();
}

void MainWindow::setVisible(bool  visible)
{
   QMainWindow::setVisible(visible);
   if (visible) {
      QTimer::singleShot(0, this, SLOT(initialize()));
   }
}

void MainWindow::createMenuBar()
{
   QMenuBar*  menuBar = new QMenuBar(this);
   this->setMenuBar(menuBar);

             programMenu = new HMenu(tr("&Program"), menuBar);
   menuBar->addMenu(programMenu);
             quitAction  = new QAction(tr("&Quit"), programMenu);
   quitAction->setShortcut(QKeySequence(tr("Ctrl+Q")));
   connect(quitAction, SIGNAL(triggered()), this, SLOT(quit()));
   programMenu->addAction(quitAction);

   HMenu*    windowsMenu = new HMenu(tr("&Windows"), menuBar);
   menuBar->addMenu(windowsMenu);
   QAction*  logAction   = new QAction(tr("&Log"), windowsMenu);
   logAction->setShortcut(QKeySequence(tr("Ctrl+L")));
   connect(logAction, SIGNAL(triggered()), this, SLOT(displayLog()));
   windowsMenu->addAction(logAction);

             repoMenu    = new HMenu(tr("&Repositories"), menuBar);
   menuBar->addMenu(repoMenu);
             editAction  = new QAction(tr("&Edit Configuration"), repoMenu);
   editAction->setShortcut(QKeySequence(tr("Ctrl+E")));
   connect(editAction, SIGNAL(triggered()), this, SLOT(editRepositories()));
   repoMenu->addAction(editAction);

             cacheMenu   = new HMenu(tr("&Cache"), menuBar);
   menuBar->addMenu(cacheMenu);
             cleanHeadersAction      = new QAction(tr("Clean Headers"),
                                                   cacheMenu);
             cleanPackagesAction     = new QAction(tr("Clean Packages"),
                                                   cacheMenu);
             cleanXmlMetadataAction  = new QAction(tr("Clean XML Metadata"),
                                                   cacheMenu);
             cleanPickledCacheAction = new QAction(tr("Clean Pickled Data"),
                                                   cacheMenu);
             cleanAllAction          = new QAction(tr("Clean All"), cacheMenu);
   connect(cleanHeadersAction,      SIGNAL(triggered()), 
           this,                    SLOT(cleanHeaders()));
   connect(cleanPackagesAction,     SIGNAL(triggered()), 
           this,                    SLOT(cleanPackages()));
   connect(cleanXmlMetadataAction,  SIGNAL(triggered()), 
           this,                    SLOT(cleanXmlMetadata()));
   connect(cleanPickledCacheAction, SIGNAL(triggered()), 
           this,                    SLOT(cleanPickledCache()));
   connect(cleanAllAction,          SIGNAL(triggered()), 
           this,                    SLOT(cleanAll()));
   cacheMenu->addAction(cleanHeadersAction);
   cacheMenu->addAction(cleanPackagesAction);
   cacheMenu->addAction(cleanXmlMetadataAction);
   cacheMenu->addAction(cleanPickledCacheAction);
   cacheMenu->addSeparator();
   cacheMenu->addAction(cleanAllAction);

#if 0
   // This only affects styles like Motif & CDE.  But for them, the Help menu
   // gets shoved off the right of the window.  Qt bug?  No matter; this
   // isn't important, so just disable it.
   menuBar->addSeparator();
#endif

   HMenu*    helpMenu    = new HMenu(tr("&Help"), menuBar);
   menuBar->addMenu(helpMenu);
   QAction*  manualHtmlAction = new QAction(tr("Manual (html)"), helpMenu);
   QAction*  manualPdfAction  = new QAction(tr("Manual (PDF)"),  helpMenu);
   QAction*  aboutAction      = new QAction(tr("&About"),        helpMenu);
   manualHtmlAction->setShortcut(Qt::Key_F1);
   connect(manualHtmlAction, SIGNAL(triggered()), this, SLOT(manualHtml()));
   connect(manualPdfAction,  SIGNAL(triggered()), this, SLOT(manualPdf()));
   connect(aboutAction,      SIGNAL(triggered()), this, SLOT(about()));
   helpMenu->addAction(manualHtmlAction);
   helpMenu->addAction(manualPdfAction);
   helpMenu->addSeparator();
   helpMenu->addAction(aboutAction);
}

void MainWindow::createMainArea()
{
   QWidget*      centralWidget = new QWidget(this);
   QVBoxLayout*  mainLayout    = new QVBoxLayout(centralWidget);

   packageTableConfig = new QWidget(centralWidget);
   QVBoxLayout*  packageTableLayout = new QVBoxLayout(packageTableConfig);

   QHBoxLayout*  controlLayout  = new QHBoxLayout();
                 modeComboBox   = new QComboBox(packageTableConfig);
   modeComboBox->addItem(tr("Filter"), PackageTable::FILTER);
   modeComboBox->addItem(tr("Search"), PackageTable::SEARCH);
   controlLayout->addWidget(modeComboBox);
                 controlEdit    = new QLineEdit(".*", packageTableConfig);
   controlLayout->addWidget(controlEdit);
                 controlButton  = new QPushButton(packageTableConfig);
   controlButton->setDefault(true);
   controlLayout->addWidget(controlButton);
                 stateComboBox   = new QComboBox(packageTableConfig);
   stateComboBox->addItem(tr("View Updatable"),   PackageTable::UPDATABLE);
   stateComboBox->addItem(tr("View Installable"), PackageTable::INSTALLABLE);
   stateComboBox->addItem(tr("View Removable"),   PackageTable::REMOVABLE);
   stateComboBox->addItem(tr("View All"),         PackageTable::ALL);
   controlLayout->addWidget(stateComboBox);
                 groupsCheckBox = new QCheckBox(tr("Include Groups"), 
                                                packageTableConfig);
   controlLayout->addWidget(groupsCheckBox);
   // modeComboBox ToolTip set in collectedExtraData().
   // controlEdit and controlButton ToolTips set in modeChanged()
   stateComboBox->setToolTip(tr("Filter for those packages which are eligible"
                                " for updating,\n"
                                "for installing, or for removing; or show all"
                                " packages"));
   groupsCheckBox->setToolTip(tr("Whether or not filter should include package"
                                 " groups"));
   connect(modeComboBox,   SIGNAL(currentIndexChanged(int)), 
           this,           SLOT(modeChanged(int)));
   connect(controlEdit,    SIGNAL(returnPressed()), this, SLOT(filtersearch()));
   connect(controlButton,  SIGNAL(clicked()),       this, SLOT(filtersearch()));
   connect(stateComboBox,  SIGNAL(currentIndexChanged(int)), 
           this,           SLOT(stateFilterChanged(int)));
   connect(groupsCheckBox, SIGNAL(stateChanged(int)), 
           this,           SLOT(groupsFilterChanged(int)));
   packageTableLayout->addLayout(controlLayout);

   yumIO = new YumIO;
   connect(yumIO, SIGNAL(collectedExtraData()),
           this,  SLOT(collectedExtraData()));
   packageTableModel = new PackageTable::Model(yumIO, packageTableConfig);
   packageTableView = new PackageTable::View(packageTableModel, 
                                             PackageTable::UPDATABLE,
                                             false,
                                             yumIO,
                                             packageTableConfig);
   connect(packageTableView, SIGNAL(currentChanged(int, int)), 
           this,             SLOT(currentChanged(int, int)));
   connect(packageTableView, SIGNAL(stateChanged(int)), 
           this,             SLOT(stateChanged(int)));
   packageTableLayout->addWidget(packageTableView);

   QHBoxLayout*  buttonLayout    = new QHBoxLayout();
                 describeButton  = new QPushButton(tr("&Describe"),
                                                   packageTableConfig);
   QPushButton*  resizeButton    = new QPushButton(tr("Re&size"),
                                                   packageTableConfig);
   QPushButton*  refreshButton   = new QPushButton(tr("Re&fresh"),
                                                   packageTableConfig);
   QPushButton*  selectedButton  = new QPushButton(tr("Set &Install/Update of"
                                                      " Selected"),
                                                   packageTableConfig);
   QPushButton*  updateAllButton = new QPushButton(updateAllIcon,
                                                   tr("Set &Update of All"
                                                      " Out-of-Date"),
                                                   packageTableConfig);
   QPushButton*  clearButton     = new QPushButton(clearIcon,
                                                   tr("&Clear Actions"),
                                                   packageTableConfig);
                 applyButton     = new QPushButton(checkIcon,
                                                   tr("&Apply Actions"),
                                                   packageTableConfig);
   describeButton->setEnabled(false);
   applyButton->setEnabled(false);
   describeButton->setToolTip(tr("Pop up a window with a detailed description"
                                 " of the currently selected package"));
   resizeButton->setToolTip(tr("Resize the columns and window to optimal"
                                " values"));
   refreshButton->setToolTip(tr("Refresh the list of installed packages and"
                                " packages available for update.\n"
                                "Use this to check for new updates in"
                                " repositories, or if any packages were"
                                " installed manually."));
   selectedButton->setToolTip(tr("Sets the Action for all selected packages"
                                 " which are Not Installed to Install, and\n"
                                 "sets the Action for all selected packages"
                                 " which are Installed but Out-of-date to"
                                 " Update.\n"
                                 "Does not Apply Actions automatically."));
   updateAllButton->setToolTip(tr("Sets the Action for all packages which are"
                                  " Installed but Out-of-date to Update.\n"
                                  "Does not Apply Actions automatically."));
   clearButton->setToolTip(tr("Clears all specified Actions"
                              " (Updates, Installs, and Removals)"));
   applyButton->setToolTip(tr("Applies the specified Actions"
                              " (Updates, Installs, and Removals)"));
   buttonLayout->addWidget(describeButton);
   buttonLayout->addStretch(1);
   buttonLayout->addWidget(resizeButton);
   buttonLayout->addWidget(refreshButton);
   buttonLayout->addWidget(selectedButton);
   buttonLayout->addWidget(updateAllButton);
   buttonLayout->addWidget(clearButton);
   buttonLayout->addWidget(applyButton);
   connect(resizeButton,     SIGNAL(clicked()),
           this,             SLOT(resize()));
   connect(refreshButton,    SIGNAL(clicked()),
           this,             SLOT(refresh()));
   connect(describeButton,   SIGNAL(clicked()),
           packageTableView, SLOT(describe()));
   connect(selectedButton,   SIGNAL(clicked()),
           this,             SLOT(setSelected()));
   connect(updateAllButton,  SIGNAL(clicked()),
           this,             SLOT(setUpdateAll()));
   connect(clearButton,      SIGNAL(clicked()),
           this,             SLOT(clearActions()));
   connect(applyButton,      SIGNAL(clicked()), 
           this,             SLOT(apply()));
   packageTableLayout->addLayout(buttonLayout);

   packageTableConfig->setLayout(packageTableLayout);
   packageTableConfig->setVisible(false);
   mainLayout->addWidget(packageTableConfig);
   
   HProgressBar*  itemProgress  = new HProgressBar(centralWidget);
   mainLayout->addWidget(itemProgress);

   HProgressBar*  totalProgress = new HProgressBar(centralWidget);
   mainLayout->addWidget(totalProgress);

   yumIO->setup(this, totalProgress, itemProgress);

   centralWidget->setLayout(mainLayout);
   setCentralWidget(centralWidget);

   collectedExtraData();
   modeChanged(PackageTable::FILTER);
   setReady(false);
}

void MainWindow::initialize()
{
   yumConfIO = new YumConfIO(yumReposEnabled, yumReposUsed);

   connect(yumConfIO, SIGNAL(changed()),
           this,      SLOT(yumConfIOChanged()));

   packageTableModel->setYumConfIO(yumConfIO);

   if (optLog) {
      displayLog();
   }

   yumIO->initialize();

   if (optRepoEditor) {
      packageTableModel->reset();
      setReady(true);
      // Use forceReposChanged so that the RepoEditorDialog will emit a
      // reposChanged signal no matter whether anything changed or not.  That
      // will cause a collect to happen when the repoEditorDialog is closed.
      // It also will cause reposChanged() to make the packageTableConfig
      // visible and to resize the packageTableView.
      editRepositories(/* forceReposChanged => */ true);
   } else {
      RepoAuth::Checker  checker(this, yumConfIO);
      if (checker.quit) {
         quit();
         return;
      }
      if (checker.changed) {
         yumIO->clear();
      }
      yumIO->collect(yumReposUsed);
      packageTableModel->reset();
      packageTableConfig->setVisible(true);
      packageTableView->resizeColumnsToContents();
      setReady(true);
   }
}

void MainWindow::yumConfIOChanged()
{
   cleanHeadersAction->setToolTip(tr("Clean RPM package headers cached locally"
                                     " in %1")
                                     .arg(yumConfIO->cachedir));
   cleanPackagesAction->setToolTip(tr("Clean RPM packages cached locally in"
                                      " %1\n"
                                      "(useful for recovering disk space)")
                                      .arg(yumConfIO->cachedir));
   cleanXmlMetadataAction->setToolTip(tr("Clean XML metadata cached locally in"
                                         " %1\n"
                                         "(useful along with Clean Pickled"
                                         " Data for\n"
                                         " forcing download of new available"
                                         " package lists)")
                                         .arg(yumConfIO->cachedir));
   cleanPickledCacheAction->setToolTip(tr("Clean pickled data cached locally in"
                                          " %1\n"
                                          "(useful along with Clean XML"
                                          " Metadata for\n"
                                          " forcing download of new available"
                                          " package lists)")
                                          .arg(yumConfIO->cachedir));
   cleanAllAction->setToolTip(tr("Clean all data cached locally in"
                                 " %1")
                                 .arg(yumConfIO->cachedir));
}

void MainWindow::quit()
{
   if (logWindow != NULL && logWindow->isVisible()) {
      logWindow->hide();
   }
   delete yumIO; // This releases the yum lock
   close();
}

void MainWindow::displayLog()
{
   if (logWindow == NULL) {
      logWindow = new Log::Window(yumIO->logEntries, this);
   }
   
   logWindow->show();
}

void MainWindow::editRepositories(bool  forceReposChanged)
{
   if (repoEditorDialog == NULL) {
      repoEditorDialog = new RepoEditor::Dialog(yumIO, yumConfIO, this);

      connect(repoEditorDialog, SIGNAL(reposChanged()),
              this,             SLOT(reposChanged()));
   }

   repoEditorDialog->forceReposChanged = forceReposChanged;

   repoEditorDialog->exec();
}

void MainWindow::cleanHeaders()
{
   setReady(false);

   yumIO->cleanHeaders();

   yumIO->clear();
   yumIO->collect(yumReposUsed);
   packageTableModel->reset();
   setReady(true);
}

void MainWindow::cleanPackages()
{
   setReady(false);
   yumIO->cleanPackages();
   setReady(true);
}

void MainWindow::cleanXmlMetadata()
{
   setReady(false);

   yumIO->cleanXmlMetadata();

   yumIO->clear();
   yumIO->collect(yumReposUsed);
   packageTableModel->reset();
   setReady(true);
}

void MainWindow::cleanPickledCache()
{
   setReady(false);

   yumIO->cleanPickledCache();

   yumIO->clear();
   yumIO->collect(yumReposUsed);
   packageTableModel->reset();
   setReady(true);
}

void MainWindow::cleanAll()
{
   setReady(false);

   yumIO->cleanAll();

   yumIO->clear();
   yumIO->collect(yumReposUsed);
   packageTableModel->reset();
   setReady(true);
}

void MainWindow::manualHtml()
{
   QString  doc    = tr("%1/html/index.html").arg(DOCDIR);
   char*    docStr = new char[doc.length()+1];
   strcpy(docStr, doc.toStdString().c_str());

   QString  tool = Share::absolute("webbrowser");
   char*    toolStr = new char[tool.length()+1];
   strcpy(toolStr, tool.toStdString().c_str());

   char*  argv[3];
   int    arg = 0;
   argv[arg++] = toolStr;
   argv[arg++] = docStr;
   argv[arg++] = NULL;

   Execute::execute(this, argv);

   delete [] toolStr;
   delete [] docStr;
}

void MainWindow::manualPdf()
{
   QString  doc    = tr("%1/manual.pdf").arg(DOCDIR);
   char*    docStr = new char[doc.length()+1];
   strcpy(docStr, doc.toStdString().c_str());

   QString  tool = Share::absolute("mimeviewer");
   char*    toolStr = new char[tool.length()+1];
   strcpy(toolStr, tool.toStdString().c_str());

   char*  mimeStr = new char[STRLEN("application/pdf")+1];
   strcpy(mimeStr, "application/pdf");

   char*  argv[4];
   int    arg = 0;
   argv[arg++] = toolStr;
   argv[arg++] = mimeStr;
   argv[arg++] = docStr;
   argv[arg++] = NULL;

   Execute::execute(this, argv);

   delete [] mimeStr;
   delete [] toolStr;
   delete [] docStr;
}

void MainWindow::about()
{
   QString  msg = tr("<h3>NUU %1</h3>"
                     "<hr>"
                     "<p>NUU, Network Update Utility, is a graphical user"
                     " interface which allows a user to make modifications to"
                     " the installed rpm packages on a system.</p>"
                     "<p>Capabilities:<p>"
                     "<ul>"
                     "<li>Install new packages or update existing packages by"
                     " downloading them from a yum package repository</li>"
                     "<li>Remove existing packages.</li>"
                     "<li>Configure sets of repositories from which to obtain"
                     " new and updated packages.</li>"
                     "<li>Examine installed packages or packages available for"
                     " installation or update</li>"
                     "<li>Install, update, or remove groups of packages</li>"
                     "<li>Install or update suites of packages</li>"
                     "</ul>"
                     "<p>For more information see "
                     "http://redhawk.ccur.com/nuu.</p>").arg(VERSION);

   QMessageBox::about(this, tr("About"), msg);
}

void MainWindow::filtersearch(void)
{
   if (modeComboBox->currentIndex() == PackageTable::SEARCH) setReady(false);

   prepareFilterSearch();
      
   packageTableView->filtersearch(controlEdit->text());

   if (modeComboBox->currentIndex() == PackageTable::SEARCH) setReady(true);
}

void MainWindow::currentChanged(int  currentRow,
                                int  previousRow)
{
   describeButton->setEnabled(currentRow != -1);
}

void MainWindow::stateChanged(int  row)
{
   bool  gotAction = false;
   if (!gotAction && row != -1) {
      gotAction = (*yumIO)[row]->action != Package::NONE;
   }
   if (!gotAction) {
      Package::EntryList::const_iterator  it = yumIO->begin();
      for (; it != yumIO->end(); ++it) {
         if ((*it)->action != Package::NONE) {
            gotAction = true;
            break;
         }
      }
   }

   applyButton->setEnabled(gotAction);
}

void MainWindow::modeChanged(int  index)
{
   if (index == PackageTable::SEARCH) setReady(false);

   switch (index) {
   case PackageTable::FILTER:
      controlButton->setText(tr("Filter"));
      controlEdit->setToolTip(tr("Regular expression filter matching textual"
                                 " fields visible in the table"));
      controlButton->setToolTip(tr("Apply the filter specified to the left"));
      break;
   case PackageTable::SEARCH:
      controlButton->setText(tr("Search"));
      controlEdit->setToolTip(tr("Regular expression to compare to all fields"
                                 " of a package"));
      controlButton->setToolTip(tr("Perform the search for the regular"
                                   " expression specified to the left"));
      break;
   }

   prepareFilterSearch();
      
   packageTableView->setMode((PackageTable::Mode)index);

   if (index == PackageTable::SEARCH) setReady(true);
}

void MainWindow::prepareFilterSearch()
{
   if (!controlEdit->text().isEmpty()
       && controlEdit->text() != QString(".*")
       && modeComboBox->currentIndex() == PackageTable::SEARCH) {
      // Do this so that PackageTable::SortFilterProxyModel::setGroupsFilter
      // doesn't have to.  But be careful!  While these entities are being
      // loaded, we must not allow any other Python code to be called.  That
      // would entail python code being run simultaneously in two QThreads.
      //
      // How can this happen?  An example is some X11 event causing
      // HTableView::sizeHint() which calls
      // QSortFilterProxyModel::columnCount and which indirectly causes
      // PackageTable::SortFilterProxyModel::filterAcceptsRow to be called.
      // And in SEARCH mode, that causes python code to be called.
      //
      // So, disable filtering while making these two calls.
      packageTableView->setFilter(false);
      yumIO->collectFileLists();
      yumIO->collectOtherData();
      packageTableView->setFilter(true);
   }
}

void MainWindow::collectedExtraData()
{
   QString  tip = tr("Specify the method for using a regular expression to"
                     " select packages in the table:\n"
                     "	Filter uses the textual fields visible in the table, as"
                     " if they were separated by spaces.\n"
                     "	Search uses all fields in packages, including list of"
                     " Files, Change Log, Requires, etc.,\n"
                     "	independently.  It is considerably slower.");
   if (!yumIO->collectedFileLists || !yumIO->collectedOtherData) {
      tip = tr("%1 (Must load extra data.)").arg(tip);
   }

   modeComboBox->setToolTip(tip);
}

void MainWindow::stateFilterChanged(int  index)
{
   if (modeComboBox->currentIndex() == PackageTable::SEARCH) setReady(false);
   packageTableView->stateFilterChanged((PackageTable::StateFilter)index);
   if (modeComboBox->currentIndex() == PackageTable::SEARCH) setReady(true);
}

void MainWindow::groupsFilterChanged(int  checked)
{
   if (modeComboBox->currentIndex() == PackageTable::SEARCH) setReady(false);
   packageTableView->groupsFilterChanged((bool)checked);
   if (modeComboBox->currentIndex() == PackageTable::SEARCH) setReady(true);
}

void MainWindow::setReady(bool  ready)
{
   {   
      Python::PyGILStateWrapper  gsw(/* needed => */ ready);

      programMenu->setEnabled(ready);
      quitAction->setEnabled(ready);
      repoMenu->setEnabled(ready);
      editAction->setEnabled(ready);
      cacheMenu->setEnabled(ready);
      cleanHeadersAction->setEnabled(ready);
      cleanPackagesAction->setEnabled(ready);
      cleanXmlMetadataAction->setEnabled(ready);
      cleanPickledCacheAction->setEnabled(ready && yumIO->hasPickledCache());
      cleanAllAction->setEnabled(ready);
      packageTableConfig->setEnabled(ready);
   }

   if (ready) {
      QApplication::restoreOverrideCursor();
   } else {
      QApplication::setOverrideCursor(QCursor(Qt::BusyCursor));
   }

   QEventLoop  eventLoop;
   eventLoop.processEvents(QEventLoop::ExcludeUserInputEvents);
}

void MainWindow::resize()
{
   packageTableView->resizeColumnsToContents();
}

void MainWindow::refresh()
{
   setReady(false);
   yumIO->clear();
   yumIO->collect(yumReposUsed);
   packageTableModel->reset();
   setReady(true);
}

void MainWindow::reposChanged()
{
   setReady(false);
   RepoAuth::Checker  checker(this, yumConfIO);
   if (checker.quit) {
      quit();
      return;
   }
   yumIO->clear();
   yumIO->collect(yumReposUsed);
   packageTableModel->reset();
   if (packageTableConfig->isHidden()) {
      // This happens after starting with the -r option and then closing
      // the repository editor.
      packageTableConfig->setVisible(true);
      packageTableView->resizeColumnsToContents();
   }
   setReady(true);
}

void MainWindow::setSelected()
{
   packageTableModel->setSelected(packageTableView->selectedRows());
   stateChanged();
}

void MainWindow::setUpdateAll()
{
   packageTableModel->setUpdateAll();
   stateChanged();
}

void MainWindow::clearActions()
{
   packageTableModel->clearActions();
   stateChanged();
}

void MainWindow::apply()
{
   setReady(false);
   bool  done = yumIO->apply();
   if (done) {
      RepoAuth::Checker  checker(this, yumConfIO);
      if (checker.quit) {
         quit();
         return;
      }
      yumIO->clear();
      yumIO->collect(yumReposUsed);
      packageTableModel->reset();
   }
   setReady(true);
}

QSize MainWindow::sizeHint() const
{
   // When the window is drawn initially, it has only progress bars.  By
   // default, they're very narrow.  Specify a miminum size of 768 pixels
   // that makes them readable.
   return QMainWindow::sizeHint().expandedTo(QSize(768, 0));
}
