#include "PackageTable.h"
#include "PackageTableModel.h"
#include <QtCore/QVariant>
#include <QtGui/QMessageBox>
#include "Share.h"
#include "Colors.h"
#include "RepoBrowse.h"

PackageTable::Model::Model(YumIO*      yumIO,
                           QObject*    parent) 
   : QAbstractTableModel(parent),
     yumIO(yumIO),
     describeDialog(NULL),
     enabled(true),
     outOfDateIcon(Share::absolute("icons/out-of-date.png")),
     upToDateIcon(Share::absolute("icons/up-to-date.png")),
     notInstalledIcon(),
     noneIcon(Share::absolute("icons/noneBox.png")),
     updateIcon(Share::absolute("icons/updateBox.png")),
     installIcon(Share::absolute("icons/installBox.png")),
     removeIcon(Share::absolute("icons/removeBox.png")),
     suiteIcon(Share::absolute("icons/suite.png")),
     groupIcon(Share::absolute("icons/group.png")),
     packageIcon(Share::absolute("icons/package.png")),
     outOfDateColor(Colors::error()),
     outOfDateInvColor(Colors::errorDisabled()),
     upToDateColor(),
     upToDateInvColor(),
     notInstalledColor(Colors::minor()),
     notInstalledInvColor(Colors::minorDisabled())
{
}

PackageTable::Model::~Model()
{
}

int
PackageTable::Model::rowCount(const QModelIndex&) const
{
   return yumIO->count();
}

int
PackageTable::Model::columnCount(const QModelIndex&) const
{
   return NUM_COLUMNS;
}

QVariant
PackageTable::Model::data(const QModelIndex&  ix, 
                          int                 role) const
{
   if (!ix.isValid()) return QVariant();

   int  r = ix.row();
   int  c = ix.column();

   if (r >= yumIO->count()) return QVariant();
   if (c >= NUM_COLUMNS)    return QVariant();

   switch (role) {
   case Qt::DisplayRole:
      switch (c) {
      case NAME:      return (*yumIO)[r]->name;
      case VERREL:    return (*yumIO)[r]->verrel;
      case REPO:      return (*yumIO)[r]->repo;
      case ARCH:      return (*yumIO)[r]->arch;
      case BUILDTIME: return (*yumIO)[r]->buildTimeText;
      case SIZE:      return (*yumIO)[r]->sizeText;
      default:        return QVariant();
      }
   case Qt::DecorationRole:
      switch (c) {
      case ACTION:
         switch ((*yumIO)[r]->action) {
         case Package::NONE:    return noneIcon;
         case Package::UPDATE:  return updateIcon;
         case Package::INSTALL: return installIcon;
         case Package::REMOVE:  return removeIcon;
         default:               return QVariant();
         }
      case STATE:
         switch ((*yumIO)[r]->state) {
         case Package::NOT_INSTALLED: return QVariant();
         case Package::OUT_OF_DATE:   return outOfDateIcon;
         case Package::INSTALLED:     return upToDateIcon;
         default:                     return QVariant();
         }
      case KIND:
         switch ((*yumIO)[r]->kind) {
         case Package::SUITE:   return suiteIcon;
         case Package::GROUP:   return groupIcon;
         case Package::PACKAGE: return packageIcon;
         default:               return QVariant();
         }
      default:
         return QVariant();
      }
   case Qt::ToolTipRole:
      switch (c) {
      case ACTION:
         switch ((*yumIO)[r]->action) {
         case Package::NONE:    return QVariant();
         case Package::UPDATE:  return tr("Update package");
         case Package::INSTALL: return tr("Install package");
         case Package::REMOVE:  return tr("Remove package");
         default:               return QVariant();
         }
      case STATE:
         switch ((*yumIO)[r]->state) {
         case Package::OUT_OF_DATE:   return tr("Package is available to update"
                                                " an out-of-date version");
         case Package::NOT_INSTALLED: return tr("Package is available for"
                                                " installation");
         case Package::INSTALLED:     return tr("Package is installed and"
                                                " up-to-date");
         default:                     return QVariant();
         }
      case KIND:
         switch ((*yumIO)[r]->kind) {
         case Package::SUITE:   return tr("Suite of packages");
         case Package::GROUP:   return tr("Group of packages");
         case Package::PACKAGE: return tr("Ordinary package");
         default:               return QVariant();
         }
      case VERREL:
         switch ((*yumIO)[r]->state) {
         case Package::OUT_OF_DATE:
            {
               QString                result;
               const Package::Entry*  entry = (*yumIO)[r];
               for (; entry != NULL; entry = entry->next) {
                  switch (entry->state) {
                  case Package::OUT_OF_DATE:
                  case Package::NOT_INSTALLED:
                     result.append(tr("Available: %1\n").arg(entry->verrel));
                     break;
                  case Package::INSTALLED:
                     result.append(tr("Installed: %1\n").arg(entry->verrel));
                     break;
                  default:
                     break;
                  }
               }
               return result;
            }
         default:
            return QVariant();
         }
      default:
         return QVariant();
      }
   case Qt::TextColorRole:
      if (c <= STATE) return QVariant();
      switch ((*yumIO)[r]->state) {
      case Package::NOT_INSTALLED: 
         if (enabled) {
            return notInstalledColor;
         } else {
            return notInstalledInvColor;
         }
      case Package::OUT_OF_DATE:
         if (enabled) {
            return outOfDateColor;
         } else {
            return outOfDateInvColor;
         }
      case Package::INSTALLED:
         if (enabled) {
            return upToDateColor;
         } else {
            return upToDateInvColor;
         }
      default:
         return QVariant();
      }
      break;
   case RawValue:
      switch (c) {
      case ACTION:    return QVariant((*yumIO)[r]->action);
      case STATE:     return QVariant((*yumIO)[r]->state);
      case KIND:      return QVariant((*yumIO)[r]->kind);
      case SIZE:      return QVariant((*yumIO)[r]->size);
      case BUILDTIME: return QVariant((qlonglong)(*yumIO)[r]->buildTime);
      default:        return QVariant::fromValue((*yumIO)[r]);
      }
   default:
      return QVariant();
   }
}

QVariant  PackageTable::Model::headerData  (int              section, 
                                            Qt::Orientation  orientation, 
                                            int              role) const
{
   if (orientation == Qt::Vertical) return QVariant();

   switch (role) {
   case Qt::DisplayRole:
      switch (section) {
      case ACTION:    return tr("Action");
      case STATE:     return tr("State");
      case KIND:      return tr("Kind");
      case NAME:      return tr("Name");
      case VERREL:    return tr("Version");
      case REPO:      return tr("Repository");
      case ARCH:      return tr("Architecture");
      case BUILDTIME: return tr("Build Time");
      case SIZE:      return tr("Size");
      default:        return QVariant();
      }
   case Qt::ToolTipRole:
      switch (section) {
      case ACTION:    return tr("Action to be performed on the package, if any:"
                                " Update, Install, Remove");
      case STATE:     return tr("Indicates whether the package is Installed or"
                                " not and, if installed, whether it is"
                                " Up-to-date or not");
      case KIND:      return tr("Whether the package is an ordinary package,"
                                " a suite of packages, or a group of packages");
      case NAME:      return tr("Name of the package");
      case VERREL:    return tr("Version and release of the package as"
                                " version-release;\n"
                                "If the Epoch is non-zero, includes it as"
                                " epoch:version-release");
      case REPO:      return tr("Repository from which the most recent version"
                                " of the package is available");
      case ARCH:      return tr("Architecture for which the package is"
                                " applicable, or \"noarch\" for general-purpose"
                                " packages");
      case BUILDTIME: return tr("Date and time at which the package was"
                                " created");
      case SIZE:      return tr("Size of the package");
      default:        return QVariant();
      }
   default:
      return QVariant();
   }
}

void  PackageTable::Model::setYumConfIO (YumConfIO*  yumConfIO)
{
   this->yumConfIO = yumConfIO;
}

Package::State  PackageTable::Model::getState (int  row)
{
   return (*yumIO)[row]->state;
}

#define STAACT(s,a)  ((s) * 16 + (a))

void  PackageTable::Model::setAction (const QModelIndex&  ix,
                                      Package::Action     action)
{
   if (!ix.isValid()) return;

   int  r = ix.row();
   int  c = ix.column();

   if (r >= yumIO->count()) return;
   if (c >= NUM_COLUMNS)    return;

   (*yumIO)[r]->action = action;
}

void PackageTable::Model::toggleAction (const QModelIndex&  ix)
{
   if (!ix.isValid()) return;

   int  r = ix.row();

   if (r >= yumIO->count()) return;

   Package::Entry&  entry = *(*yumIO)[r];

   using namespace Package;

   switch (STAACT(entry.state, entry.action)) {
   case STAACT(NOT_INSTALLED, NONE):    entry.action = INSTALL; break;
   case STAACT(NOT_INSTALLED, INSTALL): entry.action = NONE;    break;
   case STAACT(OUT_OF_DATE,   NONE):    entry.action = UPDATE;  break;
   case STAACT(OUT_OF_DATE,   UPDATE):  entry.action = REMOVE;  break;
   case STAACT(OUT_OF_DATE,   REMOVE):  entry.action = NONE;    break;
   case STAACT(INSTALLED,     NONE):    entry.action = REMOVE;  break;
   case STAACT(INSTALLED,     REMOVE):  entry.action = NONE;    break;
   default:                             entry.action = NONE;    break;
   }

   QModelIndex  actionIx = createIndex(r, ACTION);
   emit dataChanged(actionIx, actionIx);
}

void PackageTable::Model::reset()
{
   QAbstractTableModel::reset();
}

void PackageTable::Model::describe (const QModelIndex&  ix,
                                    QWidget*            parent)
{
   if (!ix.isValid()) return;

   int  r = ix.row();

   if (r >= yumIO->count()) return;

   if (describeDialog == NULL) {
      describeDialog = new Describe::Dialog(parent);
   }
   describeDialog->config(yumIO, *(*yumIO)[r], yumConfIO);
   describeDialog->exec();
}

void PackageTable::Model::browse (const QModelIndex&  ix,
                                  QWidget*            parent)
{
   if (!ix.isValid()) return;

   int  r = ix.row();

   if (r >= yumIO->count()) return;

   RepoBrowse::browse(parent, (*yumIO)[r]->repo, yumIO, yumConfIO);
}

void PackageTable::Model::setEnabled (bool  enabled)
{
   this->enabled = enabled;
}

void PackageTable::Model::setSelected(const QModelIndexList&  indices)
{
   QModelIndexList::const_iterator  it;
   
   for (it = indices.begin(); it != indices.end(); ++it) {
      Package::Entry*  entry = (*yumIO)[it->row()];
      switch (entry->state) {
      case Package::OUT_OF_DATE:
         {
            entry->action = Package::UPDATE;
            QModelIndex  ix = createIndex(it->row(), ACTION);
            emit dataChanged(ix, ix);
         }
         break;
      case Package::NOT_INSTALLED:
         {
            entry->action = Package::INSTALL;
            QModelIndex  ix = createIndex(it->row(), ACTION);
            emit dataChanged(ix, ix);
         }
         break;
      case Package::INSTALLED:
         // DO NOTHING
         break;
      default:
         // DO NOTHING
         break;
      }
   }
}

void PackageTable::Model::setUpdateAll()
{
   Package::EntryList::const_iterator  it  = yumIO->begin();
   int                                 row = 0;
   for (; it != yumIO->end(); ++it, ++row) {
      if ((*it)->state == Package::OUT_OF_DATE) {
         (*it)->action = Package::UPDATE;
         QModelIndex  ix = createIndex(row, ACTION);
         emit dataChanged(ix, ix);
      }
   }
}

void PackageTable::Model::clearActions()
{
   Package::EntryList::const_iterator  it  = yumIO->begin();
   int                                 row = 0;
   for (; it != yumIO->end(); ++it, ++row) {
      if ((*it)->action != Package::NONE) {
         (*it)->action = Package::NONE;
         QModelIndex  ix = createIndex(row, ACTION);
         emit dataChanged(ix, ix);
      }
   }
}
