#include "PackageTableView.h"
#include "PackageTableModel.h"
#include <QtGui/QHeaderView>
#include <QtCore/QObject>
#include <QtCore/QEvent>
#include <QtGui/QMenu>

PackageTable::View::View(Model*       model,
                         StateFilter  stateFilter,
                         bool         groupsFilter,
                         YumIO*       yumIO,
                         QWidget*     parent)
#if QT_VERSION >= 0x040200
   : HTableView(QStyleOptionViewItem::Left, 10, INT_MAX, parent), 
#else
   : HTableView(QStyleOptionViewItem::Bottom, 10, INT_MAX, parent), 
#endif
     model(model),
     actionContextMenu(NULL),
     otherContextMenu(NULL)
{
   sortedModel = new SortFilterProxyModel(stateFilter, groupsFilter, yumIO,
                                          parent);
   sortedModel->setSourceModel(model);
   setModel(sortedModel);

   setShowGrid(false);
   setGridStyle(Qt::NoPen);
   setSelectionBehavior(QAbstractItemView::SelectRows);
   setAlternatingRowColors(true);
   setContextMenuPolicy(Qt::CustomContextMenu);

   verticalHeader()->hide();
   setHorizontalHeader(new QHeaderView(Qt::Horizontal, this));

   horizontalHeader()->setSelectionBehavior(QAbstractItemView::SelectColumns);
   horizontalHeader()->setMovable(true);
   horizontalHeader()->setClickable(true);

   sortByColumn(KIND);
   sortedModel->setFilterKeyColumn(KIND);

   connect(horizontalHeader(), SIGNAL(sectionClicked(int)), 
           this, SLOT(handleHorizontalHeaderClicked(int)));

   connect(this, SIGNAL(clicked(const QModelIndex&)),
           this, SLOT(handleActivated(const QModelIndex&)));
   connect(this, SIGNAL(customContextMenuRequested(const QPoint&)),
           this, SLOT(handleCustomContextMenuRequested(const QPoint&)));
}

void  PackageTable::View::setMode(Mode  mode)
{
   sortedModel->setMode(mode);
}

void  PackageTable::View::setFilter(bool  filter)
{
   sortedModel->setFilter(filter);
}

void  PackageTable::View::filtersearch(const QString&  regexp)
{
   sortedModel->setFilterRegExp(regexp);
}

void  PackageTable::View::stateFilterChanged(StateFilter  stateFilter)
{
   sortedModel->setStateFilter(stateFilter);
}

void  PackageTable::View::groupsFilterChanged(bool  groupsFilter)
{
   sortedModel->setGroupsFilter(groupsFilter);
}

void  PackageTable::View::changeEvent (QEvent*  event)
{
   HTableView::changeEvent(event);
   if (event->type() == QEvent::EnabledChange) {
      model->setEnabled(isEnabled());
   }
}

QModelIndexList  PackageTable::View::selectedRows() const
{
   QModelIndexList            result = HTableView::selectedRows();
   QModelIndexList::iterator  it     = result.begin();

   for (; it != result.end(); ++it) {
      *it = sortedModel->mapToSource(*it);
   }

   return result;
}

void  PackageTable::View::resizeColumnsToContents()
{
   // Disable filtering briefly while resizing columns so they'll be big
   // enough even if the user changes the sort filter.
   bool  oldFilter = sortedModel->getFilter();
   sortedModel->setFilter(false);
   HTableView::resizeColumnsToContents();
   sortedModel->setFilter(oldFilter);
}

void  PackageTable::View::describe()
{
   model->describe(sortedModel->mapToSource(currentIndex()), this);
}

void  PackageTable::View::handleHorizontalHeaderClicked(int  column)
{
   sortByColumn(column);
   sortedModel->setFilterKeyColumn(column);
}

void  PackageTable::View::handleActivated(const QModelIndex&  ix)
{
   if (ix.column() == ACTION) {
      QModelIndex  modelIx = sortedModel->mapToSource(ix);
      model->toggleAction(modelIx);
      emit stateChanged(modelIx.row());
   }
}

void  PackageTable::View::handleCustomContextMenuRequested(const QPoint&  pos)
{
   QModelIndex  ix = sortedModel->mapToSource(indexAt(pos));

   if (ix.column() == ACTION) {
      if (actionContextMenu == NULL) {
         actionContextMenu = new ActionContextMenu(model);
      }
      actionContextMenu->exec(ix, viewport()->mapToGlobal(pos));
   } else {
      if (otherContextMenu == NULL) {
         otherContextMenu = new OtherContextMenu(model, this);
      }
      otherContextMenu->exec(ix, viewport()->mapToGlobal(pos));
   }
}

void  PackageTable::View::currentChanged(const QModelIndex&  current,
                                         const QModelIndex&  previous)
{
   int  currentRow  = sortedModel->mapToSource(current).row();
   int  previousRow = sortedModel->mapToSource(previous).row();
   if (currentRow != previousRow) {
      emit currentChanged(currentRow, previousRow);
   }
}
