#include "Progress.h"
#include <QtCore/QMetaType>
#include "globals.h"

#undef DEBUG_PROGRESS
#undef DEBUG_PROGRESS_DETAIL

bool  Progress::registeredSteps;

Progress::Progress(HProgressBar*  totalProgress,
                   HProgressBar*  itemProgress)
{
   if (!registeredSteps) {
      qRegisterMetaType<Steps>("Steps");
   }

   minion = new ProgressMinion(totalProgress, itemProgress);

   minion->moveToThread(application->thread());

   singleThread = (minion->thread() == thread());

   connect(this,   SIGNAL(startTotalSig(QString, Steps)),
           minion, SLOT(startTotal(QString, Steps)));
   connect(this,   SIGNAL(startItemSig(QString, Steps, Steps)),
           minion, SLOT(startItem(QString, Steps, Steps)));
   connect(this,   SIGNAL(changeItemActsSig(Steps)),
           minion, SLOT(changeItemActs(Steps)));
   connect(this,   SIGNAL(setItemValueSig(Steps, Steps)),
           minion, SLOT(setItemValue(Steps, Steps)));
   connect(this,   SIGNAL(endItemActsSig(Steps)),
           minion, SLOT(endItemActs(Steps)));
   connect(this,   SIGNAL(setTotalValueSig(Steps)),
           minion, SLOT(setTotalValue(Steps)));
   connect(this,   SIGNAL(resetSig()),
           minion, SLOT(reset()));

   reset();
}

Progress::~Progress()
{
   reset();

   minion->deleteLater();
}

void Progress::startTotal(QString  activity,
                          Steps    totalSteps)
{
   totalValue = 0;
   totalTail  = totalSteps;

#if defined(DEBUG_PROGRESS)
   printf("progress: startTotal: totalSteps = %llu %s\n",
          totalSteps, activity.toStdString().c_str());
#endif

   if (singleThread) {
      lastGuiUpdate.start();
   }

   emit startTotalSig(activity, totalTail);

   updateGui();
}

void Progress::startItem(QString  activity,
                         Steps    totalSteps,
                         Steps    itemSteps)
{
   totalValue = itemTotalTail;

   itemTotalStart = itemTotalTail;
   itemTotalTail  = itemTotalStart + totalSteps;

   itemValue = 0;
   itemTail  = itemSteps;

#if defined(DEBUG_PROGRESS)
   printf("progress: startItem: total = %llu:%llu (%llu) %s\n",
          itemTotalStart, itemTotalTail, totalSteps, 
          activity.toStdString().c_str());
#endif

   lastSetValue.start();

   emit startItemSig(activity, totalValue, itemTail);

   updateGui();
}

void Progress::setupItemActs (Steps  totalSteps,
                              Steps  numActs)
{
   actsTotalSteps = totalSteps;
   actsTotalTail  = itemTotalTail + totalSteps;
   this->numActs  = numActs;
   prevAct        = 0;

#if defined(DEBUG_PROGRESS)
   printf("progress: setupItemActs: total = %llu:%llu (%llu), numActs = %llu\n",
          actsTotalTail - actsTotalSteps, actsTotalTail, actsTotalSteps, 
          numActs);
#endif
}

void Progress::changeItemActs (Steps  numActs)
{
   this->numActs = numActs;

   Steps  actsTotalStart    = actsTotalTail - actsTotalSteps;
   Steps  actsTotalProgress = (numActs == 0
                               ? actsTotalSteps
                               : prevAct * actsTotalSteps / numActs);
   itemTotalTail = actsTotalStart + actsTotalProgress;

   totalValue = itemTotalTail;

#if defined(DEBUG_PROGRESS)
   printf("progress: changeItemActs: numActs = %llu,"
          " totalProgress = %llu -> %llu\n", 
          numActs, totalValue, itemTotalTail);
#endif

   emit changeItemActsSig(totalValue);

   updateGui();
}

void Progress::startItemAct (QString  activity,
                             Steps    itemSteps,
                             Steps    act)
{
   if (act > numActs) act = numActs;

#if defined(DEBUG_PROGRESS)
   printf("progress: startItemAct: acts = %llu:%llu (%llu) %s\n",
          prevAct, act, act - prevAct,
          activity.toStdString().c_str());
#endif

   Steps  prevTotalProgress = (numActs == 0
                               ? actsTotalSteps
                               : prevAct * actsTotalSteps / numActs);
   Steps  thisTotalProgress = (numActs == 0
                               ? actsTotalSteps
                               : act * actsTotalSteps / numActs);
   startItem(activity, thisTotalProgress - prevTotalProgress, itemSteps);
   prevAct = act;
}

void Progress::setItemValue(Steps  value)
{
   if (value > itemTail) {
      value = itemTail;
   }

   // Without this check, the minion's event queue is bombarded by so many
   // events that it can't keep up.  It ends up with a huge backlog of events
   // by the time the whole activity is complete and it is destroyed.  The
   // result is that the progress bars indicate only a fraction of the total
   // progress being complete when suddenly the activity is done.
   //
   // This check rate limits the most frequent events so that the minion can
   // keep up.  No more than one item value event every 10ms seems to allow it
   // to keep up pretty well.  Note that this code lets the last item value
   // event through no matter what.
   if (lastSetValue.elapsed() < 10
       && (value != itemTail || itemValue == itemTail)) {
      return;
   }
   lastSetValue.start();

   itemValue = value;

   Steps  itemLength      = itemTail;
   Steps  itemTotalLength = itemTotalTail - itemTotalStart;

   totalValue = (itemLength == 0
                 ? itemTotalStart + itemTotalLength
                 : itemTotalStart + value * itemTotalLength / itemLength);

#if defined(DEBUG_PROGRESS_DETAIL)
   printf("progress: setItemValue:"
          " total = (%llu,%llu,%llu), item = (%llu,%llu,%llu)\n",
          0ull,
          totalValue,
          totalTail,
          0ull,
          value,
          itemTail);
#endif

   emit setItemValueSig(totalValue, value);

   updateGui();
}

void Progress::endItemActs()
{
   // Sometimes the trailing act that increment the number of acts to numActs
   // will not be performed.  In case that happens, ensure the total progress
   // is updated.  This is harmless if the last act was performed.
   itemTotalStart = actsTotalTail;
   itemTotalTail  = actsTotalTail;

   totalValue = itemTotalStart;

#if defined(DEBUG_PROGRESS)
   printf("progress: endItemActs: total = %llu\n", actsTotalTail);
#endif

   emit endItemActsSig(totalValue);

   updateGui();
}

void Progress::setTotalValue (Steps  value)
{
   if (value > totalTail) {
      value = totalTail;
   }

   // Without this check, the minion's event queue is bombarded by so many
   // events that it can't keep up.  It ends up with a huge backlog of events
   // by the time the whole activity is complete and it is destroyed.  The
   // result is that the progress bars indicate only a fraction of the total
   // progress being complete when suddenly the activity is done.
   //
   // This check rate limits the most frequent events so that the minion can
   // keep up.  No more than one item value event every 10ms seems to allow it
   // to keep up pretty well.  Note that this code lets the last item value
   // event through no matter what.
   if (lastSetValue.elapsed() < 10
       && (value != totalTail || totalValue == totalTail)) {
      return;
   }
   lastSetValue.start();

   totalValue = value;

#if defined(DEBUG_PROGRESS)
   printf("progress: setTotalValue: value = %llu\n", value);
#endif

   emit setTotalValueSig(value);

   updateGui();
}

void Progress::reset()
{
   totalTail = 1;

   itemTotalStart = 0;
   itemTotalTail  = 0;

   itemTail = 1;

   emit resetSig();

   updateGui();
}

void Progress::updateGui()
{
   if (singleThread) {
      // Because there's no other thread updating the GUI< occasionally do
      // so manually here.
      if (lastGuiUpdate.elapsed() >= 10) {
         QEventLoop  eventLoop;
         eventLoop.processEvents(QEventLoop::ExcludeUserInputEvents);
         lastGuiUpdate.start();
      }
   }
}
