#ifndef Progress_h
#define Progress_h

#include "HProgressBar.h"
#include "ProgressMinion.h"
#include <QtCore/QObject>
#include <QtCore/QTime>

//
// This class controls two progress bars: total progress and item progress.
// It is for large-scale operations which are composed of a number of smaller
// operations, called items.  The total progress indicates progress for the
// whole large-scale operation.  The item progress indicates progress for
// each smaller operation.
//
// The large-scale operation is divided into a total number of steps.  Each
// item is divided into a number of steps, but they need have NO RELATION to
// the total number of steps.  That is, a large-scale operation may have 1000
// steps.  And each item also may have 1000 steps, or any other number.  As
// progress is made in an item, the item progress is scaled to its part of
// the total operation.
//
// To start a large-scale operation, call startTotal, specifying as activity
// the name of the large-scale operation, and as totalSteps the number of
// steps in the large-scale operation.
//
// To start each smaller operation, call either startItem In either case,
// activity indicates the name of the item, totalSteps indicates the number
// of total steps for the item, and itemSteps indicates the number of item
// steps for the item.
//
// Regardless of whether you use simple items or act items, the progress for
// each item is updated by calling setItemValue.  Its value is the amount
// of progress, measured in itemSteps for the current item.
//
// Sometimes a block of total steps may be reserved for a number of items,
// frequently determined dynamically.  In this case, each item is divided
// into a number of acts (although the degenerate case where an item is
// divided into only single act is useful, too).  The number of acts need
// have NO RELATION to the number of total steps or the number of item
// steps for each item in the block.
//
// To use acts, first call setupItemActs.  This does not actually start any
// item; it just specifies number of total steps in the block reserved for
// the items, and the number of acts.  Then, for each item, call
// startItemAct, specifying the name of the activity, the number of item
// steps for the act, and the number of the act (which should be one or more
// greater than the value passed in the previous call).  When all acts have
// been completed, call endItemActs.
//
// When the progress bar no longer is being used to display any progress,
// call reset.
//
// For example:
//    obj.startTotal("long operation", 100 + 50 + 200);
//    obj.startItem("part 1", 100, 5000);
//    for (i = 0; i < 5000; i++) {
//       ...
//       obj.setItemValue(i);
//    }
//    obj.startItem("part 2", 50, 3000);
//    for (i = 0; i < 3000; i++) {
//       ...
//       obj.setItemValue(i);
//    }
//    obj.setupItemActs(200, numActs * 5);
//    for (act = 0; act < numActs; act += 5) {
//       obj.startItemAct(QString("part %1").arg(act+2), 1000, act);
//       for (i = 0; i < 1000; i++) {
//          ...
//          obj.setItemValue(i);
//       }
//    }
//    obj.endItemActs();
//    obj.reset();
// And the above example shows progress broken down as follows 
// (with numActs = 3):
//    total = |0.....99|100....149|150...216|217...282|283...349|
//    acts  =                     |0...... 4|5.......9|10.....14|
//    items = |0...4999|0.....2999|0.....999|0.....999|0.....999| 
// The total progress bar tracks the first row.  The item progress bar
// tracks the last row, and thus restarts for each item.  No progress bar
// tracks the middle row; it's shown only so that the scaling relationships
// between the various numbers: total steps, acts, item steps.
//
// Sometimes, it may be useful just to use the total progress bar and leave
// the item progress bar disabled.  In that case, and only in that case, 
// setTotalValue can be used.  For example:
//    obj.startTotal("long operation", 100);
//    for (i = 0; i < 100; i++) {
//       ...
//       obj.setTotalValue(i);
//    done
//    obj.reset();
//

class Progress : public QObject {
   Q_OBJECT
public:
   explicit Progress(HProgressBar*  totalProgress,
                     HProgressBar*  itemProgress);
   ~Progress();
public:
   typedef unsigned long long  Steps;

   void startTotal     (QString  activity,
                        Steps    totalSteps);
   void startItem      (QString  activity,
                        Steps    totalSteps,
                        Steps    itemSteps = 1);
   void setupItemActs  (Steps    totalSteps,
                        Steps    numActs);
   void changeItemActs (Steps    numActs);
   void startItemAct   (QString  activity,
                        Steps    itemSteps,
                        Steps    act);
   void setItemValue   (Steps    value);
   void endItemActs    ();
   void setTotalValue  (Steps    value);
   void reset          ();
signals:
   void startTotalSig     (QString  activity,
                           Steps    totalTail);
   void startItemSig      (QString  activity,
                           Steps    totalValue,
                           Steps    itemTail);
   void changeItemActsSig (Steps    totalValue);
   void setItemValueSig   (Steps    totalValue,
                           Steps    value);
   void endItemActsSig    (Steps    totalValue);
   void setTotalValueSig  (Steps    value);
   void resetSig          ();
protected:
   void updateGui         ();
protected:
   ProgressMinion*     minion;
   bool                singleThread;
   static bool         registeredSteps;
   QTime               lastSetValue;
   QTime               lastGuiUpdate;
protected:
   Steps  totalValue;     // measured in total steps
   Steps  totalTail;      // measured in total steps

   Steps  itemTotalStart; // measured in total steps
   Steps  itemTotalTail;  // measured in total steps

   Steps  itemValue;      // measured in item steps
   Steps  itemTail;       // measured in item steps

   Steps  actsTotalSteps; // measured in total steps
   Steps  actsTotalTail;  // measured in total steps
   Steps  numActs;        // measured in acts
   Steps  prevAct;        // meausred in acts
};

// This class is just a handy way to declare a single object that, when
// destroyed (e.g. on exit from a routine) will reset the Progress.
class ProgressUser {
public:
   explicit ProgressUser(Progress*&     pg,
                         HProgressBar*  totalProgress,
                         HProgressBar*  itemProgress) 
      : pg(pg)
   {
      pg = new Progress(totalProgress, itemProgress);
   }
   ~ProgressUser() 
   { 
      delete pg;
      pg = NULL;
   }
protected:
   Progress*&  pg;
};

#endif
