#include "Python.h"
#include <signal.h>

bool       Python::initialized = false;
PyObject*  Python::main_dict   = NULL;

Python::~Python()
{
   if (initialized) {
#if defined(PYTHON_SUPPORTS_GIL)
      PyGILState_Ensure();
#endif
      Py_Finalize();

      initialized = false;
   }
}

bool Python::initialize()
{
   if (!initialized) {
      initialized = true;

      Py_Initialize();

      PyEval_InitThreads();

      PyGILStateWrapper  gsw;

      py("import signal\n"
         "defaultSigIntSigaction = signal.getsignal(signal.SIGINT)");

      PyObject*  main_mod = PyImport_ImportModule("__main__");
      if (main_mod == NULL) {
         PyErr_Clear();
         error("Internal error: cannot find python __main__ module\n");
         return false;
      }
      main_dict = PyModule_GetDict(main_mod);
      if (main_dict == NULL) {
         error("Internal error: cannot find python __main__ dict\n");
         return false;
      }

      PyEval_ReleaseLock();
   }

   return true;
}

void  Python::py (const char*  str)
{
   char*  copy = new char[strlen(str)+1];
   strcpy(copy, str);
   py(copy);
   delete [] copy;
}

PyObject*  Python::pyr (const char*  str)
{
   char*  copy = new char[strlen(str)+1];
   strcpy(copy, str);
   PyObject*  result = pyr(copy);
   delete [] copy;
   return result;
}

void  Python::py (char*  str)
{
   int  status = PyRun_SimpleString(str);
   if (status == -1 && PyErr_Occurred()) handleException();
}

PyObject*  Python::pyr (char*  str)
{
   PyObject*  result = PyRun_String(str, Py_eval_input, 
                                    main_dict, main_dict);
   if (result == NULL && PyErr_Occurred()) handleException();
   return result;
}

void Python::pyf(const char* format, ...)
{
   char buffer[10240];

   va_list  ap;

   va_start(ap, format);
   vsprintf(buffer, format, ap);
   va_end(ap);

   py(buffer);
}

PyObject* Python::pyrf(const char* format, ...)
{
   char buffer[10240];

   va_list  ap;

   va_start(ap, format);
   vsprintf(buffer, format, ap);
   va_end(ap);

   return pyr(buffer);
}

char*  Python::pyString(PyObject*  obj)
{
   if (obj == NULL) {
      error("Internal error: pyString passed NULL python object\n");
      return NULL;
   }

   if (obj == Py_None) return NULL;

   char*  result = PyString_AsString(obj);
   if (result == NULL && PyErr_Occurred()) handleException();
   return result;
}

PyObject* Python::pyCall(const char*  name,
                         const char*  format,
                         ...)
{
   char*  nameCopy   = new char[strlen(name)+1];
   char*  formatCopy = new char[strlen(format)+1];
   strcpy(nameCopy,   name);
   strcpy(formatCopy, format);

   va_list  ap;
   va_start(ap, format);
   PyObject*  result = pyVCall(nameCopy, formatCopy, ap);
   va_end(ap);

   delete [] formatCopy;
   delete [] nameCopy;
   return result;
}

PyObject* Python::pyCall(char*  name,
                         char*  format,
                         ...)
{
   va_list  ap;
   va_start(ap, format);
   PyObject*  result = pyVCall(name, format, ap);
   va_end(ap);

   return result;
}

PyObject* Python::pyVCall(char*    name,
                          char*    format,
                          va_list  ap)
{
   PyObject*  func = pyLookup(name);
   if (func == NULL) {
      char  buffer[1024];
      sprintf(buffer, "Internal error: %s not found\n", name);
      error(buffer);
      return NULL;
   }
   if (!PyCallable_Check(func)) {
      char  buffer[1024];
      sprintf(buffer, "Internal error: %s not callable\n", name);
      error(buffer);
      return NULL;
   }

   PyObject*  args = Py_VaBuildValue(format, ap);

   PyObject*  result = PyEval_CallObject(func, args);

   if (result == NULL && PyErr_Occurred()) handleException();

   Py_DECREF(args);

   return result;
}

PyObject*  Python::pyMethodCall(PyObject*    self,
                                const char*  name,
                                const char*  format,
                                ...)
{
   char*  nameCopy   = new char[strlen(name)+1];
   char*  formatCopy = new char[strlen(format)+1];
   strcpy(nameCopy,   name);
   strcpy(formatCopy, format);

   va_list  ap;
   va_start(ap, format);
   PyObject*  result = pyVMethodCall(self, nameCopy, formatCopy, ap);
   va_end(ap);

   delete [] formatCopy;
   delete [] nameCopy;
   return result;
}

PyObject*  Python::pyMethodCall(PyObject*  self,
                                char*      name,
                                char*      format,
                                ...)
{
   va_list  ap;
   va_start(ap, format);
   PyObject*  result = pyVMethodCall(self, name, format, ap);
   va_end(ap);

   return result;
}

PyObject*  Python::pyVMethodCall(PyObject*  self,
                                 char*      name,
                                 char*      format,
                                 va_list    ap)
{
   PyObject*  func = PyObject_GetAttrString(self, name);
   if (func == NULL) {
      char  buffer[1024];
      sprintf(buffer, "Internal error: method %s not found\n", name);
      error(buffer);
      return NULL;
   }
   if (!PyCallable_Check(func)) {
      char  buffer[1024];
      sprintf(buffer, "Internal error: method %s not callable\n", name);
      error(buffer);
      return NULL;
   }

   PyObject*  args = Py_VaBuildValue(format, ap);

   PyObject*  result = PyEval_CallObject(func, args);

   if (result == NULL && PyErr_Occurred()) handleException();

   Py_DECREF(func);
   Py_DECREF(args);

   return result;
}

PyObject*  Python::pyLookup(const char*  name)
{
   PyObject*  current = NULL;

   for (;;) {
      const char*  sep = strchr(name, '.');
      if (sep == NULL) {
         sep = name + strlen(name);
      }

      char*  part = (char*)malloc(sep - name + 1);
      strncpy(part, name, sep - name);
      part[sep - name] = '\0';

      if (current == NULL) {
         current = PyDict_GetItemString(main_dict, part);
         if (current == NULL) {
            PyErr_Clear();
            current = PyImport_ImportModule(part);
         }
         if (current == NULL) {
            PyErr_Clear();
            char  buffer[1024];
            sprintf(buffer, "Internal error: %s not found\n", part);
            error(buffer);
            return NULL;
         }
      } else {
         if (PyModule_Check(current)) {
            PyObject*  dict = PyModule_GetDict(current); 
            if (dict == NULL) {
               char  buffer[1024];
               sprintf(buffer, "Internal error: module %s lacks dictionary\n",
                       part);
               error(buffer);
               return NULL;
            }
            current = PyDict_GetItemString(dict, part);
            if (current == NULL) {
               char  buffer[1024];
               sprintf(buffer, "Internal error: %s not found\n", part);
               error(buffer);
               return NULL;
            }
         } else {
            current = PyDict_GetItemString(current, part);
            if (current == NULL) {
               char  buffer[1024];
               sprintf(buffer, "Internal error: %s not found\n", part);
               error(buffer);
               return NULL;
            }
         }
      }

      free(part);

      if (*sep == '\0') break;

      name = sep + 1;
   }

   return current;
}

void Python::pyExceptExpect(PyObject*  xcpt)
{
   if (xcpt == NULL) {
      error("Internal error: pyExceptExpect called with NULL\n");
   }

   xcpts.push_back(xcpt);
}

void Python::pyExceptClear()
{
   xcpts.clear();
   PyErr_Clear();
}

bool  Python::pyExceptMatches (PyObject*    result,
                               PyObject*    xcpt)
{
   return (result == NULL
           && PyErr_Occurred()
           && PyErr_ExceptionMatches(xcpt));
}

PyObject*  Python::pyExceptValue ()
{
   PyObject*  exception;
   PyObject*  v;
   PyObject*  tb;

   PyErr_Fetch(&exception, &v, &tb);
   PyErr_NormalizeException(&exception, &v, &tb);

   return v;
}

PyObject*  Python::pyExceptValuef ()
{
   PyObject*  xcpt = pyExceptValue();
   if (PyObject_HasAttrString(xcpt, "value")) {
      // In yum >= 3.1.0, the string was moved to the value attribute to hush
      // some weird errors in python 2.5.  And this change was backported to yum
      // 3.0.1 in openSUSE 10.2 because it uses python 2.5, too.
      return PyObject_GetAttrString(xcpt, "value");
   } else {
      return pyCall("__builtins__.str", "(O)", xcpt);
   }
}

PyObject*  Python::pyValue (const char*  dict,
                            const char*  name)
{
   char*  dictCopy = new char[strlen(dict)+1];
   char*  nameCopy = new char[strlen(name)+1];
   strcpy(dictCopy, dict);
   strcpy(nameCopy, name);
   PyObject*  result = pyValue(dictCopy, nameCopy);
   delete [] nameCopy;
   delete [] dictCopy;
   return result;
}

PyObject*  Python::pyValue (char*  dict,
                            char*  name)
{
   PyObject*  dictObj = pyLookup(dict);
   if (dictObj == NULL) {
      char  buffer[1024];
      sprintf(buffer, "Internal error: %s not found\n", dict);
      error(buffer);
      return NULL;
   }
   if (PyModule_Check(dictObj)) {
      dictObj = PyModule_GetDict(dictObj); 
      if (dictObj == NULL) {
         char  buffer[1024];
         sprintf(buffer, "Internal error: module %s lacks dictionary\n", dict);
         error(buffer);
         return NULL;
      }
   }

   PyObject*  result = PyDict_GetItemString(dictObj, name);
   if (result == NULL && PyErr_Occurred()) handleException();
   return result;
}

void Python::pyAssign (const char*  dict,
                       const char*  name,
                       PyObject*    obj)
{
   char*  dictCopy = new char[strlen(dict)+1];
   char*  nameCopy = new char[strlen(name)+1];
   strcpy(dictCopy, dict);
   strcpy(nameCopy, name);
   pyAssign(dictCopy, nameCopy, obj);
   delete [] nameCopy;
   delete [] dictCopy;
}

void Python::pyAssign (char*      dict,
                       char*      name,
                       PyObject*  obj)
{
   PyObject*  dictObj = pyLookup(dict);
   if (dictObj == NULL) {
      char  buffer[1024];
      sprintf(buffer, "Internal error: %s not found\n", dict);
      error(buffer);
      return;
   }
   if (PyModule_Check(dictObj)) {
      dictObj = PyModule_GetDict(dictObj); 
      if (dictObj == NULL) {
         char  buffer[1024];
         sprintf(buffer, "Internal error: module %s lacks dictionary\n", dict);
         error(buffer);
         return;
      }
   }
   int  status = PyDict_SetItemString(dictObj, name, obj);
   if (status == -1 && PyErr_Occurred()) {
      handleException();
   } else if (status == -1) {
      char  buffer[1024];
      sprintf(buffer, "Internal error: cannot assign %s.%s\n", dict, name);
      error(buffer);
      return;
   }
}

long Python::pyVersion ()
{
   return PyInt_AsLong(pyValue("sys", "hexversion"));
}

#if defined(DEBUG_PYTHON)
void Python::pyAssignGlobal (const char*  name,
                             PyObject*    obj)
{
   char*  copy = new char[strlen(name)+1];
   strcpy(copy, name);
   pyAssignGlobal(copy, obj);
   delete [] copy;
}

void Python::pyAssignGlobal (char*      name,
                             PyObject*  obj)
{
   PyDict_SetItemString(main_dict, name, obj);
}

void Python::pyDebug (PyObject*  obj)
{
   PyErr_Clear();
   pyAssignGlobal("dbg", obj);
   py("import sys\n"
      "sys.stdout = sys.__stdout__\n"
      "sys.stderr = sys.__stderr__\n"
      "import pdb\n"
      "pdb.set_trace()");
   PyErr_Clear();
}
#endif

void Python::setEnablePythonSIGINTHandler(bool value)
{
   if (value) {
      py("signal.signal(signal.SIGINT, defaultSigIntSigaction)");
   } else {
      py("signal.signal(signal.SIGINT, signal.SIG_DFL)");
   }
}

void Python::handleException()
{
   std::list<PyObject*>::const_iterator  it = xcpts.begin();
   for (; it != xcpts.end(); ++it) {
      if (PyErr_ExceptionMatches(*it)) {
         // Ignore it.  The caller intends to check for it.
         exceptionExpected();
         return;
      }
   }

   exceptionError();
}

void Python::exceptionExpected()
{
   // DO NOTHING
   // This is just here so that a derived class can override it if so desired.
}

void Python::exceptionError()
{
   PyObject*  exception;
   PyObject*  v;
   PyObject*  tb;
   PyErr_Fetch(&exception, &v, &tb);
   PyErr_NormalizeException(&exception, &v, &tb);
   PyErr_Display(exception, v, tb);

   PyErr_Clear();
}
