#include "RepoBuffet.h"
#include <QtCore/QRegExp>
#include "stringmacros.h"

#define NAMEKEY_PREFIX      "/buffet?"
#define NAME_PREFIX         "="
#define PASSKEY_PREFIX      "&"
#define PASS_PREFIX         "="
#define PATH_PREFIX         "&path="

#define NAMEKEY_PREFIX_REGEXP  "/buffet\\?"
#define NAME_PREFIX_REGEXP     "="
#define PASSKEY_PREFIX_REGEXP  "&"
#define PASS_PREFIX_REGEXP     "="
#define PATH_PREFIX_REGEXP     "&path="

#define ARG_REGEXP          "[^&\\s]*"

static inline QChar  toHex(int  c)
{
   if (0 <= c && c <= 9) {
      return QChar('0' + c);
   } else if (10 <= c && c <= 15) {
      return QChar('a' + c - 10);
   } else {
      return QChar('?');
   }
}

static inline int  fromHex(QChar  ch)
{
   int  c = ch.unicode();

   if ('0' <= c && c <= '9') {
      return c - '0';
   } else if ('a' <= c && c <= 'f') {
      return c - 'a' + 10;
   } else if ('A' <= c && c <= 'F') {
      return c - 'A' + 10;
   } else {
      return -1;
   }
}

static QString  httpUnprotect (QString  s)
{
   QRegExp  hexCode("%[0-9a-fA-F][0-9a-fA-F]");

   int  ix = 0;
   for (;;) {
      ix = s.indexOf(hexCode, ix);
      if (ix == -1) break;
      
      int  value = fromHex(s[ix+1])*16 + fromHex(s[ix+2]);

      s.replace(ix, 3, (QChar)value);

      ix++;
   }
   
   return s;
}

static QString  httpProtect (QString  s)
{
   QRegExp  special("["
                    "$&+,/:;=?@"       // http reserved characters
                    " \"<>#%{}|\\\\^~" // http "unsafe" characters
                    "]"
                    "|\\[|\\]");       // http "unsafe" characters

   int  ix = 0;
   for (;;) {
      ix = s.indexOf(special, ix);
      if (ix == -1) break;
      
      QString  hexCode("%");
      hexCode.append(toHex(s[ix].unicode() / 16));
      hexCode.append(toHex(s[ix].unicode() % 16));

      s.replace(ix, 1, hexCode);

      ix += 3;
   }
   
   return s;
}

bool  Repo::Buffet::getAuth (const QString&  url,
                             QString&        nameKey,
                             QString&        name,
                             QString&        passwordKey,
                             QString&        password)
{
   int  nameKeyPrefixIndex = url.indexOf(NAMEKEY_PREFIX);
   if (nameKeyPrefixIndex == -1) return false;
   int  nameKeyIndex       = nameKeyPrefixIndex + STRLEN(NAMEKEY_PREFIX);

   int  namePrefixIndex    = url.indexOf(NAME_PREFIX, nameKeyIndex);
   if (namePrefixIndex == -1) return false;
   int  nameIndex          = namePrefixIndex + STRLEN(NAME_PREFIX);

   int  passKeyPrefixIndex = url.indexOf(PASSKEY_PREFIX, nameIndex);
   if (passKeyPrefixIndex == -1) return false;
   int  passKeyIndex       = passKeyPrefixIndex + STRLEN(PASSKEY_PREFIX);

   int  passPrefixIndex    = url.indexOf(PASS_PREFIX, passKeyIndex);
   if (passPrefixIndex == -1) return false;
   int  passIndex          = passPrefixIndex + STRLEN(PASS_PREFIX);

   int  pathPrefixIndex    = url.indexOf(PATH_PREFIX, passIndex);
   if (pathPrefixIndex == -1) return false;

   nameKey     = url.mid(nameKeyIndex, namePrefixIndex    - nameKeyIndex);
   name        = url.mid(nameIndex,    passKeyPrefixIndex - nameIndex);
   passwordKey = url.mid(passKeyIndex, passPrefixIndex    - passKeyIndex);
   password    = url.mid(passIndex,    pathPrefixIndex    - passIndex);

   nameKey     = httpUnprotect(nameKey);
   name        = httpUnprotect(name);
   passwordKey = httpUnprotect(passwordKey);
   password    = httpUnprotect(password);

   return true;
}

bool  Repo::Buffet::getAuth (const Repo::Entry&  entry,
                             QString&            nameKey,
                             QString&            name,
                             QString&            passwordKey,
                             QString&            password)
{
   // I considered having this check for multiple urls that used buffet and
   // had different passwords, and returning false in that case to disable
   // the authentication fields.  But I worried that that would just be
   // irritating to the user, so I didn't do it.

   return (getAuth(entry.urls, 
                   nameKey, name, passwordKey, password)
           || getAuth(entry.mirrorListUrl, 
                      nameKey, name, passwordKey, password));
}

QString  Repo::Buffet::replaceAuth (QString         str,
                                    const QString&  nameKey,
                                    const QString&  name,
                                    const QString&  passwordKey,
                                    const QString&  password)
{
   QRegExp  search(NAMEKEY_PREFIX_REGEXP ARG_REGEXP
                   NAME_PREFIX_REGEXP    ARG_REGEXP
                   PASSKEY_PREFIX_REGEXP ARG_REGEXP
                   PASS_PREFIX_REGEXP    ARG_REGEXP
                   PATH_PREFIX_REGEXP);
   QString  replace;
   replace.append(NAMEKEY_PREFIX).append(httpProtect(nameKey))
          .append(NAME_PREFIX)   .append(httpProtect(name))
          .append(PASSKEY_PREFIX).append(httpProtect(passwordKey))
          .append(PASS_PREFIX)   .append(httpProtect(password))
          .append(PATH_PREFIX);

   return str.replace(search, replace);
}

QString  Repo::Buffet::replaceAuth (QString         str,
                                    const QString&  name,
                                    const QString&  password)
{
   QString  nameKey;
   QString  nameDummy;
   QString  passwordKey;
   QString  passwordDummy;

   getAuth(str, nameKey, nameDummy, passwordKey, passwordDummy);

   return replaceAuth(str, nameKey, name, passwordKey, password);
}

QString  Repo::Buffet::replaceAuth (QString         str,
                                    const QString&  password)
{
   QString  nameKey;
   QString  name;
   QString  passwordKey;
   QString  passwordDummy;

   getAuth(str, nameKey, name, passwordKey, passwordDummy);

   return replaceAuth(str, nameKey, name, passwordKey, password);
}

QRegExp  Repo::Buffet::oneUrlRegExp ()
{
   return QRegExp("^.*"
                  NAMEKEY_PREFIX_REGEXP ARG_REGEXP
                  NAME_PREFIX_REGEXP    ARG_REGEXP
                  PASSKEY_PREFIX_REGEXP ARG_REGEXP
                  PASS_PREFIX_REGEXP    "(" ARG_REGEXP ")"
                  PATH_PREFIX_REGEXP    ".*$");
}

QRegExp  Repo::Buffet::urlsRegExp ()
{
   return QRegExp(NAMEKEY_PREFIX_REGEXP ARG_REGEXP
                  NAME_PREFIX_REGEXP    ARG_REGEXP
                  PASSKEY_PREFIX_REGEXP ARG_REGEXP
                  PASS_PREFIX_REGEXP    "(" ARG_REGEXP ")"
                  PATH_PREFIX_REGEXP    "[\\s]*\\b");
}
