#include "RepoEditorConfig.h"
#include "RepoBuffet.h"
#include <QtGui/QGridLayout>
#include <QtGui/QFont>
#include "Colors.h"
#include "Fonts.h"

#define URL_DIMS      100, 3
#define OPTIONS_DIMS   80, 3

RepoEditor::Config::Config(QGridLayout*  parentLayout)
   : programmaticChangeInProgress(false)
{
   QWidget*  parent = parentLayout->parentWidget();

   QLabel*  idLabel = new QLabel(tr("Identifier"), parent);
            idEdit  = new QLineEdit(parent);
   idEdit->setPalette(Colors::makeRequired(idEdit->palette()));
   idEdit->setToolTip(tr("Unique identifier for the repository"));
   connect(idEdit, SIGNAL(textChanged(const QString&)), 
           this,   SLOT(valueChanged(const QString&)));
   parentLayout->addWidget(idLabel, 0, 0);
   parentLayout->addWidget(idEdit,  0, 1);

   QLabel*  nameLabel = new QLabel(tr("Name"), parent);
            nameEdit  = new QLineEdit(parent);
   nameEdit->setToolTip(tr("Description of the repository"));
   connect(nameEdit, SIGNAL(textChanged(const QString&)), 
           this,     SLOT(valueChanged(const QString&)));
   parentLayout->addWidget(nameLabel, 1, 0);
   parentLayout->addWidget(nameEdit,  1, 1);

   QLabel*  urlsLabel = new QLabel(tr("URLs"), parent);
            urlsEdit  = new PasswdTextEdit(Repo::Buffet::urlsRegExp(), parent);
   urlsEdit->setToolTip(tr("List of URLs to the repository.\n"
                           "If more than one is specified, they all should be"
                           " identical mirrors,\n"
                           "and will provide redundancy if some URLs are"
                           " unreachable."));
   connect(urlsEdit, SIGNAL(textChanged()), 
           this,     SLOT(urlChanged()));
   parentLayout->addWidget(urlsLabel, 2, 0);
   parentLayout->addWidget(urlsEdit,  2, 1);

   QLabel*  mirrorListUrlLabel = new QLabel(tr("Mirror list URL"), parent);
            mirrorListUrlEdit  = new PasswdLineEdit
                                        (Repo::Buffet::oneUrlRegExp(), parent);
   mirrorListUrlEdit->setToolTip(tr("URL to a file containing a list of URLs"
                                    " to the repository."));
   connect(mirrorListUrlEdit, SIGNAL(textChanged(const QString&)), 
           this,              SLOT(urlChanged(const QString&)));
   parentLayout->addWidget(mirrorListUrlLabel, 3, 0);
   parentLayout->addWidget(mirrorListUrlEdit,  3, 1);

   QGroupBox*  authGroup = createAuth(parent);
   parentLayout->addWidget(authGroup, 4, 0, 1, 2);

   QGroupBox*  optionsGroup = createOptions(parent);
   parentLayout->addWidget(optionsGroup, 5, 0, 1, 2);

   QGroupBox*  fileInfoLayout = createFileInfo(parent);
   parentLayout->addWidget(fileInfoLayout, 6, 0, 1, 2);
}

QGroupBox*  RepoEditor::Config::createAuth(QWidget*  parent)
{
   QGroupBox*    group  = new QGroupBox(tr("Buffet Authentication"), parent);
   QGridLayout*  layout = new QGridLayout;

               authNameLabel = new QLabel(tr("Name"), group);
               authNameEdit  = new QLineEdit(group);
   authNameEdit->setToolTip(tr("Authentication name for a repository using the"
                               " buffet authentication method"));
   connect(authNameEdit, SIGNAL(textEdited(const QString&)), 
           this,         SLOT(authChanged(const QString&)));
   layout->addWidget(authNameLabel, 0, 0);
   layout->addWidget(authNameEdit,  0, 1);

               authPasswordLabel = new QLabel(tr("Password"), group);
               authPasswordEdit  = new QLineEdit(group);
   authPasswordEdit->setEchoMode(QLineEdit::Password);
   authPasswordEdit->setToolTip(tr("Authentication password for a repository"
                                   " using the buffet authentication method"));
   connect(authPasswordEdit, SIGNAL(textEdited(const QString&)), 
           this,             SLOT(authChanged(const QString&)));
   layout->addWidget(authPasswordLabel, 1, 0);
   layout->addWidget(authPasswordEdit,  1, 1);

   group->setLayout(layout);

   return group;
}

QGroupBox*  RepoEditor::Config::createOptions(QWidget*  parent)
{
   QGroupBox*    group  = new QGroupBox(tr("Repository Options"), parent);
   QVBoxLayout*  layout = new QVBoxLayout;

   QHBoxLayout*  spacedSimpleOptionsLayout = new QHBoxLayout();
   QLayout*      simpleOptionsLayout       = createSimpleOptions(group);
   spacedSimpleOptionsLayout->addLayout(simpleOptionsLayout);
   spacedSimpleOptionsLayout->addStretch();
   layout->addLayout(spacedSimpleOptionsLayout);

   QLabel*  otherOptionsLabel = new QLabel(tr("Other Options:"), group);
   layout->addWidget(otherOptionsLabel);

   otherOptionsEdit  = new HTextEdit(OPTIONS_DIMS, group);
   QFont  f = Fonts::makeBold(Fonts::makeFixed(otherOptionsEdit->font()));
   otherOptionsEdit->setFont(f);
   otherOptionsEdit->setToolTip(tr("Any other yum options for this repository."
                                   "  See yum.conf(5)"));
   connect(otherOptionsEdit, SIGNAL(textChanged()), 
           this,             SLOT(valueChanged()));
   layout->addWidget(otherOptionsEdit);

   group->setLayout(layout);

   return group;
}

QLayout*  RepoEditor::Config::createSimpleOptions(QWidget*  parent)
{
   QGridLayout*  layout = new QGridLayout;

   QLabel*  enabledLabel = new QLabel(tr("Enabled"), parent);
   layout->addWidget(enabledLabel, 0, 0);
   enabledComboBox = new QComboBox(parent);
   enabledComboBox->addItem(tr("No"),      Repo::False);
   enabledComboBox->addItem(tr("Default"), Repo::Unspecified);
   enabledComboBox->addItem(tr("Yes"),     Repo::True);
   enabledComboBox->setToolTip(tr("Whether or not the repository is enabled for"
                                  " use"));
   connect(enabledComboBox, SIGNAL(currentIndexChanged(int)), 
           this,            SLOT(valueChanged(int)));
   layout->addWidget(enabledComboBox, 0, 1);

   QLabel*  checkGPGKeysLabel = new QLabel(tr("Check GPG Keys"), parent);
   layout->addWidget(checkGPGKeysLabel, 1, 0);
   checkGPGKeysComboBox = new QComboBox(parent);
   checkGPGKeysComboBox->addItem(tr("No"),      Repo::False);
   checkGPGKeysComboBox->addItem(tr("Default"), Repo::Unspecified);
   checkGPGKeysComboBox->addItem(tr("Yes"),     Repo::True);
   checkGPGKeysComboBox->setToolTip(tr("Whether or not GPG key signatures of"
                                       " packages from the repository are"
                                       " checked to ensure authenticity"));
   connect(checkGPGKeysComboBox, SIGNAL(currentIndexChanged(int)), 
           this,                 SLOT(valueChanged(int)));
   layout->addWidget(checkGPGKeysComboBox, 1, 1);

   return layout;
}

QGroupBox*  RepoEditor::Config::createFileInfo(QWidget*  parent)
{
   QGroupBox*  group  = new QGroupBox(tr("Configuration File (expert)"));
   QLayout*    layout = new QHBoxLayout;

   QLabel*          filenameLabel       = new QLabel(tr("File"), parent);
                    filenameEdit        = new QLineEdit(parent);
   filenameEdit->setToolTip(tr("Name of the yum configuration file into which"
                               " to place this repository configuration.\n"
                               "If left blank, defaults to the repository's"
                               " identifier"));
   connect(filenameEdit, SIGNAL(textChanged(const QString&)), 
           this,         SLOT(valueChanged(const QString&)));
   QLabel*          sortPositionLabel   = new QLabel(tr("Sort Position"), 
                                                     parent);
                    sortPositionSpinBox = new QDoubleSpinBox(parent);
   sortPositionSpinBox->setToolTip(tr("A numeric sort key which determines the"
                                      " order of this repository"
                                      " configuration\n"
                                      "in a configuration file which contains"
                                      " multiple configurations"));
   connect(sortPositionSpinBox, SIGNAL(valueChanged(const QString&)), 
           this,                SLOT(valueChanged(const QString&)));
   sortPositionSpinBox->setDecimals(1);
   sortPositionSpinBox->setSingleStep(1.0);
   sortPositionSpinBox->setMinimum(0.0);
   sortPositionSpinBox->setValue(0.0);

   layout->addWidget(filenameLabel);
   layout->addWidget(filenameEdit);
   layout->addWidget(sortPositionLabel);
   layout->addWidget(sortPositionSpinBox);

   group->setLayout(layout);

   return group;
}

void RepoEditor::Config::setDefaults (bool  enabled,
                                      bool  gpgcheck)
{
   enabledComboBox->setItemText(Repo::Unspecified, 
                                tr("%1 %2").arg(tr("Default"))
                                           .arg(enabled ? tr("(Yes)") 
                                                        : tr("(No)")));
   checkGPGKeysComboBox->setItemText(Repo::Unspecified, 
                                     tr("%1 %2").arg(tr("Default"))
                                                .arg(gpgcheck ? tr("(Yes)") 
                                                              : tr("(No)")));
}

void RepoEditor::Config::disable()
{
   programmaticChangeInProgress = true;

   idEdit->setText("");
   nameEdit->setText("");
   urlsEdit->setRealPlainText("");
   mirrorListUrlEdit->setRealText("");
   authNameLabel->setText("Name");
   authNameEdit->setText("");
   authPasswordLabel->setText("Password");
   authPasswordEdit->setText("");
   enabledComboBox->setCurrentIndex(Repo::Unspecified);
   checkGPGKeysComboBox->setCurrentIndex(Repo::Unspecified);
   otherOptionsEdit->setPlainText("");
   filenameEdit->setText("");
   sortPositionSpinBox->setValue(0.0);

   programmaticChangeInProgress = false;

   idEdit->setEnabled(false);
   nameEdit->setEnabled(false);
   urlsEdit->setEnabled(false);
   mirrorListUrlEdit->setEnabled(false);
   authNameEdit->setEnabled(false);
   authPasswordEdit->setEnabled(false);
   enabledComboBox->setEnabled(false);
   checkGPGKeysComboBox->setEnabled(false);
   otherOptionsEdit->setEnabled(false);
   filenameEdit->setEnabled(false);
   sortPositionSpinBox->setEnabled(false);
}

void RepoEditor::Config::setEntry(const Repo::Entry&  entry)
{
   programmaticChangeInProgress = true;

   idEdit->setText(entry.id);
   nameEdit->setText(entry.name);
   urlsEdit->setRealPlainText(entry.urls);
   mirrorListUrlEdit->setRealText(entry.mirrorListUrl);
   
   QString  authNameKey;
   QString  authName;
   QString  authPasswordKey;
   QString  authPassword;
   bool     authHas = Repo::Buffet::getAuth(entry, 
                                            authNameKey, authName,
                                            authPasswordKey, authPassword);
   authNameLabel->setText(authNameKey.isEmpty() ? "Name"
                                                : authNameKey);
   authNameEdit->setText(authName);
   authPasswordLabel->setText(authPasswordKey.isEmpty() ? "Password"
                                                        : authPasswordKey);
   authPasswordEdit->setText(authPassword);

   enabledComboBox->setCurrentIndex(entry.enabled);
   checkGPGKeysComboBox->setCurrentIndex(entry.checkGPGKeys);
   otherOptionsEdit->setPlainText(entry.otherOptions);
   filenameEdit->setText(entry.filename);
   sortPositionSpinBox->setValue(entry.sortPosition);

   programmaticChangeInProgress = false;

   idEdit->setEnabled(true);
   nameEdit->setEnabled(true);
   urlsEdit->setEnabled(true);
   mirrorListUrlEdit->setEnabled(true);
   authNameEdit->setEnabled(authHas);
   authPasswordEdit->setEnabled(authHas);
   enabledComboBox->setEnabled(true);
   checkGPGKeysComboBox->setEnabled(true);
   otherOptionsEdit->setEnabled(true);
   filenameEdit->setEnabled(true);
   sortPositionSpinBox->setEnabled(true);
}

Repo::Entry RepoEditor::Config::toEntry()
{
   Repo::Entry  entry;

   entry.id            = idEdit->text();
   entry.name          = nameEdit->text();
   entry.urls          = urlsEdit->realPlainText();
   entry.mirrorListUrl = mirrorListUrlEdit->realText();
   entry.enabled       = (Repo::TriState)enabledComboBox->currentIndex();
   entry.checkGPGKeys  = (Repo::TriState)checkGPGKeysComboBox->currentIndex();
   entry.otherOptions  = otherOptionsEdit->toPlainText();
   entry.filename      = filenameEdit->text();
   entry.sortPosition  = sortPositionSpinBox->value();

   return entry;
}

bool RepoEditor::Config::setStatus(YumConfIO*  yumConfIO,
                                   QLabel*     label)
{
   if (idEdit->text().isEmpty()) {
      label->setPalette(Colors::makeError(label->palette()));
      label->setText(tr("No Identifier specified"));
      return true;
   }

   QString  urls          = urlsEdit->realPlainText();
   QString  mirrorListUrl = mirrorListUrlEdit->realText();

   if (urls.isEmpty() && mirrorListUrl.isEmpty()) {
      label->setPalette(Colors::makeWarning(label->palette()));
      label->setText(tr("No URLs or Mirror list URL specified"));
      return true;
   }

   QString  filename = filenameEdit->text();

   if (!filename.isEmpty()
       && filename != YUM_CONF
       && !filename.startsWith(yumConfIO->reposdir + '/')) {
      label->setPalette(Colors::makeWarning(label->palette()));
      label->setText(tr("File is not in %1").arg(yumConfIO->reposdir));
      return true;
   }
   if (!filename.isEmpty()
       && filename != YUM_CONF
       && !filename.endsWith(".repo")) {
      label->setPalette(Colors::makeWarning(label->palette()));
      label->setText(tr("File lacks .repo suffix"));
      return true;
   }

   QString  nameKey;
   QString  name;
   QString  passwordKey;
   QString  password;
   if (Repo::Buffet::getAuth(urls,
                             nameKey, name, passwordKey, password)
       || Repo::Buffet::getAuth(mirrorListUrl,
                                nameKey, name, passwordKey, password)) {
      if (name.isEmpty()) {
         label->setPalette
            (Colors::makeWarning(label->palette()));
         label->setText(tr("No Buffet Authentication Name specified"));
         return true;
      } else if (password.isEmpty()) {
         label->setPalette
            (Colors::makeWarning(label->palette()));
         label->setText(tr("No Buffet Authentication Password"
                                 " specified"));
         return true;
      }
   }

   return false;
}

void RepoEditor::Config::authChanged(const QString&  text)
{
   programmaticChangeInProgress = true;

   QString  name     = authNameEdit->text();
   QString  password = authPasswordEdit->text();

   urlsEdit->setRealPlainText(Repo::Buffet::replaceAuth
                                 (urlsEdit->realPlainText(),
                                  name, password));
   mirrorListUrlEdit->setRealText(Repo::Buffet::replaceAuth
                                     (mirrorListUrlEdit->realText(),
                                      name, password));

   programmaticChangeInProgress = false;

   emit changed();
}

void RepoEditor::Config::urlChanged()
{
   // The QTextEdit emits its textChanged() signal even if the change was
   // programmatic.  It doesn't count.  Don't try to handle the signal in
   // that case.  Especially since, when authChanged changes the UrlEdit and
   // triggers this, Qt sometimes seems to go nuts and change the focus from
   // whichever authentication widget was changed to the "Other Options"
   // widget!
   if (programmaticChangeInProgress) return;

   Repo::Entry  entry = toEntry();

   QString  authNameKey;
   QString  authName;
   QString  authPasswordKey;
   QString  authPassword;
   bool     authHas = Repo::Buffet::getAuth(entry, 
                                            authNameKey,     authName, 
                                            authPasswordKey, authPassword);
   authNameLabel->setText(authNameKey.isEmpty() ? "Name"
                                                : authNameKey);
   authNameEdit->setText(authName);
   authPasswordLabel->setText(authPasswordKey.isEmpty() ? "Password"
                                                        : authPasswordKey);
   authPasswordEdit->setText(authPassword);

   authNameEdit->setEnabled(authHas);
   authPasswordEdit->setEnabled(authHas);

   emit changed();
}

void RepoEditor::Config::urlChanged(const QString&  text)
{
   urlChanged();
}

void RepoEditor::Config::valueChanged()
{
   // The QTextEdit emits its textChanged() signal even if the change was
   // programmatic.  It doesn't count.
   if (programmaticChangeInProgress) return;

   emit changed();
}

void RepoEditor::Config::valueChanged(const QString&  text)
{
   // The QDoubleSpinBox emits its valueChanged() signal even if the change
   // was programmatic.  It doesn't count.
   if (programmaticChangeInProgress) return;

   emit changed();
}

void RepoEditor::Config::valueChanged(int  state)
{
   // The QComboBox emits its currentIndexChanged() signal even if the change
   // was programmatic.  It doesn't count.
   if (programmaticChangeInProgress) return;

   emit changed();
}
