#include "MainWindow.h"
#include "RepoEditorDialog.h"
#include "RepoEditor.h"
#include "RepoBuffet.h"
#include <QtGui/QVBoxLayout>
#include <QtGui/QPushButton>
#include <QtGui/QLabel>
#include <QtGui/QFrame>
#include <QtGui/QHeaderView>
#include "Colors.h"
#include "RepoBrowse.h"

RepoEditor::Dialog::Dialog(YumIO*         yumIO,
                           YumConfIO*     yumConfIO,
                           QWidget*       parent)
   : HDialog(parent),
     forceReposChanged(false),
     parent(parent), 
     yumIO(yumIO),
     yumConfIO(yumConfIO),
     originalYumReposUsed(yumConfIO->reposUsed)
{
   setWindowTitle(tr("NUU Repository Configuration Editor"));

   setAttribute(Qt::WA_ShowModal, true);

   QVBoxLayout*  mainLayout = new QVBoxLayout(this);

   model = new Model(yumIO, yumConfIO, this);
   view  = new View(model, this);
   model->setViewFont(view->font());
   mainLayout->addWidget(view);

   connect(model, SIGNAL(useChanged()),    this, SLOT(useChanged()));
   connect(model, SIGNAL(configChanged()), this, SLOT(configChanged()));
   connect(view, SIGNAL(currentChanged(int, int)), 
           this, SLOT(currentChanged(int, int)));

#ifndef REPO_EDITOR_ONE_DIALOG
   connect(view, SIGNAL(activated(const QModelIndex&)), 
           this, SLOT(edit(const QModelIndex&)));
#endif

   QHBoxLayout*  repoButtonLayout = new QHBoxLayout();
                 browseButton     = new QPushButton(tr("Browse Repository"), 
                                                    this);
#ifndef REPO_EDITOR_ONE_DIALOG
                 editButton       = new QPushButton(tr("Edit"),   this);
#endif
   QPushButton*  addButton        = new QPushButton(tr("Add"),    this);
                 copyButton       = new QPushButton(tr("Copy"),   this);
                 removeButton     = new QPushButton(tr("Remove"), this);
   addButton->setDefault(false);
   browseButton->setToolTip(tr("Browse the first URL specified for the selected"
                               " repository using an external web browser"));
#ifndef REPO_EDITOR_ONE_DIALOG
   editButton->setToolTip(tr("Edit the selected repository configuration"));
#endif
   addButton->setToolTip(tr("Create a new repository configuration from"
                            " scratch"));
   copyButton->setToolTip(tr("Create a new repository configuration based on"
                             " the selected repository"));
   removeButton->setToolTip(tr("Remove the selected repository configuration"));
   browseButton->setEnabled(false);
#ifndef REPO_EDITOR_ONE_DIALOG
   editButton->setEnabled(false);
#endif
   copyButton->setEnabled(false);
   removeButton->setEnabled(false);
   repoButtonLayout->addWidget(browseButton);
   repoButtonLayout->addStretch(1);
#ifndef REPO_EDITOR_ONE_DIALOG
   repoButtonLayout->addWidget(editButton);
#endif
   repoButtonLayout->addWidget(addButton);
   repoButtonLayout->addWidget(copyButton);
   repoButtonLayout->addWidget(removeButton);
   connect(browseButton, SIGNAL(clicked()), this, SLOT(browse()));
#ifndef REPO_EDITOR_ONE_DIALOG
   connect(editButton,   SIGNAL(clicked()), this, SLOT(edit()));
#endif
   connect(addButton,    SIGNAL(clicked()), this, SLOT(add()));
   connect(copyButton,   SIGNAL(clicked()), this, SLOT(copy()));
   connect(removeButton, SIGNAL(clicked()), this, SLOT(remove()));

#ifndef REPO_EDITOR_ONE_DIALOG
   connect(view->getContextMenu(), SIGNAL(edit(QModelIndex)),
           this,                   SLOT(edit(QModelIndex)));
#endif

   mainLayout->addLayout(repoButtonLayout);

   QFrame*  separator = new QFrame(this);
   separator->setFrameShape(QFrame::HLine);
   separator->setFrameShadow(QFrame::Sunken);
   mainLayout->addWidget(separator);

#ifdef REPO_EDITOR_ONE_DIALOG
   QLabel*  configLabel = new QLabel(tr("Repository Configuration:"), this);
   mainLayout->addWidget(configLabel);

   QGridLayout*  configLayout = new QGridLayout();
   config = new Config(configLayout);
   connect(config, SIGNAL(changed()), this, SLOT(entryChanged()));
   config->disable();
   mainLayout->addLayout(configLayout);
#else
   configDialog = new ConfigDialog(yumConfIO, this);
#endif

   QHBoxLayout*  dialogButtonLayout = new QHBoxLayout();
                 statusLabel        = new QLabel("", this);
#ifdef REPO_EDITOR_ONE_DIALOG
                 resetButton        = new QPushButton(tr("Reset"),     this);
#endif
                 resetAllButton     = new QPushButton(tr("Reset All"), this);
   QPushButton*  cancelButton       = new QPushButton(tr("Cancel"),    this);
                 applyButton        = new QPushButton(tr("Apply"),     this);
   QPushButton*  okButton           = new QPushButton(tr("OK"),        this);
   applyButton->setDefault(true);
   statusLabelDefaultPalette = statusLabel->palette();
#ifdef REPO_EDITOR_ONE_DIALOG
   resetButton->setToolTip(tr("Undo changes to the currently selected"
                              " repository configuration since it was selected"
                              " or the last Apply"));
#endif
   resetAllButton->setToolTip(tr("Undo changes to all repository configurations"
                                 " since the last Apply or opening of the"
                                 " dialog"));
   cancelButton->setToolTip(tr("Undo changes to all repository configurations"
                               " since the last Apply or opening of the dialog,"
                               " and close the window"));
   applyButton->setToolTip(tr("Apply changes to all repository"
                              " configurations, but do not close the window"));
   okButton->setToolTip(tr("Apply changes to all repository configurations, and"
                           " close the window"));
   dialogButtonLayout->addWidget(statusLabel);
   dialogButtonLayout->addStretch(1);
#ifdef REPO_EDITOR_ONE_DIALOG
   dialogButtonLayout->addWidget(resetButton);
#endif
   dialogButtonLayout->addWidget(resetAllButton);
   dialogButtonLayout->addWidget(cancelButton);
   dialogButtonLayout->addWidget(applyButton);
   dialogButtonLayout->addWidget(okButton);
#ifdef REPO_EDITOR_ONE_DIALOG
   connect(resetButton,    SIGNAL(clicked()), this, SLOT(reset()));
#endif
   connect(resetAllButton, SIGNAL(clicked()), this, SLOT(resetAll()));
   connect(cancelButton,   SIGNAL(clicked()), this, SLOT(cancel()));
   connect(applyButton,    SIGNAL(clicked()), this, SLOT(apply()));
   connect(okButton,       SIGNAL(clicked()), this, SLOT(ok()));

   mainLayout->addLayout(dialogButtonLayout);

   setLayout(mainLayout);   
}

void RepoEditor::Dialog::setVisible(bool visible)
{
   if (visible) {
      originalYumReposUsed = yumConfIO->reposUsed;

      bool status = yumConfIO->read(parent);
      model->reset();
      if (!status) return;

      // Kludge so that sortByColumn sets the order to ascending.
      view->horizontalHeader()->setSortIndicator(NAME, Qt::DescendingOrder);
      view->sortByColumn(ID);
      view->resizeColumnsToContents();

#ifdef REPO_EDITOR_ONE_DIALOG
      config->setDefaults(yumConfIO->enabled, 
                          yumConfIO->gpgcheck);
      config->disable();
#endif
      useHasChanged = false;
      setConfigHasChanged(false);
      setEntryHasChanged(false);

      applied = false;
   }

   QDialog::setVisible(visible);

   if (!visible) {
      if (applied || useHasChanged || forceReposChanged) {
         emit reposChanged();
      }
   }
}

void RepoEditor::Dialog::currentChanged(int  currentRow,
                                        int  previousRow)
{
#ifdef REPO_EDITOR_ONE_DIALOG
   if (previousRow != -1) {
      Repo::Entry&  modelEntry = yumConfIO->entries[previousRow];
      Repo::Entry   newEntry   = config->toEntry();
      if (modelEntry.enabled != newEntry.enabled) {
         yumConfIO->reposUsed[newEntry.id]
            = (newEntry.enabled == Repo::True  ? true  :
               newEntry.enabled == Repo::False ? false :
                                                 yumConfIO->enabled);
      }
      if (newEntry.id.isEmpty()) {
         newEntry.id = modelEntry.id;
      }
      modelEntry = newEntry;

      model->dataChanged(previousRow);

      setConfigHasChanged(configHasChanged || entryHasChanged);
      setEntryHasChanged(false);
   }
#endif

#ifndef REPO_EDITOR_ONE_DIALOG
   editButton->setEnabled(currentRow != -1);
#endif
   browseButton->setEnabled(currentRow != -1);
   copyButton->setEnabled(currentRow != -1);
   removeButton->setEnabled(currentRow != -1);

#ifdef REPO_EDITOR_ONE_DIALOG
   if (currentRow != -1) {
      config->setEntry(yumConfIO->entries[currentRow]);
   } else {
      config->disable();
   }
#endif

   updateStatusLabel();
}

void  RepoEditor::Dialog::useChanged()
{
   useHasChanged = true;
}

void  RepoEditor::Dialog::configChanged()
{
   setConfigHasChanged(true);
}

#ifdef REPO_EDITOR_ONE_DIALOG
void RepoEditor::Dialog::entryChanged()
{
   setEntryHasChanged(true);
}
#endif

void RepoEditor::Dialog::browse()
{
   int  currentRow = view->currentRow();
   RepoBrowse::browse(this, yumConfIO->entries[currentRow], yumIO);
}

#ifndef REPO_EDITOR_ONE_DIALOG
void RepoEditor::Dialog::edit()
{
   int  row = view->currentRow();
   configDialog->setEntry(&yumConfIO->entries[row]);
   if (configDialog->exec() == QDialog::Accepted) {
      model->dataChanged(row);
      setConfigHasChanged(true);
   }
}

void RepoEditor::Dialog::edit(const QModelIndex&  ix)
{
   int  row = ix.row();
   configDialog->setEntry(&yumConfIO->entries[row]);
   if (configDialog->exec() == QDialog::Accepted) {
      model->dataChanged(row);
      setConfigHasChanged(true);
   }
}
#endif

void RepoEditor::Dialog::add()
{
#ifdef REPO_EDITOR_ONE_DIALOG
   view->setCurrentIndex(QModelIndex());

   int  newRow = model->addRow(Repo::Entry());
   view->setCurrentRow(newRow);

   setConfigHasChanged(true);

   config->idEdit->setFocus(Qt::OtherFocusReason);
#else
   Repo::Entry  newEntry;
   configDialog->setEntry(&newEntry);
   if (configDialog->exec() == QDialog::Accepted) {
      view->setCurrentIndex(QModelIndex());
      int  newRow = model->addRow(newEntry);
      view->setCurrentRow(newRow);
      model->dataChanged(newRow);
      setConfigHasChanged(true);
   }
#endif
}

void RepoEditor::Dialog::copy()
{
#ifdef REPO_EDITOR_ONE_DIALOG
   int currentRow = view->currentRow();

   view->setCurrentIndex(QModelIndex());

   int  newRow = model->addRow(yumConfIO->entries[currentRow]);
   view->setCurrentRow(newRow);

   setConfigHasChanged(true);
#else
   int currentRow = view->currentRow();
   Repo::Entry  newEntry = yumConfIO->entries[currentRow];
   configDialog->setEntry(&newEntry);
   if (configDialog->exec() == QDialog::Accepted) {
      view->setCurrentIndex(QModelIndex());
      int  newRow = model->addRow(newEntry);
      view->setCurrentRow(newRow);
      model->dataChanged(newRow);
      setConfigHasChanged(true);
   }
#endif
}

void RepoEditor::Dialog::remove()
{
   view->setCurrentIndex(QModelIndex());

   QModelIndexList                  rows = view->selectedRows();
   QModelIndexList::const_iterator  it;
   for (it = rows.begin(); it != rows.end(); ++it) {
      model->removeRow(it->row());
   }

   setConfigHasChanged(true);
}

void RepoEditor::Dialog::reset()
{
#ifdef REPO_EDITOR_ONE_DIALOG
   config->setEntry(yumConfIO->entries[view->currentRow()]);
   setEntryHasChanged(false);
#endif
}

void RepoEditor::Dialog::resetAll()
{
   int      oldRow = view->currentRow();
   QString  oldId;

   if (oldRow != -1) {
      oldId = yumConfIO->entries[oldRow].id;
   }

   yumConfIO->reposUsed = originalYumReposUsed;

   yumConfIO->read(this);

   model->reset();

   if (oldRow != -1) {
      setEntryToOldId(oldId);
   } else {
#ifdef REPO_EDITOR_ONE_DIALOG
      config->disable();
#endif
   }

   useHasChanged = false;
   setConfigHasChanged(false);
   setEntryHasChanged(false);
}

void RepoEditor::Dialog::cancel()
{
   resetAll();
   reject();
}

bool RepoEditor::Dialog::apply()
{
   int      oldRow = view->currentRow();
   QString  oldId;

#ifdef REPO_EDITOR_ONE_DIALOG
   // currentRow can be -1 if called from ok()
   if (oldRow != -1) {
      Repo::Entry&  modelEntry = yumConfIO->entries[oldRow];
      Repo::Entry   newEntry   = config->toEntry();
      if (entryHasChanged) {
         yumConfIO->reposUsed[newEntry.id]
            = (newEntry.enabled == Repo::True  ? true  :
               newEntry.enabled == Repo::False ? false :
                                                 yumConfIO->enabled);
      }
      modelEntry = newEntry;

      model->dataChanged(oldRow);
      oldId = yumConfIO->entries[oldRow].id;
   }
#endif

   bool  status = yumConfIO->write(this);
   if (!status) return false;

   model->reset(); // The write may have done some cleanup

   if (oldRow != -1) {
      setEntryToOldId(oldId);
   } else {
#ifdef REPO_EDITOR_ONE_DIALOG
      config->disable();
#endif
   }

   setConfigHasChanged(false);
   setEntryHasChanged(false);

   applied = true;

   return true;
}

void RepoEditor::Dialog::ok()
{
   if (entryHasChanged || configHasChanged) {
      bool  status = apply();
      if (!status) return;
   }

   accept();
}

bool RepoEditor::Dialog::hasEmptyId()
{
   Repo::EntryList::const_iterator  it  = yumConfIO->entries.begin();
   int                              row = 0;
   for (; it != yumConfIO->entries.end(); ++it, ++row) {
      QString  id;
#ifdef REPO_EDITOR_ONE_DIALOG
      if (view->currentRow() == row) {
         id = config->toEntry().id;
      } else {
         id = it->id;
      }
#else
      id = it->id;
#endif
      if (id.isEmpty()) {
         return true;
      }
   }
   return false;
}

QString RepoEditor::Dialog::getDuplicateId()
{
   // There should be a reasonably small number of repositories, so the fact
   // that this is O(n^2) shouldn't be a big deal.
   Repo::EntryList::const_iterator  outer_it  = yumConfIO->entries.begin();
   int                              outer_row = 0;
   for (; outer_it != yumConfIO->entries.end(); ++outer_it, ++outer_row) {
      QString  outer_id;
#ifdef REPO_EDITOR_ONE_DIALOG
      if (view->currentRow() == outer_row) {
         outer_id = config->toEntry().id;
      } else {
         outer_id = outer_it->id;
      }
#else
      outer_id = outer_it->id;
#endif

      Repo::EntryList::const_iterator  inner_it  = outer_it  + 1;
      int                              inner_row = outer_row + 1;
      for (; inner_it != yumConfIO->entries.end(); ++inner_it, ++inner_row) {
         QString  inner_id;
#ifdef REPO_EDITOR_ONE_DIALOG
         if (view->currentRow() == inner_row) {
            inner_id = config->toEntry().id;
         } else {
            inner_id = inner_it->id;
         }
#else
         inner_id = inner_it->id;
#endif

         if (outer_id == inner_id) {
            return outer_it->id;
         }
      }
   }
   return QString();
}

void RepoEditor::Dialog::updateStatusLabel()
{
   // Display an errors or warnings about the currently selected repository
   // first.  This means we may see a warning about the currently selected
   // repository in preference to a general error related to the state of
   // all repositories.

#ifdef REPO_EDITOR_ONE_DIALOG
   if (view->currentRow() != -1) {
      if (config->setStatus(yumConfIO, statusLabel)) {
         return;
      }
   }
#endif

   if (hasEmptyId()) {
      statusLabel->setPalette(Colors::makeError(statusLabel->palette()));
      statusLabel->setText(tr("A repository has no Identifier"));
      return;
   }

   QString  duplicateId = getDuplicateId();
   if (!duplicateId.isEmpty()) {
      statusLabel->setPalette(Colors::makeError(statusLabel->palette()));
      statusLabel->setText(tr("Two or more repositories have the Identifier,"
                              " %1")
                              .arg(duplicateId));
      return;
   }

   if (configHasChanged || entryHasChanged) {
      statusLabel->setPalette(Colors::makeOK(statusLabel->palette()));
      statusLabel->setText(tr("Unapplied Configuration Changes"));
      return;
   }

   statusLabel->setPalette(statusLabelDefaultPalette);
   statusLabel->setText("");
}

void RepoEditor::Dialog::setEntryToOldId(QString  oldId)
{
   bool  found = false;

   Repo::EntryList::const_iterator  it  = yumConfIO->entries.begin();
   int                              row = 0;
   for (; it != yumConfIO->entries.end(); ++it, ++row) {
      if (it->id == oldId) {
         view->setCurrentRow(row);
         found = true;
      }
   }

#ifdef REPO_EDITOR_ONE_DIALOG
   if (!found) {
      config->disable();
   }
#endif
}

void RepoEditor::Dialog::setEntryHasChanged(bool value)
{
   entryHasChanged = value;
#ifdef REPO_EDITOR_ONE_DIALOG
   resetButton->setEnabled(entryHasChanged);
#endif
   resetAllButton->setEnabled(entryHasChanged || configHasChanged);
   applyButton->setEnabled(entryHasChanged || configHasChanged);

   updateStatusLabel();
}

void RepoEditor::Dialog::setConfigHasChanged(bool value)
{
   configHasChanged = value;
   resetAllButton->setEnabled(entryHasChanged || configHasChanged);
   applyButton->setEnabled(entryHasChanged || configHasChanged);

   updateStatusLabel();
}
