#include "RepoEditorModel.h"
#include <QtCore/QVariant>
#include "Share.h"
#include "RepoBrowse.h"

RepoEditor::Model::Model(YumIO*      yumIO,
                         YumConfIO*  yumConfIO,
                         QObject*    parent) 
   : QAbstractTableModel(parent),
     yumIO(yumIO),
     yumConfIO(yumConfIO),
     noneBoxIcon(Share::absolute("icons/noneBox.png")),
     checkBoxIcon(Share::absolute("icons/checkBox.png"))
{
}

RepoEditor::Model::~Model()
{
}

void RepoEditor::Model::setViewFont(const QFont&  font)
{
   italicFont = font;
   italicFont.setItalic(true);
}

int
RepoEditor::Model::rowCount(const QModelIndex&) const
{
   return yumConfIO->entries.count();
}

int
RepoEditor::Model::columnCount(const QModelIndex&) const
{
   return NUM_COLUMNS;
}

QVariant
RepoEditor::Model::data(const QModelIndex&  ix, 
                        int                 role) const
{
   if (!ix.isValid()) return QVariant();

   int  r = ix.row();
   int  c = ix.column();

   if (r >= yumConfIO->entries.count()) return QVariant();
   if (c >= NUM_COLUMNS)                return QVariant();

   switch (role) {
   case Qt::DisplayRole:
      switch (c) {
      case ID:   
         if (yumConfIO->entries[r].id.isEmpty()) {
            return "(new)";
         } else {
            return yumConfIO->entries[r].id;
         }
      case NAME: 
         return yumConfIO->entries[r].name;
      default:   
         return QVariant();
      }
   case Qt::DecorationRole:
      switch (c) {
      case USE:
         if (yumConfIO->reposUsed[yumConfIO->entries[r].id]) {
            return checkBoxIcon;
         } else {
            return noneBoxIcon;
         }
      case ENABLED:
         if (yumConfIO->entries[r].enabled == Repo::True
             || (yumConfIO->entries[r].enabled == Repo::Unspecified
                 && yumConfIO->enabled)) {
            return checkBoxIcon;
         } else {
            return noneBoxIcon;
         }
      default:
         return QVariant();
      }
   case Qt::FontRole:
      switch (c) {
      case ID: 
         if (yumConfIO->entries[r].id.isEmpty()) {
            return italicFont;
         } else {
            return QVariant();
         }
      default:
         return QVariant();
      }
   case Qt::ToolTipRole:
      switch (c) {
      case USE:
         if (yumConfIO->reposUsed[yumConfIO->entries[r].id]) {
            return tr("Repository is enabled for this session");
         } else {
            return tr("Repository is not enabled for this session");
         }
      case ENABLED:
         if (yumConfIO->reposUsed[yumConfIO->entries[r].id]) {
            return tr("Repository is configured to be enabled");
         } else {
            return tr("Repository is not configured to be enabled");
         }
      default:
         return QVariant();
      }
   case Qt::TextColorRole:
      return QVariant();
   default:
      return QVariant();
   }
}

QVariant  RepoEditor::Model::headerData  (int              section, 
                                          Qt::Orientation  orientation, 
                                          int              role) const
{
   if (orientation == Qt::Vertical) return QVariant();

   switch (role) {
   case Qt::DisplayRole:
      switch (section) {
      case USE:     return tr("Sess");
      case ENABLED: return tr("Perm");
      case ID:      return tr("Identifier");
      case NAME:    return tr("Name");
      default:      return QVariant();
      }
   case Qt::ToolTipRole:
      switch (section) {
      case USE:     return tr("Whether or not the repository is used during"
                              " this session.\n"
                              "It overrides the persistent Enabled flag in the"
                              " repository configuration.");
      case ENABLED: return tr("Whether or not the repository is configured to"
                              " be enabled by default.\n"
                              "It can be overridden for a particular session.");
      case ID:      return tr("Unique identifier for the repository");
      case NAME:    return tr("Description of the repository");
      default:      return QVariant();
      }
   default:
      return QVariant();
   }
}

int  RepoEditor::Model::addRow (const Repo::Entry&  entry)
{
   beginInsertRows(QModelIndex(), 
                   yumConfIO->entries.count(), yumConfIO->entries.count());
   yumConfIO->entries.append(entry);
   endInsertRows();

   return yumConfIO->entries.count() - 1;
}

bool  RepoEditor::Model::removeRow(int                 row,
                                   const QModelIndex&  parent)
{
   beginRemoveRows(QModelIndex(), row, row);
   yumConfIO->entries.removeAt(row);
   endRemoveRows();

   return true;
}

void  RepoEditor::Model::reset()
{
   QAbstractTableModel::reset();
}

QModelIndex  RepoEditor::Model::createIndex (int  row)
{
   return QAbstractTableModel::createIndex(row, 0);
}

void  RepoEditor::Model::toggleUse (const QModelIndex&  ix)
{
   if (!ix.isValid()) return;

   int  r = ix.row();

   if (r >= yumConfIO->entries.count()) return;

   bool&  used = yumConfIO->reposUsed[yumConfIO->entries[r].id];
   used = !used;

   QModelIndex  useIx = QAbstractTableModel::createIndex(r, USE);
   emit QAbstractTableModel::dataChanged(useIx, useIx);

   emit useChanged();
}

void  RepoEditor::Model::toggleEnabled (const QModelIndex&  ix)
{
   if (!ix.isValid()) return;

   int  r = ix.row();

   if (r >= yumConfIO->entries.count()) return;

   Repo::TriState&  enabled = yumConfIO->entries[r].enabled;
   enabled = ((enabled == Repo::False 
               || (enabled == Repo::Unspecified && !yumConfIO->enabled)) 
              ? Repo::True 
              : Repo::False);

   QModelIndex  useIx = QAbstractTableModel::createIndex(r, ENABLED);
   emit QAbstractTableModel::dataChanged(useIx, useIx);

   emit configChanged();
}

void  RepoEditor::Model::browse (const QModelIndex&  ix,
                                 QWidget*            parent)
{
   if (!ix.isValid()) return;

   int  r = ix.row();

   if (r >= yumIO->count()) return;

   RepoBrowse::browse(parent, yumConfIO->entries[r], yumIO);
}

void  RepoEditor::Model::dataChanged(int  row)
{
   QModelIndex  leftIx  = QAbstractTableModel::createIndex(row, 0);
   QModelIndex  rightIx = QAbstractTableModel::createIndex(row, NUM_COLUMNS-1);

   emit QAbstractTableModel::dataChanged(leftIx, rightIx);
}
