#include "YumConfIO.h"
#include <QtGui/QMessageBox>
#include <QtCore/QtAlgorithms>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "stringmacros.h"

#define YUM_CACHEDIR_DEFAULT  "/var/cache/yum"
#define YUM_REPOSDIR_DEFAULT  "/etc/yum.repos.d"

YumConfIO::YumConfIO(const YumReposEnabled&  reposEnabled,
                     YumReposUsed&           reposUsed)
   : reposEnabled(reposEnabled),
     reposUsed(reposUsed)
{
}

bool YumConfIO::read(QWidget*  errorParent)
{
   this->errorParent = errorParent;

   options.clear();
   cachedir = YUM_CACHEDIR_DEFAULT;
   reposdir = YUM_REPOSDIR_DEFAULT;
   enabled  = true;
   gpgcheck = false;
   entries.clear();

   FILE*  file = fopen(YUM_CONF, "r");
   if (file == NULL) {
      error(tr("Unable to read yum configuration file %1").arg(YUM_CONF));
      return false;
   }

   bool  status = readConfFile(file);
   if (!status) {
      fclose(file);
      return false;
   }

   fclose(file);

   DIR*  dir = opendir(reposdir.toStdString().c_str());
   if (dir == NULL) {
      error(tr("Unable to read repository configuration directory %1")
               .arg(reposdir));
      return false;
   }

   for (;;) {
      struct dirent*  dirent = readdir(dir);
      if (dirent == NULL) break;

      if (strcmp(dirent->d_name + strlen(dirent->d_name) - 5, ".repo") == 0) {
         QString  pathname
            = QString(reposdir).append("/").append(dirent->d_name);

         struct stat  st;
         int          status = stat(pathname.toStdString().c_str(), &st);
         if (status == -1) {
            error(tr("Unable to stat repository configuration file %1")
                     .arg(pathname.toStdString().c_str()));
            closedir(dir);
            return false;
         }

         if (!(st.st_mode & S_IFDIR)) {
            FILE*  file = fopen(pathname.toStdString().c_str(), "r");
            if (file == NULL) {
               error(tr("Unable to read repository configuration file %1")
                        .arg(pathname.toStdString().c_str()));
               closedir(dir);
               return false;
            }

            bool  status = readRepoFile(file, pathname);
            if (!status) {
               fclose(file);
               closedir(dir);
               return false;
            }

            fclose(file);
         }
      }
   }

   closedir(dir);

   emit changed();

   return true;
}

bool  YumConfIO::readConfFile(FILE*  file)
{
   for (;;) {
      QString  line;
      Status   status = readLogicalLine(file, YUM_CONF, line);
      if (status == ERROR)       return false;
      if (status == END_OF_FILE) break;

      QString  value;
      if (line == "[main]") {
         // DO NOTHIHNG
      } else if (line.size() > 0 && line[0] == '[') {
         unreadLine(line);
         break;
      } else if (isOption(line, "cachedir", value)) {
         cachedir = value;
      } else if (isOption(line, "reposdir", value)) {
         reposdir = value;
      } else if (isOption(line, "enabled", value)) {
         enabled = (value != "0");
      } else if (isOption(line, "gpgcheck", value)) {
         gpgcheck = (value != "0");
      }

      options.append(line).append('\n');
   }

   return readRepoFile(file, YUM_CONF);
}

bool  YumConfIO::readRepoFile(FILE*    file,
                              QString  pathname)
{
   bool   haveEntry = false;
   bool   newUsed   = false;
   Repo::Entry  entry;
   entry.filename     = pathname;
   entry.sortPosition = 0.0;

   for (;;) {
      QString  line;
      Status   status = readLogicalLine(file, pathname, line);
      if (status == ERROR)       return false;
      if (status == END_OF_FILE) break;

      QString  value;
      if (line.size() > 0 && line[0] == '[') {
         if (line.endsWith(']')) {
            if (haveEntry) {
               cleanupRead(entry);
               if (newUsed) {
                  applyUsedOptions(entry);
               }
               entries.append(entry);
            }

            double  sortPosition = entry.sortPosition;
            entry = Repo::Entry();
            entry.id = line.mid(1, line.count() - 2);
            newUsed = !reposUsed.contains(entry.id);
            if (newUsed) {
               reposUsed[entry.id] = enabled;
            }
            entry.filename     = pathname;
            entry.sortPosition = sortPosition + 1;

            haveEntry = true;
         } else {
            error(tr("Malformed section name in line: %1").arg(line));
            return false;
         }
      } else if (isOption(line, "name", value)) {
         entry.name = value;
      } else if (isOption(line, "baseurl", value)) {
         entry.urls = value;
      } else if (isOption(line, "mirrorlist", value)) {
         entry.mirrorListUrl = value;
      } else if (isOption(line, "enabled", value)) {
         entry.enabled = (value == "0" ? Repo::False : Repo::True);
         if (newUsed) {
            reposUsed[entry.id] = (entry.enabled == Repo::True);
         }
      } else if (isOption(line, "gpgcheck", value)) {
         entry.checkGPGKeys = (value == "0" ? Repo::False : Repo::True);
      } else if (!line.isEmpty()) {
         entry.otherOptions.append(line).append('\n');
      }
   }

   if (haveEntry) {
      cleanupRead(entry);
      if (newUsed) {
         applyUsedOptions(entry);
      }
      entries.append(entry);
   }

   return true;
}

YumConfIO::Status  YumConfIO::readLogicalLine (FILE*     file,
                                               QString   pathname,
                                               QString&  line)
{
   // readAhead contains logical lines and possibly as its last element the
   // first part of a logical line.  readAhead never contains an RFC 822
   // continuation line as a standalone entity.

   if (!readAhead.empty()) {
      line = readAhead.front();
      readAhead.pop_front();
      if (!readAhead.empty()) {
         // If there are more lines in the readAhead, then this must be a
         // whole logical line.  Just return it.
         return READ;
      }
   } else {
      Status  status = readLine(file, pathname, line);
      if (status != READ) return status;
   }

   // readAhead is empty

   for (;;) {
      QString  physicalLine;
      Status  status = readLine(file, pathname, physicalLine);
      if (status == ERROR) return ERROR;
      if (status == END_OF_FILE) {
         readAhead.clear();
         return READ;
      }
      if (physicalLine[0] == ' ' || physicalLine[0] == '\t') {
         // RFC 822 continuation line
         line.append("\n").append(physicalLine);
      } else {
         readAhead.append(physicalLine);
         return READ;
      }
   }

   // NOTREACHED
}

YumConfIO::Status  YumConfIO::readLine (FILE*     file,
                                        QString   pathname,
                                        QString&  line)
{
   line = "";

   for (;;) {
      errno = 0;
      int  ch = fgetc(file);
      if (ch == EOF) {
         if (errno != 0) {
            error(tr("Error reading repository configuration file %1")
                  .arg(pathname));
            return ERROR;
         } else {
            if (line != "") {
               return READ;
            } else {
               return END_OF_FILE;
            }
         }
      } else if (ch == '\n') {
         return READ;
      } else {
         line.append((char)ch);
      }
   }
}

void YumConfIO::unreadLine(QString  line)
{
   readAhead.push_front(line);
}

bool YumConfIO::isOption(QString   line,
                         QString   name,
                         QString&  value)
{
   int  delim = line.indexOf(QRegExp("[:=]"));
   if (delim == -1) return false;

   int  optionFin = delim - 1;
   while (line[optionFin] == ' ' || line[optionFin] == '\t') {
      optionFin--;
   }

   int  optionLen = optionFin + 1;
   if (optionLen != name.count() || line.mid(0, optionLen) != name) {
      return false;
   }

   unsigned int  valueIx = delim + 1;
   while (line[valueIx] == ' ' || line[valueIx] == '\t') {
      valueIx++;
   }

   value = line.mid(valueIx);

   return true;
}

void YumConfIO::cleanupRead(Repo::Entry&  entry)
{
   entry.urls.replace(QRegExp("[ \t\n]+"), "\n");
}

void YumConfIO::applyUsedOptions(Repo::Entry&  entry)
{
   // Apply any --enablerepo or --disablerepo options to override the default
   // enablement or the enablement as specified for the repo.  They are applied
   // as wildcards.  If two wildcards both apply, the one specified later takes
   // precedence.

   YumReposEnabled::const_iterator  it  = reposEnabled.begin();
   for (; it != reposEnabled.end(); ++it) {
      if (it->re.exactMatch(entry.id)) {
         reposUsed[entry.id] = it->enabled;
      }
   }
}

bool YumConfIO::write(QWidget*  errorParent)
{
   this->errorParent = errorParent;

   if (!checkErrors()) return false;
      
   cleanupWrite(entries);

   Repo::EntryList  e = entries;
   qSort(e.begin(), e.end(), filePositionLessThan);

   QSet<QString>  filenamesSeen;

   QString  filename;
   FILE*    file = NULL;

   Repo::EntryList::const_iterator  it = e.begin();
   for (; it != e.end(); ++it) {
      if (file != NULL && filename != it->filename) {
         fclose(file);
         file = NULL;
      }
      if (file == NULL) {
         filename = it->filename;
         filenamesSeen << it->filename;
         file = fopen(filename.toStdString().c_str(), "w");
         if (file == NULL) {
            error(tr("Unable to write repository configuration file %1")
                     .arg(filename.toStdString().c_str()));
            return false;
         }
         if (it->filename == YUM_CONF && !options.isEmpty()) {
            fprintf(file, "%s", options.toStdString().c_str());
         }
      } else {
         fprintf(file, "\n");
      }

      bool  status = writeRepo(file, *it);
      if (!status) return false;
   }

   if (file != NULL) {
      fclose(file);
   }

   if (!filenamesSeen.contains(YUM_CONF)) {
      file = fopen(YUM_CONF, "w");
      if (file == NULL) {
         error(tr("Unable to write repository configuration file %1")
                  .arg(YUM_CONF));
         return false;
      }
      if (!options.isEmpty()) {
         fprintf(file, "%s", options.toStdString().c_str());
      }
      fclose(file);
   }

   removeUnusedFiles(filenamesSeen);

   return true;
}

bool  YumConfIO::checkErrors()
{
   QString  msg;

   if (hasEmptyId()) {
      if (!msg.isEmpty()) {
         msg.append('\n');
      }
      msg.append(tr("A repository has no Identifier.  It will be removed."));
   }

   QString  duplicateId = getDuplicateId();
   if (!duplicateId.isEmpty()) {
      if (!msg.isEmpty()) {
         msg.append('\n');
      }
      msg.append(tr("Two or more repositories have the same Identifier(s), %1."
                    "  Only one repository will be applied for each Identifier.")
                    .arg(duplicateId));
   }

   if (!msg.isEmpty()) {
      int  result = QMessageBox::critical(errorParent,
                                          QObject::tr("Repository Editor"),
                                          msg.append("  Proceed?"),
                                          QMessageBox::Yes,
                                          (QMessageBox::No
                                           | QMessageBox::Default
                                           | QMessageBox::Escape));
      if (result != QMessageBox::Yes) return false;

      // Clean it up

      // There should be a reasonably small number of repositories, so the fact
      // that this is O(n^2) shouldn't be a big deal.
      Repo::EntryList::iterator  outer_it  = entries.begin();
      while (outer_it != entries.end()) {
         if (outer_it->id.isEmpty()) {
            outer_it = entries.erase(outer_it);
         } else {
            Repo::EntryList::iterator  inner_it  = outer_it + 1;
            while (inner_it != entries.end()) {
               if (outer_it->id == inner_it->id) {
                  inner_it = entries.erase(inner_it);
               } else {
                  ++inner_it;
               }
            }
            ++outer_it;
         }
      }
   }

   return true;
}

bool  YumConfIO::hasEmptyId()
{
   Repo::EntryList::const_iterator  it  = entries.begin();
   int                              row = 0;
   for (; it != entries.end(); ++it, ++row) {
      if (it->id.isEmpty()) {
         return true;
      }
   }
   return false;
}

QString  YumConfIO::getDuplicateId()
{
   QString  result;

   // There should be a reasonably small number of repositories, so the fact
   // that this is O(n^2) shouldn't be a big deal.
   Repo::EntryList::const_iterator  outer_it  = entries.begin();
   for (; outer_it != entries.end(); ++outer_it) {
      Repo::EntryList::const_iterator  inner_it  = outer_it  + 1;
      for (; inner_it != entries.end(); ++inner_it) {
         if (outer_it->id == inner_it->id) {
            if (!result.isEmpty()) {
               result.append(", ");
            }
            result.append(outer_it->id);
         }
      }
   }

   return result;
}

void  YumConfIO::cleanupWrite(Repo::EntryList&  e)
{
   Repo::EntryList::iterator  it = e.begin();
   for (; it != e.end(); ++it) {
      if (it->filename.isEmpty()) {
         it->filename = it->id;
         it->filename.append(".repo");
      }
      if (it->filename[0] != '/') {
         it->filename.prepend('/').prepend(reposdir);
      }
   }   
}

bool  YumConfIO::filePositionLessThan(const Repo::Entry&  l, 
                                      const Repo::Entry&  r)
{
   if (l.filename < r.filename) {
      return true;
   } else if (l.filename > r.filename) {
      return false;
   } else if (l.sortPosition < r.sortPosition) {
      return true;
   } else if (l.sortPosition > r.sortPosition) {
      return false;
   } else {
      return false;
   }
}

bool  YumConfIO::writeRepo(FILE*               file,
                           const Repo::Entry&  entry)
{
   fprintf(file, "[%s]\n", entry.id.toStdString().c_str());
   if (!entry.name.isEmpty()) {
      fprintf(file, "name=%s\n", entry.name.toStdString().c_str());
   }
   if (!entry.urls.isEmpty()) {
      QString  urls = entry.urls;
      urls.replace("\n", "\n        ");
      fprintf(file, "baseurl=%s\n", urls.toStdString().c_str());
   }
   if (!entry.mirrorListUrl.isEmpty()) {
      fprintf(file,
              "mirrorlist=%s\n", entry.mirrorListUrl.toStdString().c_str());
   }
   if (entry.enabled != Repo::Unspecified) {
      fprintf(file, 
              "enabled=%s\n", entry.enabled == Repo::True ? "1" : "0");
   }
   if (entry.checkGPGKeys != Repo::Unspecified) {
      fprintf(file,
              "gpgcheck=%s\n", entry.checkGPGKeys == Repo::True ? "1" : "0");
   }
   if (!entry.otherOptions.isEmpty()) {
      fprintf(file, "%s", entry.otherOptions.toStdString().c_str());
   }

   return true;
}

bool  YumConfIO::removeUnusedFiles (QSet<QString>&  filenamesSeen)
{
   DIR*  dir = opendir(reposdir.toStdString().c_str());
   if (dir == NULL) {
      error(tr("Unable to read repository configuration directory %1")
               .arg(reposdir));
      return false;
   }

   for (;;) {
      struct dirent*  dirent = readdir(dir);
      if (dirent == NULL) break;

      QString  pathname;
      pathname.append(reposdir).append("/").append(dirent->d_name);

      struct stat  st;
      int          status = stat(pathname.toStdString().c_str(), &st);
      if (status == -1) {
         error(tr("Unable to stat repository configuration file %1")
                  .arg(pathname.toStdString().c_str()));
         closedir(dir);
         return false;
      }

      if (!(st.st_mode & S_IFDIR)) {
         if (!filenamesSeen.contains(pathname)) {
            int  status = unlink(pathname.toStdString().c_str());
            if (status == -1) {
               error(tr("Unable to unlink repository configuration file %1")
                        .arg(pathname.toStdString().c_str()));
               closedir(dir);
               return false;
            }
         }
      }
   }

   closedir(dir);

   return true;
}

void  YumConfIO::error(QString  text)
{
   QMessageBox::critical(errorParent, 
                         QObject::tr("NUU Repository Editor Error"), 
                         text.append(tr("; errno = %1 (%2)."))
                             .arg(errno)
                             .arg(strerror(errno)));
}

#ifdef DEBUG_YUMCONFIO
void  YumConfIO::dump()
{
   fprintf(stderr, "----------------------------------------\n");

   Repo::EntryList  e = entries;
   
   Repo::EntryList::const_iterator  it = e.begin();
   for (; it != e.end(); ++it) {
      fprintf(stderr, "=== %s:\n", it->filename.toStdString().c_str());
      bool  status = writeRepo(stderr, *it);
      if (!status) fprintf(stderr, "ERROR\n");
      fprintf(stderr, "\n");
   }
}
#endif
