#include "YumIO.h"
#include "YumIOC.h"
#include <QtGui/QMessageBox>
#include <QtCore/QRegExp>
#include <QtCore/QEventLoop>
#include <QtCore/QFile>
#include "ScrollableErrorDialog.h"
#include <QtCore/qglobal.h>
#include <QtCore/QThread>
#include <QtCore/QMutex>
#include "Share.h"
#include <rpm/rpmlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include "stringmacros.h"
#include "globals.h"

#undef DEBUG_GRAB_METER
#undef DEBUG_GRAB_METER_DETAIL
#undef DEBUG_TS
#undef DEBUG_TS_DETAIL
#undef DEBUG_EXPECTED_EXCEPTIONS
#undef DEBUG_DISABLE_CAPTURERS

#if PY_VERSION_HEX < 0x02050000
typedef int  Py_ssize_t;
#endif

//
// Values passed to output.YumCliRPMCallback by rpm in yum >= 3.2.3
//
#define TS_UPDATE       10
#define TS_INSTALL      20
#define TS_TRUEINSTALL  30
#define TS_ERASE        40
#define TS_OBSOLETED    50
#define TS_OBSOLETING   60
#define TS_AVAILABLE    70
#define TS_UPDATED      90

#define IS_INSTALL(action) \
   ((action) == TS_INSTALL || (action) == TS_TRUEINSTALL)
#define IS_UPDATE(action) \
   ((action) == TS_UPDATE || (action) == TS_OBSOLETING)
#define IS_REMOVE(action) \
   ((action) == TS_ERASE)
#define IS_COMPLETING_UPDATE(action) \
   ((action) == TS_UPDATED || (action) == TS_OBSOLETED)

//
// Known yum strings for collect's progress indications
//
#define COLL_GRAB_METADATA  "MD Read"

//
// Known yum strings for apply's depsolve
//
// I hate having to know these strings, but there is no callback for a
// package added to the transaction by depSolve.  (There's one for the
// Requires: string that caused it, but that's not good enough, because
// the Requires: might not mention the package name but just something
// else Provided by the package.)
//
// The actual strings look like:
//    TSINFO: Updating %s to resolve dep.
//    TSINFO: Updating %s to resolve conflict.
//    TSINFO: Obsoleting %s with %s to resolve dep.
//    TSINFO: Marking %s as update for %s
//    TSINFO: Marking %s as install for %s
//    TSINFO: %s package requiring %s marked as erase
//
#define APPLY_DEPSOLVE_PREFIX           "TSINFO: "
#define APPLY_DEPSOLVE_UPDATING         "Updating "
#define APPLY_DEPSOLVE_OBSOLETING       "Obsoleting "
#define APPLY_DEPSOLVE_OBSOLETING_WITH  " with "
#define APPLY_DEPSOLVE_MARKING          "Marking "
#define APPLY_DEPSOLVE_MARKING_UPDATE   " update "
#define APPLY_DEPSOLVE_ERASE            " marked as erase"

//
// Known yum strings for available packages
//
#define YUM_DECODE_PRCO_FLAGS(f)  (strcmp((f), "GT") == 0 ? ">"  : \
                                   strcmp((f), "GE") == 0 ? ">=" : \
                                   strcmp((f), "LT") == 0 ? "<"  : \
                                   strcmp((f), "LE") == 0 ? "<=" : \
                                   strcmp((f), "EQ") == 0 ? "="  : \
                                                            (f))

// When using package sizes to estimate the amount of time for a package,
// very small packages still will have overhead time.  So, use a lower bound
// of PKG_MIN as the "size" of those packages when figuring the total
// progress.  (Continue to use the actual size for item progress, though.)
#define PKG_MIN  10240ULL

//
// Progress values for collect
//
#define PG_C_INITIALIZE     100
#define PG_C_REPOS_START    900
#define PG_C_REPOS        14000
#define PG_C_UPDATES       4000
#define PG_C_INSTALL       4000
#define PG_C_GROUPS        1000

#define PG_C_TOTAL_TEXT        "Initializing: Total Progress"
#define PG_C_INITIALIZE_TEXT   "Preparing"
#define PG_C_REPOS_START_TEXT  "Reading repositories"
//      PG_C_REPOS_TEXT        "Downloading/Reading repository metadata for ..."
#define PG_C_UPDATES_TEXT      "Collecting updatable packages"
#define PG_C_INSTALL_TEXT      "Collecting installed and available packages"
#define PG_C_GROUPS_TEXT       "Collecting groups"

#define PG_C_TOTAL  (  PG_C_INITIALIZE  \
                     + PG_C_REPOS_START \
                     + PG_C_REPOS       \
                     + PG_C_UPDATES     \
                     + PG_C_INSTALL     \
                     + PG_C_GROUPS)

#define PG_APR    2  // Acts per repository: grab & read

//
// Progress values for collectFileLists
//
#define PG_F_REPOS_START   1000
#define PG_F_REPOS        14000

#define PG_F_TOTAL_TEXT        "Loading file lists: Total Progress"
#define PG_F_REPOS_START_TEXT  "Reading repositories"
//      PG_F_REPOS_TEXT        "Downloading/Reading repository metadata for ..."

#define PG_F_TOTAL  (  PG_F_REPOS_START \
                     + PG_F_REPOS)

//
// Progress values for collectOtherData
//
#define PG_O_REPOS_START   1000
#define PG_O_REPOS        14000

#define PG_O_TOTAL_TEXT        "Loading miscellaneous data: Total Progress"
#define PG_O_REPOS_START_TEXT  "Reading repositories"
//      PG_O_REPOS_TEXT        "Downloading/Reading repository metadata for ..."

#define PG_O_TOTAL  (  PG_O_REPOS_START \
                     + PG_O_REPOS)

//
// Progress values for apply's Check Dependencies (i.e. before confirmation)
//
#define PG_D_INITIALIZE    100
#define PG_D_DEPSOLVE     9900

#define PG_D_TOTAL_TEXT        "Checking Dependencies: Total Progress"
#define PG_D_INITIALIZE_TEXT   "Preparing"
//      PG_D_DEPSOLVE_TEXT     "Downloading ... header"
//                             "Resolve Dependencies"


#define PG_D_TOTAL        (  PG_D_INITIALIZE  \
                           + PG_D_DEPSOLVE)

//
// Progress values for apply's Run Transaction (i.e. after confirmation)
//
#define PG_R_INITIALIZE           100
#define PG_R_GRAB_RPMS_NOMINAL  24000 // used to define ts.run.pgRGrabRpms
#define PG_R_TRANS_TEST_INIT     1000
#define PG_R_TRANS_TEST          1000
#define PG_R_RUN_TRANS_INIT      1000
#define PG_R_RUN_TRANS           8000

#define PG_R_TOTAL_TEXT            "Running Transaction: Total Progress"
#define PG_R_INITIALIZE_TEXT       "Preparing"
//      PG_R_GRAB_RPMS             "Downloading ... header"
#define PG_R_TRANS_TEST_INIT_TEXT  "Preparing to Test Transaction"
//      PG_R_TRANS_TEST            "Transaction Test"
#define PG_R_RUN_TRANS_INIT_TEXT   "Preparing to Run Transaction"
//      PG_R_RUN_TRANS             "Transaction Test"
//                                 "Installing: ..."
//                                 "Updating: ..."
//                                 "Removing: ..."
//                                 "Completing update for: ..."

#define PG_R_TOTAL        (  PG_R_INITIALIZE      \
                           + ts.run.pgRGrabRpms   \
                           + PG_R_TRANS_TEST_INIT \
                           + PG_R_TRANS_TEST      \
                           + PG_R_RUN_TRANS_INIT  \
                           + PG_R_RUN_TRANS)

// The number of acts given the number of grabs
#define PG_D_GRAB(grabs)  ((grabs) * 10)

// Estimate that a dependency resolution is 10% of the total time for
// grabbing the headers
#define PG_D_DEPSOLVE_ESTIMATE(pkgs)  ((pkgs) * 10 / 10)

// The number of acts for each download based on the size of the package
#define PG_R_GRAB(size)   (qMax((size), PKG_MIN))

// The number of acts for each operation based on the size of the package
//
// The size of an installed package is reported in uncompressed bytes.
// But the size of an uninstalled package is reported as its file size.
// Convert uninstalled package sizes by multiplying by 3 (a heuristic value).
//
// The COMPLETE_UPDATE is the erase part of the update.  It usually is pretty 
// fast because most of the time it doesn't have to delete much (or anything).
// So divide its size by 5 (a heuristic value).
#define PG_R_INSTALL(size)          (qMax((size) * 3, PKG_MIN))
#define PG_R_REMOVE(size)           (qMax((size),     PKG_MIN))
#define PG_R_UPDATE(size)           (qMax((size) * 3, PKG_MIN))
#define PG_R_COMPLETE_UPDATE(size)  (qMax((size) / 5, PKG_MIN))

// Estimate that a transaction test is 5% of the total time for the main
// part of the transaction
#define PG_R_TEST_TRANS_ESTIMATE(acts)  ((acts) / 20)

class YumIO::EntryKey {
public:
   EntryKey(bool     isGroup,
            QString  name)
      : isGroup(isGroup), 
        name(name)
   {}
   EntryKey(Package::Kind  kind,
            QString        name)
      : isGroup(kind == Package::GROUP), 
        name(name)
   {}
   inline bool operator==(EntryKey&  that) const
   {
      return this->isGroup == that.isGroup && this->name == that.name;
   }
   bool     isGroup;
   QString  name;
};

inline uint qHash(const YumIO::EntryKey&  key)
{
   extern uint qHash(const QString &key);
   return (qHash(key.name) << 1) + key.isGroup;
}

YumIO::YumIO()
   : logStartOp(-1),
     collectedFileLists(false),
     collectedOtherData(false),
     dialogs(NULL),
     capturers(NULL),
     rpm_dict(NULL),
     classYum(NULL),
     yumIO(NULL),
     reposDict(NULL),
     locked(false),
     yumVerKnown(false),
     totalProgressBar(NULL),
     itemProgressBar(NULL),
     pg(NULL)
{
   wr.stderrKind  = Log::ERROR;
   wr.stderrValue = 0;
}

YumIO::~YumIO()
{
   Python::PyGILStateWrapper  gsw;

   if (initialized) {
      // Since the log window is toast now, and we're destroying the yumIO
      // now, which will render the YumIOStdout and YumIOStderr python
      // objects useless, let's change the python stdout & stderr to their
      // defaults.  That way, if any errors happen, they'll be seen and won't
      // cause either cascading hangs or SIGSEGV's.
      py("sys.stdout = sys.__stdout__\n"
         "sys.stderr = sys.__stderr__\n");
   }

   if (locked) {
      PyObject*  xcptLockError = pyLookup("yum.Errors.LockError");
      pyExceptExpect(xcptLockError);
      PyObject*  unlockStatus = pyMethodCall(yumIO, "unlock", "()");
      if (pyExceptMatches(unlockStatus, xcptLockError)) {
         pythonException(tr("Unable to release yum lock"));
      }
      pyExceptClear();
   }
}

void YumIO::setup(QWidget*       parent,
                  HProgressBar*  totalProgressBar,
                  HProgressBar*  itemProgressBar)
{
   dialogs = new YumIODialogs(parent);
   this->parent           = parent;
   this->totalProgressBar = totalProgressBar;
   this->itemProgressBar  = itemProgressBar;
}

bool YumIO::initialize()
{
   Python::initialize();
   if (yumIO == NULL) {
      YumIOC::initialize(this);

      // This is necessary for newer versions of yum (e.g. 2.4.2) when
      // calling getOptionsConfig in python.  And it's harmless for older
      // versions of yum (e.g. 2.2.2).
      pyAssign("sys", "argv", Py_BuildValue("[s,s]", program, "info"));

      PyObject*  rpm_mod  = PyImport_ImportModule("rpm");
      if (rpm_mod == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: cannot find python rpm module"));
         return false;
      }

      rpm_dict = PyModule_GetDict(rpm_mod);
      if (rpm_dict == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: cannot find python rpm dict"));
         return false;
      }

      setEnablePythonSIGINTHandler(false);

      FILE*  yumIOFile
         = fopen(Share::absolute("python/YumIO.py").toStdString().c_str(), "r");
      if (yumIOFile == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: unable to open YumIO.py"));
         return false;
      }
      int  status = PyRun_SimpleFile(yumIOFile, "YumIO.py");
      if (status == -1) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: unable to execute YumIO.py"));
         return false;
      }
      fclose(yumIOFile);

      classYum = PyDict_GetItemString(main_dict, "YumIO");
      if (classYum == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: cannot find python YumIO"));
         return false;
      }
      PyObject*  args = Py_BuildValue("()");
      yumIO = PyObject_Call(classYum, args, NULL);
      Py_DECREF(args);
      if (yumIO == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: cannot create python yumIO"));
         return false;
      }
   }

   return true;
}

void  YumIO::execWorkThread (QEventLoop&  eventLoop,
                             QThread&     thread)
{
   blocker.prepare();

   thread.start(QThread::HighPriority);

   sigset_t     blocked;
   sigset_t     original;
   sigemptyset(&blocked);
   sigaddset(&blocked, SIGCHLD);
   sigprocmask(SIG_BLOCK, &blocked, &original);

   blocker.release();

   eventLoop.exec();
   thread.wait();
   eventLoop.processEvents(QEventLoop::ExcludeUserInputEvents);

   // Undo the obnoxious SIGINT handler that either yum or rpm installed again
   {
      Python::PyGILStateWrapper  gsw;
      setEnablePythonSIGINTHandler(false);
   }

   sigprocmask(SIG_SETMASK, &original, NULL);
}

class CollectThread : public QThread 
{
public:
   CollectThread(YumIO*               yumIO,
                 QEventLoop&          eventLoop,
                 const YumReposUsed&  reposUsed)
      : yumIO(yumIO),
        eventLoop(eventLoop),
        reposUsed(reposUsed)
   {}
protected:
   YumIO*               yumIO;
   QEventLoop&          eventLoop;
   const YumReposUsed&  reposUsed;
public:
   bool                 result;
protected:
   void run()
   {
      result = yumIO->collectReal(reposUsed);
      eventLoop.exit();
   }
};

bool YumIO::collect (const YumReposUsed&  reposUsed)
{
   QEventLoop     eventLoop;
   CollectThread  thread(this, eventLoop, reposUsed);
   execWorkThread(eventLoop, thread);
   return thread.result;
}

bool YumIO::collectReal (const YumReposUsed&  reposUsed)
{
   ProgressUser      pgUser(pg, totalProgressBar, itemProgressBar);
   YumIODialogsUser  yumIODialogsUser(dialogs, parent);

   log(Log::ACTION, 0, tr("Initialization"));

   Python::PyGILStateWrapper  gsw;

   collectedFileLists = false;
   collectedOtherData = false;

   pg->startTotal(tr(PG_C_TOTAL_TEXT), PG_C_TOTAL);

   pg->startItem(tr(PG_C_INITIALIZE_TEXT), PG_C_INITIALIZE);
   if (!locked) {
      PyObject*  xcptLockError = pyLookup("yum.Errors.LockError");
      pyExceptExpect(xcptLockError);
      PyObject*  lockStatus = pyMethodCall(yumIO, "lock", "()");
      if (pyExceptMatches(lockStatus, xcptLockError)) {
         pythonException(tr("Unable to obtain yum lock"));
         return false;
      }
      pyExceptClear();

      locked = true;
   }

   PyObject*  repos     = PyObject_GetAttrString(yumIO, "repos");
              reposDict = PyObject_GetAttrString(repos, "repos");
   if (reposDict == NULL) {
      error(tr("Internal error: yumIO.repos.repos not evaluable"));
      return false;
   }
   if (!PyDict_Check(reposDict)) {
      error(tr("Internal error: yumIO.repos.repos not a dictionary"));
      return false;
   }
   
   useRepos(repos, reposUsed);

   repoProgress.repos = collectRepos(reposDict);
   repoProgress.kind  = tr("metadata");
   repoProgress.act   = 0;

   if (PyDict_Size(reposDict) == 0) {
      error(tr("No repositories are configured.\n"
               "Under the Repositories menu, Edit Configuration can be used"
               " to configure repositories."));
      return false;
   } else if (repoProgress.repos.count() == 0) {
      error(tr("No repositories are enabled.\n"
               "Under the Repositories menu, Edit Configuration can be used"
               " to configure repositories."));
      return false;
   }

   pg->startItem(tr(PG_C_REPOS_START_TEXT), PG_C_REPOS_START);

   pg->setupItemActs(PG_C_REPOS, repoProgress.repos.count() * PG_APR);
   PyObject*  xcptRepoError = pyLookup("yum.Errors.RepoError");
   pyExceptExpect(xcptRepoError);
   PyObject*  reposStatus = pyMethodCall(yumIO, "setupRepos", "()");
   if (pyExceptMatches(reposStatus, xcptRepoError)) {
      pythonExceptionV(tr("Unable to initialize"));
      return false;
   }
   pyExceptClear();
   pg->endItemActs();

   pg->startItem(tr(PG_C_UPDATES_TEXT), PG_C_UPDATES);
   PyObject*  xcptYumBaseError = pyLookup("yum.Errors.YumBaseError");
   pyExceptExpect(xcptYumBaseError);
   PyObject*  setupTsStatus = pyMethodCall(yumIO, "setupTs", "()");
   if (pyExceptMatches(setupTsStatus, xcptYumBaseError)) {
      pythonExceptionV(tr("Unable to setup transaction"));
      return false;
   }
   pyExceptClear();
   pyExceptExpect(xcptYumBaseError);
   PyObject*  updates = pyMethodCall(yumIO, "collectUpdates", "()");
   if (pyExceptMatches(updates, xcptYumBaseError)) {
      pythonExceptionV(tr("Unable to collect updatable packages"));
      return false;
   } else if (updates == NULL) {
      error(tr("Internal error: execution of yumIO.collectUpdates failed"));
      return false;
   }
   pyExceptClear();

   if (!collectSet(hash, updates, Package::OUT_OF_DATE)) {
      return false;
   }

   pg->startItem(tr(PG_C_INSTALL_TEXT), PG_C_INSTALL);
   pyExceptExpect(xcptYumBaseError);
   PyObject*  info = pyMethodCall(yumIO, "collectAll", "()");
   if (pyExceptMatches(info, xcptYumBaseError)) {
      pythonExceptionV(tr("Unable to collect installed and available"
                          " packages"));
      return false;
   } else if (info == NULL) {
      error(tr("Internal error: execution of yumIO.collectAll failed"));
      return false;
   }
   pyExceptClear();

   if (   !collectField(hash, info, "installed", Package::INSTALLED)
       || !collectField(hash, info, "available", Package::NOT_INSTALLED)) {
      return false;
   }

   pg->startItem(tr(PG_C_GROUPS_TEXT), PG_C_GROUPS);
   PyObject*  xcptGroupsError = pyLookup("yum.Errors.GroupsError");
   pyExceptExpect(xcptGroupsError);
   pyExceptExpect(xcptYumBaseError);
   PyObject*  setupTsGroupResult = pyMethodCall(yumIO, "setupTsGroup", "()");
   if (pyExceptMatches(setupTsGroupResult, xcptGroupsError)) {
      QString  value = pyString(pyExceptValuef());
      if (value.contains("No Groups Available")) {
         // DO NOTHING: It's fairly normal.  (Although why it's lumped with 
         //             actually serious errors as the single 
         //             yum.Errors.GroupsErrors exception I have no idea!)
      } else {
         error(tr("Internal error:"
                  " execution of yumIO.setupTsGroup failed"));
         return false;
      }
   } else if (pyExceptMatches(setupTsGroupResult, xcptYumBaseError)) {
      pythonExceptionV(tr("Unable to collect groups"));
      return false;
   }
   pyExceptClear();
   PyObject*  groupInfo = pyMethodCall(yumIO, "collectGroupInfo", "()");
   if (groupInfo == NULL) {
      error(tr("Internal error: execution of yumIO.collectGroupInfo failed"));
      return false;
   }

   if (!collectGroups(hash, groupInfo)) {
      return false;
   }

   // This is OK because collectGroup() had to INCREF it once for each group
   // anyway, so that the number of references matches number of
   // Package::Entry's that actually reference it.  That way, when
   // Package::Entry::~Entry decrements the counts, they aren't
   // over-decremented.
   Py_DECREF(groupInfo);

   EntryHash::const_iterator  it = hash.begin();
   for (; it != hash.end(); ++it) {
      list.append(*it);
   }

   return true;
}

void  YumIO::useRepos(PyObject*            repos,
                      const YumReposUsed&  reposUsed)
{
   YumReposUsed::const_iterator  it = reposUsed.begin();
   for (; it != reposUsed.end(); ++it) {
      PyObject*  repo = pyDictGetItem(reposDict, it.key());
      if (repo != NULL) {
         char*  id = strdup(it.key().toStdString().c_str());
         if (*it) {
            pyMethodCall(repos, "enableRepo", "(s)", id);
         } else {
            pyMethodCall(repos, "disableRepo", "(s)", id);
         }
         free(id);
      }
   }
}

QStringList  YumIO::collectRepos (PyObject*  dict)
{
   QStringList  result;

   Py_ssize_t  pos = 0;
   for (;;) {
      PyObject*  key;
      PyObject*  value;
      if (!PyDict_Next(dict, &pos, &key, &value)) break;

      PyObject*  enabled = PyObject_GetAttrString(value, "enabled");
      if (enabled != NULL && PyInt_AsLong(enabled) != 0) {
         result.append(pyString(key));
      }
   }

   return result;
}

bool  YumIO::collectField (EntryHash&      hash,
                           PyObject*       info,
                           const char*     field,
                           Package::State  state)
{
   char*  copy = new char[strlen(field)+1];
   strcpy(copy, field);
   bool  result = collectField(hash, info, copy, state);
   delete [] copy;
   return result;
}

bool  YumIO::collectField (EntryHash&      hash,
                           PyObject*       info,
                           char*           field,
                           Package::State  state)
{
   PyObject*  set = PyObject_GetAttrString(info, field);
   if (set == NULL) {
      error(tr("Internal error: yumIO.collectAll.%1 not found").arg(field));
      return false;
   }

   return collectSet(hash, set, state);
}

bool  YumIO::collectSet (EntryHash&      hash,
                         PyObject*       set,
                         Package::State  state)
{
   PyObject*  it = PyObject_GetIter(set);
   for (;;) {
      PyObject*  pkg = PyIter_Next(it);
      if (pkg == NULL) break;

      EntryKey  key(false, pyStringDec(PyObject_GetAttrString(pkg, "name")));
      if (!hash.contains(key)) {
         bool  status = collectPkg(hash[key], state, pkg);
         if (!status) return false;
      } else {
         Package::Entry*  entry = hash[key];
         if (entry->state == Package::OUT_OF_DATE
             && state == Package::INSTALLED) {
            // For an out-of-date package, in addition to the version
            // available for installation, keep track of the installed
            // version.  The Describe dialog displays both.
            while (entry->next != NULL) {
               entry = entry->next;
            }
            collectPkg(entry->next, state, pkg);
         }
      }

      // Don't Py_DECREF(pkg) because pkg is stashed in entry.obj
   }

   Py_DECREF(it);

   return true;
}

bool  YumIO::collectPkg (Package::Entry*&  entry,
                         Package::State    state,
                         PyObject*         pkg)
{
   entry = new Package::Entry;
   
   entry->action        = Package::NONE;
   entry->state         = state;
   entry->kind          = getPkgKind(pkg);
   entry->name          = pyStringDec(PyObject_GetAttrString(pkg, "name"));
   entry->verrel        = pyStringDec(pyMethodCall(pkg, "printVer", "()"));
   entry->repo          = simpleString(pkg, "repoid");
   entry->arch          = pyStringDec(PyObject_GetAttrString(pkg, "arch"));
   entry->size          = sizeOf(pkg);
   entry->sizeText      = bytesFormat(entry->size);
   entry->summary       = simpleString(pkg, "summary");
   entry->buildTime     = buildTimeOf(pkg);
   entry->buildTimeText = timeFormat(entry->buildTime);
   entry->obj           = pkg;

   return true;
}

bool  YumIO::collectGroups (EntryHash&  hash,
                            PyObject*   groupInfo)
{
   PyObject*  list;

   if (hasDistribGroups()) {
      list = groupInfo;
   } else {
      list = PyObject_GetAttrString(groupInfo, "visible_groups");
      if (list == NULL) {
         error(tr("Internal error: groupInfo.visible_groups not evaluable"));
         return false;
      }
   }

   PyObject*  it = PyObject_GetIter(list);
   for (;;) {
      PyObject*  obj = PyIter_Next(it);
      if (obj == NULL) break;

      PyObject*  nameObj;
      PyObject*  groupObj;
      if (hasDistribGroups()) {
         PyObject*   user_visible = PyObject_GetAttrString(obj, "user_visible");
         if (user_visible == NULL || !PyObject_IsTrue(user_visible)) continue;
         nameObj = PyObject_GetAttrString(obj, "name");
         if (nameObj == NULL) {
            error(tr("Internal error: group without name"));
            return false;
         }
         groupObj = obj;
      } else {
         nameObj  = obj;
         groupObj = groupInfo;
      }

      QString   name = pyString(nameObj);
      EntryKey  key(true, name);

      if (!hash.contains(key)) {
         bool  status = collectGroup(hash[key], groupObj, name);
         if (!status) return false;
      }

      Py_DECREF(nameObj);
   }

   Py_DECREF(it);

   return true;
}

bool  YumIO::collectGroup (Package::Entry*&  entry,
                           PyObject*         groupObj,
                           QString           name)
{
   PyObject*  installedObj;
   if (hasDistribGroups()) {
      // groupObj is a yum.comps.Group
      installedObj = PyObject_GetAttrString(groupObj, "installed");
      if (installedObj == NULL) {
         error(tr("Internal error: cannot find group's installed flag"));
         return false;
      }
   } else {
      // groupObj is a groupInfo
      PyObject*  groupInstalled = PyObject_GetAttrString(groupObj,
                                                         "group_installed");
      char*  nameCopy = strdup(name.toStdString().c_str());
      installedObj = PyDict_GetItemString(groupInstalled, nameCopy);
      free(nameCopy);
      if (installedObj == NULL) {
         error(tr("Internal error:"
                  " cannot find groupInfo.group_installed['%s']").arg(name));
         return false;
      }
   }
   bool  installed = PyObject_IsTrue(installedObj);

   Py_INCREF(groupObj);

   entry = new Package::Entry;
   
   entry->action = Package::NONE;
   entry->state  = (installed ? Package::INSTALLED 
                              : Package::NOT_INSTALLED);
   entry->kind   = Package::GROUP;
   entry->name   = name;
   entry->obj    = groupObj;

   if (entry->state == Package::INSTALLED && !upToDateGroup(*entry)) {
      entry->state = Package::OUT_OF_DATE;
   }

   return true;
}

void  YumIO::clear()
{
   Python::PyGILStateWrapper  gsw;

   pyMethodCall(yumIO, "flush", "()");

   qDeleteAll(list.begin(), list.end());
   list.clear();
   hash.clear();

   Py_DECREF(yumIO);
   PyObject*  args = Py_BuildValue("()");
   yumIO = PyObject_Call(classYum, args, NULL);
   Py_DECREF(args);
}

QString  YumIO::getState (const Package::Entry&  entry)
{
   switch (entry.state) {
   case Package::OUT_OF_DATE:   return tr("Available for update");
   case Package::NOT_INSTALLED: return tr("Available for installation");
   case Package::INSTALLED:     return tr("Installed");
   default:                     return QString();
   }
}

QString  YumIO::getKind (const Package::Entry&  entry)
{
   switch (entry.kind) {
   case Package::SUITE:   return tr("Suite of packages");
   case Package::GROUP:   return tr("Group of packages");
   case Package::PACKAGE: return tr("Ordinary package");
   default:               return QString();
   }
}

QString  YumIO::getDescription (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "description");
}

QString  YumIO::getGroup (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "group");
}

QString  YumIO::getVendor (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "vendor");
}

QString  YumIO::getBuildHost (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "buildhost");
}

QString  YumIO::getInstallTime (const Package::Entry&  entry)
{
   return getRpmTime(entry.obj, "INSTALLTIME");
}

QString  YumIO::getBinaryRpm (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "relativepath");
}

QString  YumIO::getSourceRpm (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "sourcerpm");
}

QString  YumIO::getLicense (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "license");
}

QString  YumIO::getPackager (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "packager");
}

QString  YumIO::getUrl (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "url");
}

QString  YumIO::getRelocations (const Package::Entry&  entry)
{
   return getRpmList(entry.obj, "PREFIXES");
}

QString  YumIO::getProvides (const Package::Entry&  entry)
{
   return getPrco(entry.obj, "PROVIDE", "provides");
}

QString  YumIO::getRequires (const Package::Entry&  entry)
{
   return getPrco(entry.obj, "REQUIRE", "requires");
}

QString  YumIO::getConflicts (const Package::Entry&  entry)
{
   return getPrco(entry.obj, "CONFLICT", "conflicts");
}

QString  YumIO::getObsoletes (const Package::Entry&  entry)
{
   return getPrco(entry.obj, "OBSOLETE", "obsoletes");
}

QString  YumIO::getFiles (const Package::Entry&  entry)
{
   Python::PyGILStateWrapper  gsw;

   QStringList  values;

   switch (entry.kind) {
   case Package::SUITE:
   case Package::PACKAGE:
      if (PyObject_HasAttrString(entry.obj, "hdr")) {
         // Must be a YumInstalledPackage.  Get info from the hdr

         PyObject*  baseNamesList  = getRpmHdrValue(entry.obj, "BASENAMES");
         if (baseNamesList == NULL || baseNamesList == Py_None) {
            return QString();
         }
         PyObject*  dirNamesList   = getRpmHdrValue(entry.obj, "DIRNAMES");
         PyObject*  dirIndexesList = getRpmHdrValue(entry.obj, "DIRINDEXES");

         PyObject*  baseNamesIt  = PyObject_GetIter(baseNamesList);
         PyObject*  dirIndexesIt = (PyList_Check(dirIndexesList)
                                    ? PyObject_GetIter(dirIndexesList)
                                    : NULL);
         for (;;) {
            PyObject*  baseNameItem = PyIter_Next(baseNamesIt);
            if (baseNameItem == NULL) break;
            PyObject*  dirIndexItem = (dirIndexesIt != NULL
                                       ? PyIter_Next(dirIndexesIt)
                                       : dirIndexesList);

            QString        baseName = pyString(baseNameItem);
            unsigned long  dirIndex = PyInt_AsLong(dirIndexItem);

            PyObject*  dirNameItem = (dirNamesList != NULL
                                      ? PyList_GetItem(dirNamesList, dirIndex)
                                      : NULL);
            QString    dirName     = pyString(dirNameItem);

            values.append(QString(dirName).append(baseName));

            Py_DECREF(dirIndexItem);
            Py_DECREF(baseNameItem);
         }      

         Py_XDECREF(dirIndexesIt);
         Py_DECREF(baseNamesIt);

         Py_DECREF(dirIndexesList);
         Py_DECREF(dirNamesList);
         Py_DECREF(baseNamesList);
      } else if (PyObject_HasAttrString(entry.obj, "files")) {
         // Must be a YumAvailablePackage.  First make sure we've collected
         // the File Lists, and then get the files from returnFileEntries()
         gsw.release();
         collectFileLists();
         gsw.ensure();

         PyObject*  list = pyMethodCall(entry.obj, "returnFileEntries", 
                                        "(s)", "file");
         if (list == NULL) {
            error(tr("Internal error: cannot find package's files"));
            return QString();
         }

         PyObject*  it = PyObject_GetIter(list);
         for (;;) {
            PyObject*  item = PyIter_Next(it);
            if (item == NULL) break;

            values.append(pyString(item));

            Py_DECREF(item);
         }

         Py_DECREF(it);
      }
      break;
   case Package::GROUP:
      // DO NOTHING
      break;
   }

   return makeSortedUniq(values);
}

QString  YumIO::getChangeLog (const Package::Entry&  entry)
{
   Python::PyGILStateWrapper  gsw;

   QString  result;

   switch (entry.kind) {
   case Package::SUITE:
   case Package::PACKAGE:
      if (PyObject_HasAttrString(entry.obj, "hdr")) {
         // Must be a YumInstalledPackage.  Get info from the hdr

         PyObject*  namesList = getRpmHdrValue(entry.obj, "CHANGELOGNAME");
         if (namesList == NULL || namesList == Py_None) return QString();
         PyObject*  timesList = getRpmHdrValue(entry.obj, "CHANGELOGTIME");
         PyObject*  textsList = getRpmHdrValue(entry.obj, "CHANGELOGTEXT");

         PyObject*  namesIt = PyObject_GetIter(namesList);
         PyObject*  timesIt = (PyList_Check(timesList)
                               ? PyObject_GetIter(timesList)
                               : NULL);
         PyObject*  textsIt = PyObject_GetIter(textsList);
         for (;;) {
            PyObject*  nameItem = PyIter_Next(namesIt);
            if (nameItem == NULL) break;
            PyObject*  timeItem = (timesIt != NULL
                                   ? PyIter_Next(timesIt)
                                   : timesList);
            PyObject*  textItem = PyIter_Next(textsIt);

            QString  name = pyString(nameItem);
            QString  time = dateFormat(pyTime(timeItem));
            QString  text = pyString(textItem);

            if (!result.isEmpty()) {
               result.append("\n\n");
            }

            result.append("* ").append(time)
                  .append(' ').append(name)
                  .append('\n');
            result.append(text);

            Py_DECREF(textItem);
            Py_DECREF(timeItem);
            Py_DECREF(nameItem);
         }      

         Py_DECREF(textsIt);
         Py_XDECREF(timesIt);
         Py_DECREF(namesIt);

         Py_DECREF(textsList);
         Py_DECREF(timesList);
         Py_DECREF(namesList);
      } else if (PyObject_HasAttrString(entry.obj, "changelog")) {
         // Must be a YumAvailablePackage.  First make sure we've collected
         // the File Lists, and then get the files from returnFileEntries()
         gsw.release();
         collectOtherData();
         gsw.ensure();

         PyObject*  list;
         if (yumVersion() >= "2.3.1") {
            list = pyMethodCall(entry.obj, "returnChangelog", "()");
         } else {
            list = PyObject_GetAttrString(entry.obj, "changelog");
         }
         if (list == NULL) {
            error(tr("Internal error: cannot find package's changelog"));
            return QString();
         }
         if (list == Py_None) {
            return QString();
         }

         PyObject*  it = PyObject_GetIter(list);
         for (;;) {
            PyObject*  item = PyIter_Next(it);
            if (item == NULL) break;

            QString  time = dateFormat(pyTime(PyTuple_GetItem(item, 0)));
            QString  name = pyString(PyTuple_GetItem(item, 1));
            QString  text = pyString(PyTuple_GetItem(item, 2));

            if (!result.isEmpty()) {
               result.append("\n\n");
            }

            result.append("* ").append(time)
                  .append(' ').append(name)
                  .append('\n');
            result.append(text);

            Py_DECREF(item);
         }

         Py_DECREF(it);
      }
      break;
   case Package::GROUP:
      // DO NOTHING
      break;
   }

   return result;
}

QString  YumIO::getRequiredGroups (const Package::Entry&  entry)
{
   return getGroupGrps(entry, "sub_groups");
}

QString  YumIO::getDefaultMetaPkgs (const Package::Entry&  entry)
{
   return getGroupGrps(entry, "default_metapkgs");
}

QString  YumIO::getOptionalMetaPkgs (const Package::Entry&  entry)
{
   return getGroupGrps(entry, "optional_metapkgs");
}

QString  YumIO::getMandatoryPkgs (const Package::Entry&  entry)
{
   return getGroupPkgs(entry, "mandatory_pkgs", "mandatory_packages");
}

QString  YumIO::getDefaultPkgs (const Package::Entry&  entry)
{
   return getGroupPkgs(entry, "default_pkgs", "default_packages");
}

QString  YumIO::getOptionalPkgs (const Package::Entry&  entry)
{
   return getGroupPkgs(entry, "optional_pkgs", "optional_packages");
}

QString  YumIO::getConditionalPkgs (const Package::Entry&  entry)
{
   if (hasDistribGroups()) {
      return getGroupPkgs(entry, "", "conditional_packages");
   } else {
      return QString();
   }
}

QString  YumIO::getYumVar (QString  name)
{
   PyObject*  conf   = PyObject_GetAttrString(yumIO, "conf");
   PyObject*  yumvar = PyObject_GetAttrString(conf, "yumvar");

   char*  nameStr = new char[name.length()+1];
   strcpy(nameStr, name.toStdString().c_str());

   PyObject*  resultObj = PyMapping_GetItemString(yumvar, nameStr);

   delete [] nameStr;

   return pyStringDec(resultObj);
}

class CollectFileListsThread : public QThread 
{
public:
   CollectFileListsThread(YumIO*       yumIO,
                          QEventLoop&  eventLoop)
      : yumIO(yumIO),
        eventLoop(eventLoop)
   {}
protected:
   YumIO*       yumIO;
   QEventLoop&  eventLoop;
public:
   bool         result;
protected:
   void run()
   {
      result = yumIO->collectFileListsReal();
      eventLoop.exit();
   }
};

bool  YumIO::collectFileLists ()
{
   if (collectedFileLists) return true;

   QEventLoop               eventLoop;
   CollectFileListsThread   thread(this, eventLoop);
   execWorkThread(eventLoop, thread);
   return thread.result;
}

bool  YumIO::collectFileListsReal ()
{
   ProgressUser      pgUser(pg, totalProgressBar, itemProgressBar);
   YumIODialogsUser  yumIODialogsUser(dialogs, parent);

   log(Log::ACTION, 0, tr("Collecting File Lists"));

   repoProgress.kind = tr("file lists");
   repoProgress.act  = 0;

   Python::PyGILStateWrapper  gsw;

   pg->startTotal(tr(PG_F_TOTAL_TEXT), PG_F_TOTAL);

   pg->startItem(tr(PG_F_REPOS_START_TEXT), PG_F_REPOS_START);

   pg->setupItemActs(PG_F_REPOS, repoProgress.repos.count() * PG_APR);
   PyObject*  xcptYumBaseError = pyLookup("yum.Errors.YumBaseError");
   pyExceptExpect(xcptYumBaseError);
   PyObject*  info = pyMethodCall(yumIO, "collectFileLists", "()");
   if (pyExceptMatches(info, xcptYumBaseError)) {
      pythonExceptionV(tr("Unable to collect file lists"));
      return false;
   } else if (info == NULL) {
      error(tr("Internal error: execution of yumIO.collectFileLists failed"));
      return false;
   }
   pyExceptClear();
   pg->endItemActs();

   collectedFileLists = true;

   emit collectedExtraData();

   return true;
}

class CollectOtherDataThread : public QThread 
{
public:
   CollectOtherDataThread(YumIO*       yumIO,
                          QEventLoop&  eventLoop)
      : yumIO(yumIO),
        eventLoop(eventLoop)
   {}
protected:
   YumIO*       yumIO;
   QEventLoop&  eventLoop;
public:
   bool         result;
protected:
   void run()
   {
      result = yumIO->collectOtherDataReal();
      eventLoop.exit();
   }
};

bool  YumIO::collectOtherData ()
{
   if (collectedOtherData) return true;

   QEventLoop               eventLoop;
   CollectOtherDataThread   thread(this, eventLoop);
   execWorkThread(eventLoop, thread);
   return thread.result;
}

bool  YumIO::collectOtherDataReal ()
{
   ProgressUser      pgUser(pg, totalProgressBar, itemProgressBar);
   YumIODialogsUser  yumIODialogsUser(dialogs, parent);

   log(Log::ACTION, 0, tr("Collecting Miscellaneous data"));

   repoProgress.kind = tr("miscellaneous data");
   repoProgress.act  = 0;

   Python::PyGILStateWrapper  gsw;

   pg->startTotal(tr(PG_O_TOTAL_TEXT), PG_O_TOTAL);

   pg->startItem(tr(PG_O_REPOS_START_TEXT), PG_O_REPOS_START);

   pg->setupItemActs(PG_O_REPOS, repoProgress.repos.count() * PG_APR);
   PyObject*  xcptYumBaseError = pyLookup("yum.Errors.YumBaseError");
   pyExceptExpect(xcptYumBaseError);
   PyObject*  info = pyMethodCall(yumIO, "collectOtherData", "()");
   if (pyExceptMatches(info, xcptYumBaseError)) {
      pythonExceptionV(tr("Unable to collect miscellaneous data"));
      return false;
   } else if (info == NULL) {
      error(tr("Internal error: execution of yumIO.collectOtherData failed"));
      return false;
   }
   pyExceptClear();
   pg->endItemActs();

   collectedOtherData = true;

   emit collectedExtraData();

   return true;
}

bool  YumIO::upToDateGroup (const Package::Entry&  entry)
{
   return (upToDateGroupGrps(entry, "sub_groups")
           && upToDateGroupPkgs(entry, "mandatory_pkgs", "mandatory_packages"));
}

bool  YumIO::upToDateGroupGrps (const Package::Entry&  entry,
                                const char*            attr_22)
{
   char*  copy = new char[strlen(attr_22)+1];
   strcpy(copy, attr_22);
   bool  result = upToDateGroupGrps(entry, copy);
   delete [] copy;
   return result;
}

bool  YumIO::upToDateGroupGrps (const Package::Entry&  entry,
                                char*                  attr_22)
{
   if (hasDistribGroups()) {
      // DO NOTHING
   } else {
      PyObject*  seq = getGroupItem_22(entry, attr_22);
      if (seq == NULL) {
         return true;
      }
      if (PySequence_Length(seq) > 0) {
         QStringList  values;

         PyObject*  it = PyObject_GetIter(seq);
         for (;;) {
            PyObject*  grpId = PyIter_Next(it);
            if (grpId == NULL) break;
            EntryKey  key(true, pyString(grpId));
            if (hash.contains(key)) {
               if (!upToDateGroup(*hash[key])) {
                  return false;
               }
            }
            Py_DECREF(grpId);
         }

         Py_DECREF(it);
      }
   }

   return true;
}

bool  YumIO::upToDateGroupPkgs (const Package::Entry&  entry,
                                const char*            attr_22,
                                const char*            attr_25)
{
   char*  attr_22_copy = new char[strlen(attr_22)+1];
   char*  attr_25_copy = new char[strlen(attr_25)+1];
   strcpy(attr_22_copy, attr_22);
   strcpy(attr_25_copy, attr_25);
   bool  result = upToDateGroupPkgs(entry, attr_22_copy, attr_25_copy);
   delete [] attr_25_copy;
   delete [] attr_22_copy;
   return result;
}

bool  YumIO::upToDateGroupPkgs (const Package::Entry&  entry,
                                char*                  attr_22,
                                char*                  attr_25)
{
   if (hasDistribGroups()) {
      PyObject*  dict = getGroupItem_25(entry, attr_25);
      if (dict == NULL) {
         return true;
      }

      if (PyDict_Size(dict) > 0) {
         Py_ssize_t  pos = 0;
         for (;;) {
            PyObject*  pkgName;
            PyObject*  value;
            if (!PyDict_Next(dict, &pos, &pkgName, &value)) break;

            EntryKey  key(false, pyString(pkgName));
            if (hash.contains(key)) {
               if (hash[key]->state == Package::OUT_OF_DATE) {
                  return false;
               }
            }
         }
      }
   } else {
      PyObject*  seq = getGroupItem_22(entry, attr_22);
      if (seq == NULL) {
         return true;
      }

      if (PySequence_Length(seq) > 0) {
         PyObject*  it = PyObject_GetIter(seq);
         for (;;) {
            PyObject*  pkgName = PyIter_Next(it);
            if (pkgName == NULL) break;
            EntryKey  key(false, pyString(pkgName));
            if (hash.contains(key)) {
               if (hash[key]->state == Package::OUT_OF_DATE) {
                  return false;
               }
            }
            Py_DECREF(pkgName);
         }

         Py_DECREF(it);
      }
   }

   return true;
}

QString  YumIO::getPrco (PyObject*    pkg,
                         const char*  hdrKey,
                         const char*  prcoKey)
{
   char*  procKeyCopy = new char[strlen(prcoKey)+1];
   strcpy(procKeyCopy, prcoKey);
   QString  result = getPrco(pkg, hdrKey, procKeyCopy);
   delete [] procKeyCopy;
   return result;
}

QString  YumIO::getPrco (PyObject*    pkg,
                         const char*  hdrKey,
                         char*        prcoKey)
{
   Python::PyGILStateWrapper  gsw;

   QStringList  values;

   if (PyObject_HasAttrString(pkg, "hdr")) {
      // Must be a YumInstalledPackage.  Get the prco from the hdr
      char  nameKey[1024];
      char  flagsKey[1024];
      char  versionKey[1024];
      sprintf(nameKey,    "%sNAME",    hdrKey);
      sprintf(flagsKey,   "%sFLAGS",   hdrKey);
      sprintf(versionKey, "%sVERSION", hdrKey);

      PyObject*  nameList    = getRpmHdrValue(pkg, nameKey);
      if (nameList == NULL || nameList == Py_None) return QString();
      PyObject*  flagsList   = getRpmHdrValue(pkg, flagsKey);
      PyObject*  versionList = getRpmHdrValue(pkg, versionKey);

      PyObject*  nameIt    = PyObject_GetIter(nameList);
      PyObject*  flagsIt   = (flagsList != NULL ? PyObject_GetIter(flagsList)
                                                : NULL);
      PyObject*  versionIt = (flagsList != NULL ? PyObject_GetIter(versionList)
                                                : NULL);
      for (;;) {
         PyObject*  nameItem    = PyIter_Next(nameIt);
         if (nameItem == NULL) break;
         PyObject*  flagsItem   = (flagsIt != NULL   ? PyIter_Next(flagsIt)
                                                     : NULL);
         PyObject*  versionItem = (versionIt != NULL ? PyIter_Next(versionIt)
                                                     : NULL);

         QString  value = pyString(nameItem);

         if (   flagsItem   != NULL && flagsItem   != Py_None
             && versionItem != NULL && versionItem != Py_None) {
            unsigned long  flags = PyInt_AsLong(flagsItem);
            QString        flagsStr;
            if (flags & RPMSENSE_LESS)    flagsStr.append('<');
            if (flags & RPMSENSE_GREATER) flagsStr.append('>');
            if (flags & RPMSENSE_EQUAL)   flagsStr.append('=');
            value.append(' ').append(flagsStr);
            value.append(' ').append(pyString(versionItem));
         }

         values.append(value);

         Py_XDECREF(versionItem);
         Py_XDECREF(flagsItem);
         Py_DECREF(nameItem);
      }      

      Py_XDECREF(versionIt);
      Py_XDECREF(flagsIt);
      Py_DECREF(nameIt);

      Py_DECREF(versionList);
      Py_DECREF(flagsList);
      Py_DECREF(nameList);
   } else if (PyObject_HasAttrString(pkg, "prco")) {
      // Must be a YumAvailablePackage.  
      //
      // If the package has a returnPrco function, as is the case with newer
      // versions of yum (e.g. 2.4.2), then call it.  This is necessary even
      // if they have a prco attribute, because it is filled in lazily.
      //
      // However, for older versions of yum (e.g. 2.2.2), there is no
      // returnPrco function, so use the prco attribute directly.  Note that
      // returnSimple won't return the prco attribute.
      PyObject*  list;
      if (PyObject_HasAttrString(pkg, "returnPrco")) {
         list = pyMethodCall(pkg, "returnPrco", "(s)", prcoKey);
      } else {
         PyObject*  prco = PyObject_GetAttrString(pkg, "prco");
         list = PyDict_GetItemString(prco, prcoKey);
      }
      if (list == NULL) {
         error(tr("Internal error:"
                  " cannot find package's prco.%1").arg(prcoKey));
         return QString();
      }

      PyObject*  it = PyObject_GetIter(list);
      for (;;) {
         PyObject*  item = PyIter_Next(it);
         if (item == NULL) break;

         PyObject*  name  = PyTuple_GetItem(item, 0);
         PyObject*  flags = PyTuple_GetItem(item, 1);
         PyObject*  evr   = PyTuple_GetItem(item, 2);

         QString  value = pyString(name);
         
         if (   flags != NULL && flags != Py_None
             && evr   != NULL && evr   != Py_None) {
            const char*  flagsCode = Python::pyString(flags);
            const char*  flagsStr  = YUM_DECODE_PRCO_FLAGS(flagsCode);
            PyObject*    epoch     = PyTuple_GetItem(evr, 0);
            PyObject*    ver       = PyTuple_GetItem(evr, 1);
            PyObject*    rel       = PyTuple_GetItem(evr, 2);
            if (ver != NULL && ver != Py_None) {
               value.append(' ').append(flagsStr).append(' ');
               if (epoch != NULL && epoch != Py_None) {
                  char*  epochStr = Python::pyString(epoch);
                  if (strcmp(epochStr, "0") != 0) {
                     value.append(epochStr).append(':');
                  }
               }
               value.append(pyString(ver));
               if (rel != NULL && rel != Py_None) {
                  value.append('-').append(pyString(rel));
               }
            }
         }

         values.append(value);

         Py_DECREF(item);
      }

      Py_DECREF(it);
   }

   return makeSorted(values);
}

#if 0
QString  YumIO::getRpm(PyObject*  pkg,
                       char*      rpmKey)
{
   Python::PyGILStateWrapper  gsw;

   if (PyObject_HasAttrString(pkg, "hdr")) {
      return pyStringXDec(getRpmHdrValue(pkg, rpmKey));
   } else {
      return QString();
   }
}
#endif

QString  YumIO::getRpmTime(PyObject*    pkg,
                           const char*  rpmKey)
{
   Python::PyGILStateWrapper  gsw;

   if (PyObject_HasAttrString(pkg, "hdr")) {
      return timeFormat(pyTimeDec(getRpmHdrValue(pkg, rpmKey)));
   } else {
      return QString();
   }
}

QString  YumIO::getRpmList(PyObject*    pkg,
                           const char*  rpmKey)
{
   Python::PyGILStateWrapper  gsw;

   if (PyObject_HasAttrString(pkg, "hdr")) {
      QString  value;

      PyObject*  list = getRpmHdrValue(pkg, rpmKey);
      if (list != NULL && list != Py_None) {
         PyObject*  it = PyObject_GetIter(list);
         for (;;) {
            PyObject*  item = PyIter_Next(it);
            if (item == NULL) break;
            if (!value.isEmpty()) {
               value.append('\n');
            }
            value.append(pyString(item));
            Py_DECREF(item);
         }
         Py_DECREF(it);
         Py_DECREF(list);
      }

      return value;
   } else {
      return QString();
   }
}

QString  YumIO::getGroupGrps (const Package::Entry&  entry,
                              const char*            attr_22)
{
   char*  copy = new char[strlen(attr_22)+1];
   strcpy(copy, attr_22);
   QString  result = getGroupGrps(entry, copy);
   delete [] copy;
   return result;
}

QString  YumIO::getGroupGrps (const Package::Entry&  entry,
                              char*                  attr_22)
{
   Python::PyGILStateWrapper  gsw;

   if (hasDistribGroups()) {
      return QString();
   } else {
      PyObject*  seq = getGroupItem_22(entry, attr_22);
      if (seq == NULL) {
         return QString();
      }

      if (PySequence_Length(seq) > 0) {
         QStringList  values;

         PyObject*  it = PyObject_GetIter(seq);
         for (;;) {
            PyObject*  grpId = PyIter_Next(it);
            if (grpId == NULL) break;
            values.append(pyString(grpId));
            Py_DECREF(grpId);
         }

         Py_DECREF(it);

         return makeSorted(values);
      } else {
         return QString();
      }
   }
}

QString  YumIO::getGroupPkgs (const Package::Entry&  entry,
                              const char*            attr_22,
                              const char*            attr_25)
{
   char*  attr_22_copy = new char[strlen(attr_22)+1];
   char*  attr_25_copy = new char[strlen(attr_25)+1];
   strcpy(attr_22_copy, attr_22);
   strcpy(attr_25_copy, attr_25);
   QString  result = getGroupPkgs(entry, attr_22_copy, attr_25_copy);
   delete [] attr_25_copy;
   delete [] attr_22_copy;
   return result;
}

QString  YumIO::getGroupPkgs (const Package::Entry&  entry,
                              char*                  attr_22,
                              char*                  attr_25)
{
   Python::PyGILStateWrapper  gsw;

   if (hasDistribGroups()) {
      PyObject*  dict = getGroupItem_25(entry, attr_25);
      if (dict == NULL) {
         return QString();
      }

      if (PyDict_Size(dict) > 0) {
         QStringList  values;

         Py_ssize_t  pos = 0;
         for (;;) {
            PyObject*  pkgName;
            PyObject*  value;
            if (!PyDict_Next(dict, &pos, &pkgName, &value)) break;

            values.append(pyString(pkgName));
         }

         return makeSorted(values);
      } else {
         return QString();
      }
   } else {
      PyObject*  seq = getGroupItem_22(entry, attr_22);
      if (seq == NULL) {
         return QString();
      }

      if (PySequence_Length(seq) > 0) {
         QStringList  values;

         PyObject*  it = PyObject_GetIter(seq);
         for (;;) {
            PyObject*  pkgName = PyIter_Next(it);
            if (pkgName == NULL) break;
            values.append(pyString(pkgName));
            Py_DECREF(pkgName);
         }

         Py_DECREF(it);

         return makeSorted(values);
      } else {
         return QString();
      }
   }
}

QString  YumIO::makeSorted (QStringList&  list)
{
   qSort(list.begin(), list.end());

   QString  result;

   QStringList::const_iterator  it = list.begin();
   if (it != list.end()) {
      result.append(*it); ++it;
   }
   for (; it != list.end(); ++it) {
      result.append("\n").append(*it);
   }

   return result;
}

QString  YumIO::makeSortedUniq (QStringList&  list)
{
   qSort(list.begin(), list.end());

   QString  result;

   QStringList::const_iterator  it   = list.begin();
   QStringList::const_iterator  prev = list.begin();
   if (it != list.end()) {
      result.append(*it); prev = it; ++it;
   }
   for (; it != list.end(); prev = it, ++it) {
      if (*it != *prev) {
         result.append("\n").append(*it);
      }
   }

   return result;
}

PyObject*  YumIO::getRpmHdrValue(PyObject*    pkg,
                                 const char*  rpmKey)
{
   if (!PyObject_HasAttrString(pkg, "hdr")) {
      error(tr("getRpmHdrValue called for package with no hdr"
               " (i.e. not installed)"));
      return NULL;
   }

   PyObject*  hdr = PyObject_GetAttrString(pkg, "hdr");

   char  key[1024];
   sprintf(key, "RPMTAG_%s", rpmKey);

   PyObject*  tag = PyDict_GetItemString(rpm_dict, key);
   if (tag == NULL) {
      error(tr("Internal error: cannot find rpm.%1").arg(key));
      return NULL;
   }

   return PyObject_GetItem(hdr, tag);
}

class ApplyThread : public QThread 
{
public:
   ApplyThread(YumIO*       yumIO,
               QEventLoop&  eventLoop)
      : yumIO(yumIO),
        eventLoop(eventLoop)
   {}
protected:
   YumIO*       yumIO;
   QEventLoop&  eventLoop;
public:
   bool         result;
protected:
   void run()
   {
      result = yumIO->applyReal();
      eventLoop.exit();
   }
};

bool YumIO::apply () 
{
   QEventLoop   eventLoop;
   ApplyThread  thread(this, eventLoop);
   execWorkThread(eventLoop, thread);
   return thread.result;
}

bool YumIO::applyReal () 
{
   blocker.block();

   ProgressUser      pgUser(pg, totalProgressBar, itemProgressBar);
   YumIODialogsUser  yumIODialogsUser(dialogs, parent);

   log(Log::ACTION, 0, tr("Applying"));

   Python::PyGILStateWrapper  gsw;

   pg->startTotal(tr(PG_D_TOTAL_TEXT), PG_D_TOTAL);

   pg->startItem(tr(PG_D_INITIALIZE_TEXT), PG_D_INITIALIZE);

   ts.descr               = QString();
   ts.confirmed           = false;
   ts.act                 = 0;
   ts.updates.clear();
   ts.dep.totalTs         = 0;
   ts.dep.totalGrabs      = 0;
   ts.run.nominalGrabSize = 0;
   ts.run.totalGrabSize   = 0;
   ts.run.totalTsSize     = 0;
   ts.run.lastTs          = 0;

   PyObject*  installs = PyList_New(0);
   PyObject*  updates  = PyList_New(0);
   PyObject*  removes  = PyList_New(0);

   {
      Package::EntryList::const_iterator  it = list.begin();
      for (; it != list.end(); ++it) {
         Package::Entry*  entry = *it;
         entry->apply.action = entry->action;
         entry->apply.reason = entry;
      }
   }

   bool  status = applyGroups();
   if (!status) {
      return false;
   }

   // I'm not bothering to Py_INCREF and Py_DECREF the packages themselves,
   // because I already am holding a reference for them because they're in
   // the entries list.  That's good enough.

   {
      Package::EntryList::const_iterator  it = list.begin();
      for (; it != list.end(); ++it) {
         Package::Entry*  entry = *it;

         if (entry->kind != Package::GROUP) {
            switch (entry->apply.action) {
            case Package::UPDATE:
               {
                  PyObject*  pair = PyList_New(0);
                  status = PyList_Append(pair, entry->obj);
                  if (status != 0) {
                     error(tr("Unable to add new package to pair for updated"
                              " package %1").arg(entry->name));
                  }
                  status = PyList_Append(pair, entry->next->obj);
                  if (status != 0) {
                     error(tr("Unable to add old package to pair for updated"
                              " package %1").arg(entry->name));
                  }
                  status = PyList_Append(updates, pair);
                  if (status != 0) {
                     error(tr("Unable to add package %1 to updates"
                              " list").arg(entry->name));
                  }
                  ts.dep.totalTs += 2;
                  if (!QFile::exists(getLocalHdr(entry->obj))) {
                     ts.dep.totalGrabs++;
                  }
               }
               break;
            case Package::INSTALL:
               {
                  int  status = PyList_Append(installs, entry->obj);
                  if (status != 0) {
                     error(tr("Unable to add package %1 to installs"
                              " list").arg(entry->name));
                  }
                  ts.dep.totalTs++;
                  if (!QFile::exists(getLocalHdr(entry->obj))) {
                     ts.dep.totalGrabs++;
                  }
               }
               break;
            case Package::REMOVE:
               {
                  Package::Entry*  e = entry;
                  if (e->next != NULL) {
                     e = e->next;
                  }
                  int  status = PyList_Append(removes, e->obj);
                  if (status != 0) {
                     error(tr("Unable to add package %1 to removes"
                              " list").arg(entry->name));
                  }
                  ts.dep.totalTs++;
               }
               break;
            case Package::NONE:
               // DO NOTHING
               break;
            }
         }
      }
   }

   // Assume that all headers are roughly the same size and use number of
   // grabs as the number of acts.
   unsigned int  grabActs = PG_D_GRAB(ts.dep.totalGrabs);
   ts.dep.depSolveActs      = PG_D_DEPSOLVE_ESTIMATE(ts.dep.totalTs);
   ts.dep.totalDepSolveActs = ts.dep.depSolveActs;
   unsigned int  numActs = grabActs + ts.dep.totalDepSolveActs;
   pg->setupItemActs(PG_D_DEPSOLVE, numActs);

   {
#ifndef DEBUG_DISABLE_CAPTURERS
      YumIOCapturersUser  capturersUser(capturers, this);
#endif
      PyObject*  xcptYumBaseError = pyLookup("yum.Errors.YumBaseError");
      pyExceptExpect(xcptYumBaseError);
      pyExceptExpect(PyExc_IOError);
      PyObject*  applyStatus = pyMethodCall(yumIO, "apply", "(O,O,O)", 
                                            installs, updates, removes);
      if (pyExceptMatches(applyStatus, xcptYumBaseError)) {
         pythonExceptionV(tr("Error applying changes"));
         return false;
      } else if (pyExceptMatches(applyStatus, PyExc_IOError)) {
         pythonException(tr("Error applying changes: Exiting on Broken Pipe"));
         return false;
      }
      pyExceptClear();
   }

   pyMethodCall(yumIO, "flush", "()");

   Py_DECREF(removes);
   Py_DECREF(updates);
   Py_DECREF(installs);

   pg->reset();

   ts.updates.clear();

   return ts.confirmed;
}

bool  YumIO::cleanHeaders ()
{
   log(Log::ACTION, 0, tr("Cleaning Headers"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanHeaders", "()");

   return (PyInt_AsLong(PyTuple_GetItem(pair, 0)) == 0);
}

bool  YumIO::cleanPackages ()
{
   log(Log::ACTION, 0, tr("Cleaning Packages"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanPackages", "()");

   return (PyInt_AsLong(PyTuple_GetItem(pair, 0)) == 0);
}

bool  YumIO::cleanXmlMetadata ()
{
   log(Log::ACTION, 0, tr("Cleaning Xml Metadata"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanMetadata", "()");

   return (PyInt_AsLong(PyTuple_GetItem(pair, 0)) == 0);
}

bool  YumIO::cleanPickledCache ()
{
   log(Log::ACTION, 0, tr("Cleaning Pickled Cache"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanPickles", "()");

   return (PyInt_AsLong(PyTuple_GetItem(pair, 0)) == 0);
}

bool  YumIO::cleanAll ()
{
   log(Log::ACTION, 0, tr("Cleaning All"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  hdrPair  = pyMethodCall(yumIO, "cleanHeaders", "()");
   PyObject*  pkgPair  = pyMethodCall(yumIO, "cleanPackages", "()");
   PyObject*  metaPair = pyMethodCall(yumIO, "cleanMetadata", "()");

   long  hdrValue  = PyInt_AsLong(PyTuple_GetItem(hdrPair, 0));
   long  pkgValue  = PyInt_AsLong(PyTuple_GetItem(pkgPair, 0));
   long  metaValue = PyInt_AsLong(PyTuple_GetItem(metaPair, 0));
   long  pickValue = 0;
   
   if (hasPickledCache()) {
      PyObject*  pickPair = pyMethodCall(yumIO, "cleanPickles", "()");
      pickValue = PyInt_AsLong(PyTuple_GetItem(pickPair, 0));
   }

   return (hdrValue == 0
           && pkgValue == 0
           && metaValue == 0
           && pickValue == 0);
}

/*
** yumIO callback functions from Python
*/

PyObject*  YumIO::stdoutWrite(PyObject*  self, PyObject*  args)
{ 
   return write(self, args, Log::STDOUT, 0, wr.stdoutBuffer);
}

PyObject*  YumIO::stderrWrite(PyObject*  self, PyObject*  args)
{ 
   return write(self, args, wr.stderrKind, wr.stderrValue, wr.stderrBuffer);
}

PyObject*  YumIO::write(PyObject*     self, 
                        PyObject*     args,
                        Log::Kind     kind,
                        unsigned int  value,
                        QString&      buffer)
{
   char*  content;
   
   if (!PyArg_ParseTuple(args, "s", &content)) return NULL;

   write(content, kind, value, buffer);

   Py_INCREF(Py_None);
   return Py_None;
}

void  YumIO::write (QString       content,
                    Log::Kind     kind,
                    unsigned int  value,
                    QString&      buffer)
{
   buffer.append(content);

   QStringList  lines = buffer.split("\n");

   QStringList::const_iterator  it   = lines.begin();
   QStringList::const_iterator  last = lines.end() - 1;

   if (it != last) {
      logEntries.lock();

      int i = 0;
      for (; it != last; ++it, ++i) {
         logEntries.append(Log::Entry(kind, value, *it));
      }

      logEntries.unlock();
      logEntries.emitAppended();
   }

   // The last line in the list is, by definition, unfinished.  (Note that
   // a very common case would be when it's the empty string.)  Sit on it
   // until further text comes along to finish it with a newline.
   buffer = *it;
}

void  YumIO::writeStdout (QString  content)
{
   write(content, Log::STDOUT, 0, wr.stdoutBuffer);
}

void  YumIO::writeStderr (QString  content)
{
   write(content, Log::ERROR, 0, wr.stderrBuffer);
}

PyObject*  YumIO::log(PyObject*  self, PyObject*  args)
{ 
   long   value;
   char*  msg;
   
   if (!PyArg_ParseTuple(args, "ls", &value, &msg)) return NULL;

   log(Log::NORMAL, value, msg);

   if (strncmp(msg,
               APPLY_DEPSOLVE_PREFIX,
               STRLEN(APPLY_DEPSOLVE_PREFIX)) == 0) {
      // I hate having to know this string, but there is no callback
      // for a package added to the transaction by depSolve.  (There's
      // one for the Requires: string that caused it, but that's not
      // good enough.)
      dsTsInfo(msg + STRLEN(APPLY_DEPSOLVE_PREFIX));
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::errorlog(PyObject*  self, PyObject*  args)
{ 
   long   value;
   char*  msg;
   
   if (!PyArg_ParseTuple(args, "ls", &value, &msg)) return NULL;

   log(Log::ERROR, value, msg);

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::collectProgress(PyObject*  self, PyObject*  args)
{ 
   long       current;
   long       total;
   PyObject*  nameObj;
   
   if (!PyArg_ParseTuple(args, "llO", &current, &total, &nameObj)) return NULL;

   QString  name;
   if (PyString_Check(nameObj)) {
      name = pyString(nameObj);
   } else if (nameObj == Py_None) {
      name = "";
   } else {
      error(tr("Internal error: unexpected progress name"));
      name = "";
   }

   if (repoProgress.act > 0
       && total == repoProgress.total
       && name == repoProgress.name) {
      pg->setItemValue(current);
   } else {
      repoProgress.total = total;
      repoProgress.name  = name;
      QString  activity;
      if (name == COLL_GRAB_METADATA) {
         activity = tr("Downloading repository %1 for %2")
                       .arg(repoProgress.kind)
                       .arg(repoProgress.repos[repoProgress.act / PG_APR]);
         repoProgress.act++;
      } else {
         if (name == "") {
            activity = tr("Reading repository %1 for %2").arg(repoProgress.kind)
                                                         .arg(name);
         } else {
            activity = tr("Reading repository %1").arg(repoProgress.kind);
         }

         // If the MD Read happened, this just increments.  If not, this
         // effectively increments by two.
         repoProgress.act = (repoProgress.act / PG_APR * PG_APR + PG_APR);
      }
      pg->startItemAct(activity, total, repoProgress.act);
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::grabMeterStart (PyObject*  self, PyObject*  args)
{
   char*          filename;
   char*          url;
   char*          basename;
   Package::Size  size;
   PyObject*      text;

   if (!PyArg_ParseTuple(args, "sssLO", 
                         &filename, &url, &basename, &size, &text)) return NULL;

#if defined(DEBUG_GRAB_METER)
   printf("grabMeterStart: %s size=%llu\n", basename(filename), size);
#endif

   QString  name = filename;
   int      slash = name.lastIndexOf('/');
   if (slash != -1) {
      name.remove(0, slash+1);
   }
   name.remove(QRegExp("buffet\?.*=.*&.*=.*&path="));
   if (name.endsWith(".rpm") || name.endsWith(".hdr")) {
      name.chop(4);
   }

   QString  activity;
   if (ts.confirmed) {
      activity = tr("Downloading %1").arg(name);
      // All rpms are not the same size, so use size as the number of acts
      ts.act += PG_R_GRAB(size);
   } else {
      activity = tr("Downloading %1 header").arg(name);
      // Assume that all headers are roughly the same size and use number of
      // grabs as the number of acts.
      ts.act += PG_D_GRAB(1);
   }
   pg->startItemAct(activity, size, ts.act);

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::grabMeterUpdate (PyObject*  self, PyObject*  args)
{
   long long  amount;

   if (!PyArg_ParseTuple(args, "L", &amount)) return NULL;

#if defined(DEBUG_GRAB_METER_DETAIL)
   printf("grabMeterUpdate: %llu\n", amount);
#endif

   pg->setItemValue(amount);

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::grabMeterEnd (PyObject*  self, PyObject*  args)
{
   long long  amount;

   if (!PyArg_ParseTuple(args, "L", &amount)) return NULL;

#if defined(DEBUG_GRAB_METER)
   printf("grabMeterEnd\n");
#endif

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::dsTsCheck(PyObject*  self, PyObject*  args)
{ 
#if defined(DEBUG_TS)
   printf("dsTsCheck\n");
#endif

   ts.act += ts.dep.depSolveActs;
   pg->startItemAct(tr("Resolve Dependencies"), 1, ts.act);

   Py_INCREF(Py_None);
   return Py_None;
}

void  YumIO::dsTsInfo (char*  msg)
{
#if defined(DEBUG_TS)
   printf("dsTsInfo\n");
#endif

   if (strncmp(msg, 
               APPLY_DEPSOLVE_UPDATING, 
               STRLEN(APPLY_DEPSOLVE_UPDATING)) == 0) {
      QString  name(msg + STRLEN(APPLY_DEPSOLVE_UPDATING));
      int      space = name.indexOf(' ');
      if (space == -1) {
         error(tr("Internal error: unexpected TSINFO: %1").arg(msg));
         return;
      }
      name.resize(space);
      dsTsNewPkg(name, 2, true);
   } else if (strncmp(msg, 
                      APPLY_DEPSOLVE_OBSOLETING, 
                      STRLEN(APPLY_DEPSOLVE_OBSOLETING)) == 0) {
      QString  oldName(msg + STRLEN(APPLY_DEPSOLVE_OBSOLETING));
      int      oldSpace = oldName.indexOf(' ');
      if (oldSpace == -1) {
         error(tr("Internal error: unexpected TSINFO: %1").arg(msg));
         return;
      }
      int      withIx = oldName.indexOf(APPLY_DEPSOLVE_OBSOLETING_WITH, 
                                        oldSpace + 1);
      if (withIx == -1) {
         error(tr("Internal error: unexpected TSINFO: %1").arg(msg));
         return;
      }
      QString  newName(oldName
                       + withIx + STRLEN(APPLY_DEPSOLVE_OBSOLETING_WITH));
      int      newSpace = newName.indexOf(' ');
      if (newSpace == -1) {
         error(tr("Internal error: unexpected TSINFO: %1").arg(msg));
         return;
      }
      oldName.resize(oldSpace);
      newName.resize(newSpace);
      dsTsNewPkg(oldName, 1, false);
      dsTsNewPkg(newName, 1, true);
   } else if (strncmp(msg, 
                      APPLY_DEPSOLVE_MARKING, 
                      STRLEN(APPLY_DEPSOLVE_MARKING)) == 0) {
      QString  name(msg + STRLEN(APPLY_DEPSOLVE_MARKING));
      int      space = name.indexOf(' ');
      if (space == -1) {
         error(tr("Internal error: unexpected TSINFO: %1").arg(msg));
         return;
      }
      name.resize(space);
      dsTsNewPkg(name, 
                 strstr(msg, APPLY_DEPSOLVE_MARKING_UPDATE) != NULL ? 2 : 1,
                 true);
   } else if (strstr(msg, APPLY_DEPSOLVE_ERASE) != NULL) {
      // DO NOTHING
   } else {
      error(tr("Internal error: unexpected TSINFO: %1").arg(msg));
      return;
   }
}

void  YumIO::dsTsNewPkg (QString       name,
                         unsigned int  numTs,
                         bool          needHdr)
{
#if defined(DEBUG_TS)
   printf("dsTsNewPkg\n");
#endif

   Package::Entry*  entry = hash[EntryKey(false, name)];
   if (entry == NULL) {
      error(tr("Internal error:"
               " TSINFO mentions unknown package: %1").arg(name));
      return;
   }
   if (entry->state == Package::INSTALLED) {
      error(tr("Internal error:"
               " TSINFO mentions intent to install or upgrade package that"
               " already seems up-to-date: %1").arg(name));
      return;
   }

   ts.dep.totalTs += numTs;
   if (needHdr && !QFile::exists(getLocalHdr(entry->obj))) {
      ts.dep.totalGrabs++;
   }
   unsigned int  grabActs = PG_D_GRAB(ts.dep.totalGrabs);
   unsigned int  numActs  = grabActs + ts.dep.totalDepSolveActs;
   pg->changeItemActs(numActs); // PG_D_DEPSOLVE
}


PyObject*  YumIO::dsRestartLoop(PyObject*  self, PyObject*  args)
{ 
#if defined(DEBUG_TS)
   printf("dsRestartLoop\n");
#endif

   // We'll be doing another depsolve, so re-estimate and add the new
   // estimate.
   unsigned int  grabActs = PG_D_GRAB(ts.dep.totalGrabs);
   ts.dep.depSolveActs       = PG_D_DEPSOLVE_ESTIMATE(ts.dep.totalTs);
   ts.dep.totalDepSolveActs += ts.dep.depSolveActs;
   unsigned int  numActs = grabActs + ts.dep.totalDepSolveActs;
   pg->changeItemActs(numActs); // PG_D_DEPSOLVE

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::userconfirm(PyObject*  self, PyObject*  args)
{ 
#if defined(DEBUG_TS)
   printf("userconfirm\n");
#endif

   pg->endItemActs(); // PG_D_DEPSOLVE

   pg->reset(); // PG_D_*

   int  result = dialogs->confirm(tr("Apply"), ts.descr);

   ts.confirmed = (result == QDialog::Accepted);

   if (ts.confirmed) {
      log(Log::ACTION, 0, tr("Applying: OK"));
   } else {
     log(Log::ACTION, 0, tr("Not Applying: Canceled"));
   }

   if (ts.confirmed) {
      // The PG_R_GRAB_RPMS_NOMINAL fractional part of PG_R_TOTAL is based
      // on actually having to grab all the rpm's.  If some already are in
      // the cache, then scale it back accordinly.
      ts.run.pgRGrabRpms = (ts.run.nominalGrabSize == 0
                            ? 0
                            : 
                            (ts.run.totalGrabSize * PG_R_GRAB_RPMS_NOMINAL
                             / ts.run.nominalGrabSize));

      pg->startTotal(tr(PG_R_TOTAL_TEXT), PG_R_TOTAL);

      pg->startItem(tr(PG_R_INITIALIZE_TEXT), PG_R_INITIALIZE);

      // All rpms are not the same size, so use size as the number of acts
      pg->setupItemActs(ts.run.pgRGrabRpms, ts.run.totalGrabSize);
   }

   return PyInt_FromLong(ts.confirmed);
}

PyObject*  YumIO::reportDownloadSize(PyObject*  self, PyObject*  args)
{ 
   PyObject*  packages;
   
   if (!PyArg_ParseTuple(args, "O", &packages)) return NULL;

#if defined(DEBUG_TS)
   printf("reportDownloadSize\n");
#endif

   Package::Size  size = 0;

   PyObject*  it = PyObject_GetIter(packages);
   for (;;) {
      PyObject*  pkg = PyIter_Next(it);
      if (pkg == NULL) break;
      size += sizeOf(pkg);
      Py_DECREF(pkg);
   }
   
   Py_DECREF(it);

   ts.descr.append(tr("Total download size: %1\n").arg(bytesFormat(size)));

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::listTransaction(PyObject*  self, PyObject*  args)
{ 
   PyObject*  yumIO;
   
   if (!PyArg_ParseTuple(args, "O", &yumIO)) return NULL;

#if defined(DEBUG_TS)
   printf("listTransaction\n");
#endif

   PyObject*  tsInfo;
   PyObject*  lists;
   PyObject*  updated;
   PyObject*  installed;
   PyObject*  removed;
   PyObject*  obsoleted;
   PyObject*  depup;
   PyObject*  depinst;
   PyObject*  deprem;

   if (!makeLists(yumIO, tsInfo, lists, 
                  updated, installed, removed, obsoleted, 
                  depup, depinst, deprem)) {
      return NULL;
   }

   ts.descr.append(tr("Transaction to be performed:\n"));

   if (   !displayTransactionPkgsRepo(tr("Install"), installed)
       || !displayTransactionPkgsRepo(tr("Update"),  updated)
       || !displayTransactionPkgs    (tr("Remove"),  removed)) {
      Py_DECREF(lists);
      Py_DECREF(tsInfo);
      return NULL;
   }

   if (   PySequence_Length(depinst) > 0
       || PySequence_Length(depup) > 0
       || PySequence_Length(deprem) > 0) {
      ts.descr.append(tr("Performing the following to resolve"
                         " dependencies:\n"));
      if (   !displayTransactionPkgsRepo(tr("Install"), depinst)
          || !displayTransactionPkgsRepo(tr("Update"),  depup)
          || !displayTransactionPkgs    (tr("Remove"),  deprem)) {
         Py_DECREF(lists);
         Py_DECREF(tsInfo);
         return NULL;
      }
   }

   if (PySequence_Length(obsoleted) > 0) {
      ts.descr.append(tr("Other Transactions:\n"));
      if (!displayTransactionPkgsObs(tr("Obsoleting"), obsoleted)) {
         Py_DECREF(lists);
         Py_DECREF(tsInfo);
         return NULL;
      }
   }

   noteTransactionPkgs(Package::INSTALL, installed);
   noteTransactionPkgs(Package::UPDATE,  updated);
   noteTransactionPkgs(Package::REMOVE,  removed);

   noteTransactionPkgs(Package::INSTALL, depinst);
   noteTransactionPkgs(Package::UPDATE,  depup);
   noteTransactionPkgs(Package::REMOVE,  deprem);

   noteTransactionPkgs(Package::REMOVE,  obsoleted);

   Py_DECREF(lists);
   Py_DECREF(tsInfo);

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::tsInitCallback(PyObject*  self, PyObject*  args)
{
   long  output;

   if (!PyArg_ParseTuple(args, "l", &output)) return NULL;

#if defined(DEBUG_TS)
   printf("tsInitCallback (output=%lu)\n", output);
#endif

   ts.act = 0;

   if (output) {
      pg->endItemActs(); // for PG_R_TRANS_TEST

      pg->startItem(tr(PG_R_RUN_TRANS_INIT_TEXT), PG_R_RUN_TRANS_INIT);
      // yum does another transaction test as part of the transaction
      Progress::Steps  testActs = PG_R_TEST_TRANS_ESTIMATE(ts.run.totalTsSize);
      Progress::Steps  numActs  = ts.run.totalTsSize + testActs;
      pg->setupItemActs(PG_R_RUN_TRANS, numActs);
   } else {
      pg->endItemActs(); // for PG_R_GRAB_RPMS

      pg->startItem(tr(PG_R_TRANS_TEST_INIT_TEXT), PG_R_TRANS_TEST_INIT);
      Progress::Steps  testActs = PG_R_TEST_TRANS_ESTIMATE(ts.run.totalTsSize);
      Progress::Steps  numActs  = testActs;
      pg->setupItemActs(PG_R_TRANS_TEST, numActs);
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::tsCallback(PyObject*  self, PyObject*  args)
{
   // This routine is called in yum <= 3.2.2.  tsEvent is called instead in
   // yum >= 3.2.3.

   long             what;
   Progress::Steps  progress;
   Progress::Steps  total;
   PyObject*        h;
   PyObject*        user;
   bool             understood = false;

   if (!PyArg_ParseTuple(args, "lLLOO", 
                         &what, &progress, &total, &h, &user)) return NULL;

#if defined(DEBUG_TS_DETAIL)
   QString  whatStr;
   if (what & RPMCALLBACK_INST_OPEN_FILE)  whatStr.append(" INST_OPEN_FILE");
   if (what & RPMCALLBACK_INST_CLOSE_FILE) whatStr.append(" INST_CLOSE_FILE");
   if (what & RPMCALLBACK_TRANS_START)     whatStr.append(" TRANS_START");
   if (what & RPMCALLBACK_TRANS_PROGRESS)  whatStr.append(" TRANS_PROGRESS");
   if (what & RPMCALLBACK_TRANS_STOP)      whatStr.append(" TRANS_STOP");
   if (what & RPMCALLBACK_INST_START)      whatStr.append(" INST_START");
   if (what & RPMCALLBACK_INST_PROGRESS)   whatStr.append(" INST_PROGRESS");
   if (what & RPMCALLBACK_UNINST_START)    whatStr.append(" UNINST_START");
   if (what & RPMCALLBACK_UNINST_PROGRESS) whatStr.append(" UNINST_PROGRESS");
   if (what & RPMCALLBACK_UNINST_STOP)     whatStr.append(" UNINST_STOP");
   whatStr = whatStr.trimmed();
   printf("tsCallback (what=%s, progress=%llu/%llu)\n", 
          whatStr.toStdString().c_str(), progress, total);
#endif

   if (what & RPMCALLBACK_INST_OPEN_FILE) {
      // DO NOTHING: handled by Python code
      understood = true;
   }
   if (what & RPMCALLBACK_INST_CLOSE_FILE) {
      // DO NOTHING: handled by Python code
      understood = true;
   }
   if (what & RPMCALLBACK_TRANS_START) {
      ts.act += PG_R_TEST_TRANS_ESTIMATE(ts.run.totalTsSize);
      pg->startItemAct(tr("Transaction Test"), total, ts.act);
      understood = true;
   }
   if (what & RPMCALLBACK_TRANS_PROGRESS) {
      pg->setItemValue(progress);
      understood = true;
   }
   if (what & RPMCALLBACK_TRANS_STOP) {
      pg->setItemValue(total);
      understood = true;
   }
   if (what & RPMCALLBACK_INST_START) {
      PyObject*  hdr   = PyTuple_GetItem(h, 0);
      QString    name  = pyMappingString(hdr, "name");
#if defined(DEBUG_TS_DETAIL)
      printf("           (name=%s)\n", name.toStdString().c_str());
#endif
      Package::Entry*  entry = hash[EntryKey(false, name)];
      QString          activity;
      if (ts.updates.contains(name)) {
         activity = tr("Updating %1").arg(name);
         ts.act += PG_R_UPDATE(entry->size);
      } else {
         activity = tr("Installing %1").arg(name);
         ts.act += PG_R_INSTALL(entry->size);
      }
      // RPMInstallCallback.callback fails to say anything for this case,
      // so do it manually here.
      log(Log::NORMAL, 1, activity);
      pg->startItemAct(activity, total, ts.act);
      understood = true;
   }
   if (what & RPMCALLBACK_INST_PROGRESS) {
      pg->setItemValue(progress);
      understood = true;
   }
   if (what & RPMCALLBACK_UNINST_START) {
      QString               name  = pyString(h);
#if defined(DEBUG_TS_DETAIL)
      printf("           (name=%s))\n", name.toStdString().c_str());
#endif
      Package::Entry*  entry = hash[EntryKey(false, name)];
      QString          activity;
      if (ts.updates.contains(name)) {
         // update: "removal" of old package
         activity = tr("Completing update for: %1").arg(name);
         ts.act += PG_R_COMPLETE_UPDATE(entry->next->size);
      } else {
         activity = tr("Removing %1").arg(name);
         ts.act += PG_R_REMOVE(entry->size);
      }
      pg->startItemAct(activity, total, ts.act);
      understood = true;
   }
   if (what & RPMCALLBACK_UNINST_PROGRESS) {
      pg->setItemValue(progress);
      understood = true;
   }
   if (what & RPMCALLBACK_UNINST_STOP) {
      pg->setItemValue(total);
      understood = true;
   }

   if (!understood) {
      warning(tr("Internal error: YumIO::tsCallback what=%1 not understood;"
                 " ignoring").arg(what));
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::tsEvent(PyObject*  self, PyObject*  args)
{
   // This routine is called in yum >= 3.2.3.  tsCallback is called instead in
   // yum <= 3.2.2.

   PyObject*        pkg;
   long             action;
   Progress::Steps  progress;
   Progress::Steps  total;
   unsigned long    currentTs;
   unsigned long    totalTs;
   bool             understood = false;

   if (!PyArg_ParseTuple(args, "OlLLll",
                         &pkg, &action,
                         &progress, &total,
                         &currentTs, &totalTs)) return NULL;

#if defined(DEBUG_TS_DETAIL)
   QString  actStr;
   if      (action == TS_UPDATE)      actStr = "TS_UPDATE";
   else if (action == TS_INSTALL)     actStr = "TS_INSTALL";
   else if (action == TS_TRUEINSTALL) actStr = "TS_TRUEINSTALL";
   else if (action == TS_ERASE)       actStr = "TS_ERASE";
   else if (action == TS_OBSOLETED)   actStr = "TS_OBSOLETED";
   else if (action == TS_OBSOLETING)  actStr = "TS_OBSOLETING";
   else if (action == TS_AVAILABLE)   actStr = "TS_AVAILABLE";
   else if (action == TS_UPDATED)     actStr = "TS_UPDATED";
   actStr = actStr.trimmed();
   printf("tsEvent (action=%s, progress=%llu/%llu)\n", 
          actStr.toStdString().c_str(), progress, total);
#endif

   if (IS_INSTALL(action) || IS_UPDATE(action)) {
      if (currentTs != ts.run.lastTs) {
         QString  name  = pyStringDec(PyObject_GetAttrString(pkg, "name"));
#if defined(DEBUG_TS_DETAIL)
         printf("        (name=%s)\n", name.toStdString().c_str());
#endif
         Package::Entry*  entry = hash[EntryKey(false, name)];
         QString          activity;
         if (IS_UPDATE(action)) {
            activity = tr("Updating %1").arg(name);
            ts.act += PG_R_UPDATE(entry->size);
         } else {
            activity = tr("Installing %1").arg(name);
            ts.act += PG_R_INSTALL(entry->size);
         }
         log(Log::NORMAL, 1, activity);
         pg->startItemAct(activity, total, ts.act);
         ts.run.lastTs = currentTs;
      }
      pg->setItemValue(progress);
      understood = true;
   } else if (IS_REMOVE(action) || IS_COMPLETING_UPDATE(action)) {
      if (currentTs != ts.run.lastTs) {
         QString  name  = pyString(pkg);
#if defined(DEBUG_TS_DETAIL)
         printf("        (name=%s)\n", name.toStdString().c_str());
#endif
         Package::Entry*  entry = hash[EntryKey(false, name)];
         QString          activity;
         if (IS_COMPLETING_UPDATE(action)) {
            activity = tr("Completing update for: %1").arg(name);
            ts.act += PG_R_COMPLETE_UPDATE(entry->next->size);
         } else {
            activity = tr("Removing %1").arg(name);
            ts.act += PG_R_REMOVE(entry->size);
         }
         pg->startItemAct(activity, total, ts.act);
         ts.run.lastTs = currentTs;
      }
      pg->setItemValue(progress);
      understood = true;
   } else if (action == TS_AVAILABLE) {
      // DO NOTHING
      understood = true;
   }

   if (!understood) {
      warning(tr("Internal error: YumIO::tsEvent action=%1 not understood;"
                 " ignoring").arg(action));
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::postTransactionOutput(PyObject*  self, PyObject*  args)
{ 
   PyObject*  yumIO;
   
   if (!PyArg_ParseTuple(args, "O", &yumIO)) return NULL;

   pg->endItemActs(); // for PG_R_RUN_TRANS

   PyObject*  tsInfo;
   PyObject*  lists;
   PyObject*  updated;
   PyObject*  installed;
   PyObject*  removed;
   PyObject*  obsoleted;
   PyObject*  depup;
   PyObject*  depinst;
   PyObject*  deprem;

   if (!makeLists(yumIO, tsInfo, lists, 
                  updated, installed, removed, obsoleted, 
                  depup, depinst, deprem)) {
      return NULL;
   }

#if defined(DEBUG_TS)
   printf("postTransactionOutput\n");
#endif

   if (   !displayTransactionPkgs(tr("Removed"),              removed)
       || !displayTransactionPkgs(tr("Dependency Removed"),   deprem)
       || !displayTransactionPkgs(tr("Installed"),            installed)
       || !displayTransactionPkgs(tr("Dependency Installed"), depinst)
       || !displayTransactionPkgs(tr("Updated"),              updated)
       || !displayTransactionPkgs(tr("Dependency Updated"),   depup)
       || !displayTransactionPkgs(tr("Obsoleted"),            obsoleted)) {
      Py_DECREF(lists);
      Py_DECREF(tsInfo);
      return NULL;
   }

   Py_DECREF(lists);
   Py_DECREF(tsInfo);

   return PyString_FromString("");
}

/*
** Utility functions
*/

QString  YumIO::getLocalPkg (PyObject*  pkg)
{
   PyObject*  obj = pyMethodCall(pkg, "localPkg", "()");
   if (obj == NULL || obj == Py_None) return QString();
   return pyStringDec(obj);
}

QString  YumIO::getLocalHdr (PyObject*  pkg)
{
   PyObject*  obj = pyMethodCall(pkg, "localHdr", "()");
   if (obj == NULL || obj == Py_None) return QString();
   return pyStringDec(obj);
}

Package::Kind  YumIO::getPkgKind (PyObject*  pkg)
{
   return simpleString(pkg, "group") == "Suite" ? Package::SUITE 
                                                : Package::PACKAGE;
}

PyObject*  YumIO::getGroupItem_22 (const Package::Entry&  entry,
                                   char*                  attr)
{
   if (hasDistribGroups()) {
      error(tr("getGroupItem_22 called for incorrect version of yum"));
      return NULL;
   }

   PyObject*  dict = PyObject_GetAttrString(entry.obj, attr);
   if (dict == NULL) {
      error(tr("Internal error: groupInfo.%1 not evaluable").arg(attr));
      return NULL;
   }
   PyObject*  result = pyDictGetItem(dict, entry.name);
   if (result == NULL) {
      error(tr("Internal error: groupInfo.%1[%2]"
               " not evaluable").arg(attr).arg(entry.name));
      return NULL;
   }
   return result;
}

PyObject*  YumIO::getGroupItem_25 (const Package::Entry&  entry,
                                   char*                  attr)
{
   if (!hasDistribGroups()) {
      error(tr("getGroupItem_25 called for incorrect version of yum"));
      return NULL;
   }
      
   PyObject*  result = PyObject_GetAttrString(entry.obj, attr);
   if (result == NULL) {
      error(tr("Internal error: group %1.%2 not evaluable").arg(attr)
                                                           .arg(entry.name));
      return NULL;
   }
   return result;
}

bool  YumIO::applyGroups()
{
   Package::EntryList::const_iterator  it = list.begin();
   for (; it != list.end(); ++it) {
      Package::Entry*  entry = *it;

      if (entry->kind == Package::GROUP) {
         switch (entry->apply.action) {
         case Package::UPDATE:
            if (!applyGrpUpdate(entry)) return false;
            break;
         case Package::INSTALL:
            if (!applyGrpInstall(entry)) return false;
            break;
         case Package::REMOVE:
            if (!applyGrpRemove(entry)) return false;
            break;
         case Package::NONE:
            // DO NOTHING
            break;
         }
      }
   }

   return true;
}

bool  YumIO::applyGrpUpdate (const Package::Entry*  grp)
{
   QSet<QString>  reqs;
   {
      PyObject*  reqPkgs = requiredPkgs(grp);
      if (reqPkgs == NULL) return false;
      PyObject*  it      = PyObject_GetIter(reqPkgs);
      for (;;) {
         PyObject*  pkgItem = PyIter_Next(it);
         if (pkgItem == NULL) break;
         reqs.insert(pyString(pkgItem));
         Py_DECREF(pkgItem);
      }
      Py_DECREF(it);
      Py_DECREF(reqPkgs);
   }
   
   PyObject*  pkgs = pkgTree(grp);
   if (pkgs == NULL) return false;
   PyObject*  it   = PyObject_GetIter(pkgs);
   for (;;) {
      PyObject*  pkgItem = PyIter_Next(it);
      if (pkgItem == NULL) break;

      EntryKey   key(false, pyString(pkgItem));
      if (hash.contains(key)) {
         Package::Entry*  entry = hash[key];
         switch (entry->apply.action) {
         case Package::NONE:
            // Only install required packages.  But if an optional package is
            // installed, do update it if necessary.
            switch (entry->state) {
            case Package::OUT_OF_DATE:
               entry->apply.action = Package::UPDATE;
               applyGrpLog(grp, entry);
               break;
            case Package::NOT_INSTALLED:
               if (reqs.contains(entry->name)) {
                  entry->apply.action = Package::INSTALL;
                  applyGrpLog(grp, entry);
               } else {
                  applyGrpLogNot(grp, entry, Package::INSTALL,
                                 tr("the package is optional"));
               }
               break;
            case Package::INSTALLED:
               // DO NOTHING
            default:
               // DO NOTHING
               break;
            }
            break;
         case Package::UPDATE:
         case Package::INSTALL:
            // DO NOTHING: no conflict
            applyGrpLogWas(grp, entry);
            break;
         case Package::REMOVE:
         default:
            if (!applyGrpConflict(grp, entry)) {
               return false;
            }
            break;
         }
      }

      Py_DECREF(pkgItem);
   }
   Py_DECREF(it);

   return true;
}

bool  YumIO::applyGrpInstall (const Package::Entry*  grp)
{
   PyObject*  pkgs = pkgTree(grp);
   if (pkgs == NULL) return false;
   PyObject*  it   = PyObject_GetIter(pkgs);
   for (;;) {
      PyObject*  pkgItem = PyIter_Next(it);
      if (pkgItem == NULL) break;

      EntryKey   key(false, pyString(pkgItem));
      if (hash.contains(key)) {
         Package::Entry*  entry = hash[key];
         switch (entry->apply.action) {
         case Package::NONE:
            // Install or update all packages
            switch (entry->state) {
            case Package::OUT_OF_DATE:
               entry->apply.action = Package::UPDATE;
               applyGrpLog(grp, entry);
               break;
            case Package::NOT_INSTALLED:
               entry->apply.action = Package::INSTALL;
               applyGrpLog(grp, entry);
               break;
            case Package::INSTALLED:
            default:
               // DO NOTHING
               break;
            }
            break;
         case Package::UPDATE:
         case Package::INSTALL:
            // DO NOTHING: no conflicts
            applyGrpLogWas(grp, entry);
            break;
         case Package::REMOVE:
         default:
            if (!applyGrpConflict(grp, entry)) {
               return false;
            }
            break;
         }
      }

      Py_DECREF(pkgItem);
   }
   Py_DECREF(it);

   return true;
}

bool  YumIO::applyGrpRemove (const Package::Entry*  grp)
{
   PyObject*    pkgs = allPkgs(grp);
   PyObject*    it   = PyObject_GetIter(pkgs);
   for (;;) {
      PyObject*  pkgItem = PyIter_Next(it);
      if (pkgItem == NULL) break;

      EntryKey   key(false, pyString(pkgItem));
      if (hash.contains(key)) {
         Package::Entry*  entry = hash[key];
         switch (entry->apply.action) {
         case Package::NONE:
            switch (entry->state) {
            case Package::OUT_OF_DATE:
            case Package::INSTALLED:
               entry->apply.action = Package::REMOVE;
               applyGrpLog(grp, entry);
               break;
            case Package::NOT_INSTALLED:
            default:
               // DO NOTHING
               break;
            }
            break;
         case Package::REMOVE:
            // DO NOTHING: no conflict
            applyGrpLog(grp, entry);
            break;
         case Package::UPDATE:
         case Package::INSTALL:
         default:
            if (!applyGrpConflict(grp, entry)) {
               return false;
            }
            break;
         }
      }

      Py_DECREF(pkgItem);
   }
   Py_DECREF(it);

   return true;
}

bool  YumIO::applyGrpConflict (const Package::Entry*  grp,
                               Package::Entry*        pkg)
{
   Package::Action  grpPkgAction;

   switch (pkg->state) {
   case Package::OUT_OF_DATE:   grpPkgAction = Package::UPDATE;  break;
   case Package::NOT_INSTALLED: grpPkgAction = Package::INSTALL; break;
   case Package::INSTALLED:     grpPkgAction = Package::REMOVE;  break;
   default:
      error("Internal error: applyGrpConflict unexpected package state\n");
      return false;
   }

   QString  grpImage    = ACTION_IMAGE(grp->apply.action);
   QString  grpPkgImage = ACTION_IMAGE(grpPkgAction);
   QString  pkgImage    = ACTION_IMAGE(pkg->apply.action);

   QString  msg;
   if (pkg->apply.reason == pkg) {
      msg = tr("Package %1 is included in group %2 which was marked for %3,\n"
               "so the package would be marked for %4,\n"
               "but it was marked individually for %5.\n"
               "How do you wish to proceed?")
               .arg(pkg->name)
               .arg(grp->name)
               .arg(grpImage)
               .arg(grpPkgImage)
               .arg(pkgImage);
   } else if (pkg->apply.reason != NULL) {
      msg = tr("Package %1 is included in group %2 which was marked for %3,\n"
               "so the package would be marked for %4,\n"
               "but it also was part of group %5 and so was marked for %6.\n"
               "How do you wish to proceed?")
               .arg(pkg->name)
               .arg(grp->name)
               .arg(grpImage)
               .arg(grpPkgImage)
               .arg(pkg->apply.reason->name)
               .arg(pkgImage);
   } else {
      msg = tr("Package %1 is included in group %2 which was marked for %3,\n"
               "so the package would be marked for %4,\n"
               "but you resolved an earlier conflict by marking it for %5.\n"
               "How do you wish to proceed?")
               .arg(pkg->name)
               .arg(grp->name)
               .arg(grpImage)
               .arg(grpPkgImage)
               .arg(pkgImage);
   }

   int  result = dialogs->critical(/* caption             => */ tr("Apply"),
                                   /* text                => */ msg,
                                   /* button0Text         => */ grpPkgImage,
                                   /* button1Text         => */ pkgImage,
                                   /* button2Text         => */ tr("Cancel"),
                                   /* defaultButtonNumber => */ 2,
                                   /* escapeButtonNumber  => */ 2);
   
   switch (result) {
   case 0:
      pkg->apply.action = grpPkgAction;
      break;
   case 1:
      // DO NOTHING
      break;
   case 2:
      return false;
   }

   log(Log::NORMAL, 4, tr("Manually resolving conflict: marking package %1 for"
                          " %3").arg(pkg->name)
                                .arg(ACTION_IMAGE(pkg->apply.action)));
   
      
   pkg->apply.reason = NULL; // to indicate a user conflict resolution
   return true;
}

void  YumIO::applyGrpLog (const Package::Entry*  grp,
                          const Package::Entry*  pkg)
{
   log(Log::NORMAL, 4, tr("For group %1, marking package %2 for"
                          " %3").arg(grp->name)
                                .arg(pkg->name)
                                .arg(ACTION_IMAGE(pkg->apply.action)));
}

void  YumIO::applyGrpLogWas (const Package::Entry*  grp,
                             const Package::Entry*  pkg)
{
   log(Log::NORMAL, 4, tr("For group %1, package %2 already was marked for"
                          " %3").arg(grp->name)
                                .arg(pkg->name)
                                .arg(ACTION_IMAGE(pkg->apply.action)));
}

void  YumIO::applyGrpLogNot (const Package::Entry*  grp,
                             const Package::Entry*  pkg,
                             Package::Action        action,
                             QString                reason)
{
   log(Log::NORMAL, 4, tr("For group %1, not marking package %2 for"
                          " %3, becuase %4").arg(grp->name)
                                            .arg(pkg->name)
                                            .arg(ACTION_IMAGE(action))
                                            .arg(reason));
}

PyObject*  YumIO::requiredPkgs (const Package::Entry*  grp)
{
   if (hasDistribGroups()) {
      PyObject*  mand = PyObject_GetAttrString(grp->obj, "mandatory_packages");
      PyObject*  def  = PyObject_GetAttrString(grp->obj, "default_packages");
      if (mand == NULL || def == NULL) {
         error(tr("Internal error: group lacks required package lists"));
         return NULL;
      }
      PyObject*  mand_keys = pyMethodCall(mand, "keys", "()");
      PyObject*  def_keys  = pyMethodCall(def,  "keys", "()");
      if (mand_keys == NULL || def_keys == NULL) {
         error(tr("Internal error:"
                  " group lacks required proper required package lists"));
         return NULL;
      }
      PyObject*  result = PySequence_Concat(mand_keys, def_keys);
      Py_DECREF(def_keys);
      Py_DECREF(mand_keys);
      return result;
   } else {
      char*      name   = strdup(grp->name.toStdString().c_str());
      PyObject*  result = pyMethodCall(grp->obj, "requiredPkgs", "(s)", name);
      free(name);
      return result;
   }
}

PyObject*  YumIO::pkgTree (const Package::Entry*  grp)
{
   if (hasDistribGroups()) {
      // There are no subgroups in yum >= 2.5, so these two are equivalent.
      return requiredPkgs(grp);
   } else {
      char*      name   = strdup(grp->name.toStdString().c_str());
      PyObject*  result = pyMethodCall(grp->obj, "pkgTree", "(s)", name);
      free(name);
      return result;
   }
}

PyObject*  YumIO::allPkgs (const Package::Entry*  grp)
{
   if (hasDistribGroups()) {
      PyObject*  mand = PyObject_GetAttrString(grp->obj, "mandatory_packages");
      PyObject*  def  = PyObject_GetAttrString(grp->obj, "default_packages");
      PyObject*  opt  = PyObject_GetAttrString(grp->obj, "optional_packages");
      PyObject*  cond = PyObject_GetAttrString(grp->obj, 
                                               "conditional_packages");
      if (mand == NULL || def == NULL || opt == NULL || cond == NULL) {
         error(tr("Internal error: group lacks package lists"));
         return NULL;
      }
      PyObject*  mand_keys = pyMethodCall(mand, "keys", "()");
      PyObject*  def_keys  = pyMethodCall(def,  "keys", "()");
      PyObject*  opt_keys  = pyMethodCall(opt,  "keys", "()");
      PyObject*  cond_keys = pyMethodCall(cond, "keys", "()");
      if (mand_keys == NULL || def_keys == NULL
          || opt_keys == NULL || cond_keys == NULL) {
         error(tr("Internal error:"
                  " group lacks required proper package lists"));
         return NULL;
      }
      PyObject*  temp1   = PySequence_Concat(mand_keys, def_keys);
      PyObject*  temp2   = PySequence_Concat(opt_keys,  cond_keys);
      PyObject*  result = PySequence_Concat(temp1, temp2);
      Py_DECREF(temp2);
      Py_DECREF(temp1);
      Py_DECREF(cond_keys);
      Py_DECREF(opt_keys);
      Py_DECREF(def_keys);
      Py_DECREF(mand_keys);
      return result;
   } else {
      char*      name   = strdup(grp->name.toStdString().c_str());
      PyObject*  result = pyMethodCall(grp->obj, "requiredPkgs", "(s)", name);
      free(name);
      return result;
   }
}

bool  YumIO::makeLists (PyObject*   yumIO,
                        PyObject*&  tsInfo,
                        PyObject*&  lists,
                        PyObject*&  updated,
                        PyObject*&  installed,
                        PyObject*&  removed,
                        PyObject*&  obsoleted,
                        PyObject*&  depup,
                        PyObject*&  depinst,
                        PyObject*&  deprem)
{
   tsInfo = PyObject_GetAttrString(yumIO, "tsInfo");
   lists  = pyMethodCall(tsInfo, "makelists", "()");

   if (yumVersion() >= "2.3.2") {
      updated   = PyObject_GetAttrString(tsInfo, "updated");
      installed = PyObject_GetAttrString(tsInfo, "installed");
      removed   = PyObject_GetAttrString(tsInfo, "removed");
      obsoleted = PyObject_GetAttrString(tsInfo, "obsoleted");
      depup     = PyObject_GetAttrString(tsInfo, "depupdated");
      depinst   = PyObject_GetAttrString(tsInfo, "depinstalled");
      deprem    = PyObject_GetAttrString(tsInfo, "depremoved");
   } else {
      if (!PyTuple_Check(lists)) {
         error("Internal error: yumIO.makelists() did not return a tuple\n");
         Py_DECREF(tsInfo);
         Py_DECREF(lists);
         return false;
      }
      if (PyTuple_Size(lists) != 7) {
         error("Internal error: yumIO.makelists() returned tuple of incorrect"
               " size\n");
         Py_DECREF(tsInfo);
         Py_DECREF(lists);
         return false;
      }
      updated   = PyTuple_GetItem(lists, 0);
      installed = PyTuple_GetItem(lists, 1);
      removed   = PyTuple_GetItem(lists, 2);
      obsoleted = PyTuple_GetItem(lists, 3);
      depup     = PyTuple_GetItem(lists, 4);
      depinst   = PyTuple_GetItem(lists, 5);
      deprem    = PyTuple_GetItem(lists, 6);
   }

   return true;
}

QString  YumIO::simpleString (PyObject*    pkg,
                              const char*  key)
{
   char*  copy = new char[strlen(key)+1];
   strcpy(copy, key);
   QString  result = simpleString(pkg, copy);
   delete [] copy;
   return result;
}

QString  YumIO::simpleString (PyObject*  pkg,
                              char*      key)
{
   Python::PyGILStateWrapper  gsw;

   pyExceptExpect(PyExc_KeyError);
   PyObject*  obj = pyMethodCall(pkg, "returnSimple", "(s)", key);
   pyExceptClear();
   if (obj == NULL || obj == Py_None) return QString();

   QString  result = pyStringXDec(obj);
   result.remove(QRegExp("\n*$"));

   return result;
}

Package::Size  YumIO::sizeOf(PyObject*  pkg)
{
   PyObject*      sizeObj = PyObject_GetAttrString(pkg, "size");
   Package::Size  result  = 0;

   if (PyLong_Check(sizeObj)) {
      result = PyLong_AsLongLong(sizeObj);
   } else if (PyInt_Check(sizeObj)) {
      result = PyInt_AsLong(sizeObj);
   } else if (PyString_Check(sizeObj)) {
      QString  str = pyString(sizeObj);
      errno = 0;
      result = strtoull(str.toStdString().c_str(), NULL, 10);
      if (errno != 0) {
         warning(tr("Internal error: package size not understood:"
                    " %1").arg(str));
      }
   } else if (PyCallable_Check(sizeObj)) {
      PyObject*  obj = pyMethodCall(pkg, "size", "()");

      if (PyLong_Check(obj)) {
         result = PyLong_AsLongLong(obj);
         Py_DECREF(obj);
      } else if (PyInt_Check(obj)) {
         result = PyInt_AsLong(obj);
         Py_DECREF(obj);
      } else if (PyString_Check(obj)) {
         QString  str = pyStringDec(obj);
         errno = 0;
         result = strtoull(str.toStdString().c_str(), NULL, 10);
         if (errno != 0) {
            warning(tr("Internal error: package size not understood:"
                       " %1").arg(str));
         }
      } else {
         warning(tr("Internal error: package size result not understood"));
         Py_DECREF(obj);
      }
   } else {
      warning(tr("Internal error: package size not understood"));
   }
   
   return result;
}

QString  YumIO::bytesFormat(Package::Size  bytes)
{
   QString  result;
   double   units  = (double)bytes;
   QString  suffix = tr("b");

   if (bytes >= 1000ULL*1024*1024*1024*1024*1024) {
      units /= 1024.0*1024*1024*1024*1024*1024;
      suffix = tr("Eb");
   } else if (bytes >= 1000ULL*1024*1024*1024*1024) {
      units /= 1024.0*1024*1024*1024*1024;
      suffix = tr("Pb");
   } else if (bytes >= 1000ULL*1024*1024*1024) {
      units /= 1024.0*1024*1024*1024;
      suffix = tr("Tb");
   } else if (bytes >= 1000ULL*1024*1024) {
      units /= 1024.0*1024*1024;
      suffix = tr("Gb");
   } else if (bytes >= 1000ULL*1024) {
      units /= 1024.0*1024;
      suffix = tr("Mb");
   } else if (bytes >= 1000ULL) {
      units /= 1024.0;
      suffix = tr("Kb");
   }

   if (0.0 < units && units < 9.95) {
      result.sprintf("%.1f", units);
   } else {
      result.sprintf("%.0f", units);
   }
   result.append(suffix);

   return result;
}

time_t  YumIO::buildTimeOf(PyObject*  pkg)
{
   return pyTimeDec(pyMethodCall(pkg, "returnSimple", "(s)", "buildtime"));
}

#define TIME_1  "Mon Dec 31 "
#define TIME_2  "Mon Dec 31 03:00:00 "

char*  YumIO::dateFormat(time_t  secs)
{
   static char  buffer[1024];
   ctime_r(&secs, buffer);
   buffer[strlen(buffer)-1] = '\0'; // Strip the newline
   strcpy(&buffer[STRLEN(TIME_1)], &buffer[STRLEN(TIME_2)]);
   return buffer;
}

char*  YumIO::timeFormat(time_t  secs)
{
   static char  buffer[1024];
   ctime_r(&secs, buffer);
   buffer[strlen(buffer)-1] = '\0'; // Strip the newline
   return buffer;
}

QString  YumIO::yumVersion()
{
   if (!yumVerKnown) {
      PyObject*  ver = pyValue("yum", "__version__");
      if (ver == NULL) {
         ver = pyValue("cli", "__version__");
      }

      if (ver != NULL) {
         yumVer = pyString(ver);
      }
      yumVerKnown = true;
   }

   return yumVer;
}

bool  YumIO::hasDistribGroups()
{
   // This implies that group information now is distributed into a separate
   // class for each individual group.  It also implies that the concept of
   // sub-groups (and metapkgs) have been removed.
   return yumVersion() >= "2.5";
}

bool  YumIO::hasPickledCache()
{
   return yumVersion() < "2.9.2";
}

QString  YumIO::pyString(PyObject*  obj)
{
   if (PyString_Check(obj)) {
      return Python::pyString(obj);
   } else if (PyUnicode_Check(obj)) {
      Py_ssize_t  size   = PyUnicode_GetSize(obj);
      wchar_t*    buffer = new wchar_t[size+1];
      Py_ssize_t  actual = PyUnicode_AsWideChar((PyUnicodeObject*)obj,
                                                buffer, size+1);
      if (actual == -1 && PyErr_Occurred()) handleException();
      if (actual >= 0) {
         buffer[actual] = '\0';
      }
      QString  result = QString::fromStdWString(buffer);
      delete [] buffer;
      return result;
   } else if (obj == Py_None) {
      return QString();
   } else {
      warning(tr("Internal error: YumIO::pyString did not understand type"));
      return QString();
   }
}

QString  YumIO::pyStringDec(PyObject*  obj)
{
   QString  result = pyString(obj);
   Py_XDECREF(obj);
   return result;
}

// Generally the same as pyString, except that it returns a QString.  But if
// a yum string is parenthesized (e.g. "(null)"), then it gets turned into a
// list.  This copes with that and turns it back into a string.
QString  YumIO::pyStringX (PyObject*  obj)
{
   if (PyString_Check(obj) || PyUnicode_Check(obj)) {
      return pyString(obj);
   } else if (PyList_Check(obj)) {
      if (PyList_Size(obj) == 0) {
         return QString();
      } else {
#if 1
         warning(tr("Internal error:"
                    " YumIO::pyStringX did not understand non-empty list"));
         return QString();
#else
         QString  result("(");
         PyObject*  it = PyObject_GetIter(obj);
         for (;;) {
            PyObject*  item = PyIter_Next(it);
            if (item == NULL) break;
            result.append(pyStringX(item));
            Py_DECREF(item);
         }
         result.append(")");
         return result;
#endif
      }
   } else {
      warning(tr("Internal error: YumIO::pyStringX did not understand type"));
      return QString();
   }      
}

QString  YumIO::pyStringXDec (PyObject*  obj)
{
   QString  result = pyStringX(obj);
   Py_XDECREF(obj);
   return result;
}

QString  YumIO::pyMappingString (PyObject*    obj,
                                 const char*  attr)
{
   char*  copy = new char[strlen(attr)+1];
   strcpy(copy, attr);
   QString  result = pyStringDec(PyMapping_GetItemString(obj, copy));
   delete [] copy;
   return result;
}

time_t  YumIO::pyTime(PyObject*  obj)
{
   time_t  result = 0;

   if (PyInt_Check(obj)) {
      result = PyInt_AsLong(obj);
   } else if (PyString_Check(obj)) {
      QString  str = pyString(obj);
      errno = 0;
      result = strtoul(str.toStdString().c_str(), NULL, 10);
      if (errno != 0) {
         warning(tr("Internal error: time not understood: %1").arg(str));
      }
   } else {
      warning(tr("Internal error: time not understood"));
   }

   return result;
}

time_t  YumIO::pyTimeDec(PyObject*  obj)
{
   time_t  result = 0;

   if (PyInt_Check(obj)) {
      result = PyInt_AsLong(obj);
      Py_DECREF(obj);
   } else if (PyString_Check(obj)) {
      QString  str = pyStringDec(obj);
      errno = 0;
      result = strtoul(str.toStdString().c_str(), NULL, 10);
      if (errno != 0) {
         warning(tr("Internal error: time not understood: %1").arg(str));
      }
   } else {
      warning(tr("Internal error: time not understood"));
      Py_DECREF(obj);
   }

   return result;
}

PyObject*  YumIO::pyDictGetItem (PyObject*       dict,
                                 const QString&  key)
{
   // This is because PyDict_GetItemString's 2nd parameter is a char* and not
   // a const char*.  It probably is just that its parameter profile isn't
   // const-correct.  But I'm not chancing it.
   char*      keyCopy = strdup(key.toStdString().c_str());
   PyObject*  result  = PyDict_GetItemString(dict, keyCopy);
   free(keyCopy);
   return result;
}

void  YumIO::log (Log::Kind     kind,
                  unsigned int  value,
                  QString       text)
{
   QString  edited(text);
   edited.replace(QRegExp(": (\\[Errno [0-9]*\\] )"), ":\n\\1");

   QStringList  lines = edited.split("\n");
   
   logEntries.lock();

   QStringList::const_iterator  it = lines.begin();
   for (; it != lines.end(); ++it) {
      logEntries.append(Log::Entry(kind, value, *it));
   }
   
   logEntries.unlock();
   logEntries.emitAppended();
}

void  YumIO::pyExceptExpect (PyObject*  xcpt)
{
   logEntries.lock();
   logStartOp = logEntries.count();
   logEntries.unlock();
   Python::pyExceptExpect(xcpt);
}

void  YumIO::pyExceptClear ()
{
   logStartOp = -1;
   Python::pyExceptClear();
}

void  YumIO::pythonException (QString  text)
{
   int  startOp;
   if (logStartOp != -1) {
      startOp = logStartOp;
   } else {
      logEntries.lock();
      startOp = logEntries.count();
      logEntries.unlock();
   }

   log(Log::ERROR, 0, text);

   QString  msg;

   logEntries.lock();
   int  i;
   for (i = startOp; i < logEntries.count(); i++) {
      if (logEntries[i].kind == Log::ERROR && logEntries[i].value <= 2) {
         if (!msg.isEmpty()) {
            msg.append('\n');
         }
         msg.append(logEntries[i].text);
      }
   }
   logEntries.unlock();

   dialogs->scrollable(tr("NUU Yum Error"), msg);
}

void  YumIO::pythonExceptionV (QString  text)
{
   int  startOp;
   if (logStartOp != -1) {
      startOp = logStartOp;
   } else {
      logEntries.lock();
      startOp = logEntries.count();
      logEntries.unlock();
   }

   log(Log::ERROR, 0, pyString(pyExceptValuef()));
   log(Log::ERROR, 0, text);

   QString  msg;

   logEntries.lock();
   int  i;
   for (i = startOp; i < logEntries.count(); i++) {
      if (logEntries[i].kind == Log::ERROR && logEntries[i].value <= 2) {
         if (!msg.isEmpty()) {
            msg.append('\n');
         }
         msg.append(logEntries[i].text);
      }
   }
   logEntries.unlock();

   dialogs->scrollable(tr("NUU Yum Error"), msg);
}

void  YumIO::exceptionExpected()
{
#ifdef DEBUG_EXPECTED_EXCEPTIONS
   // This exception was totally expected.  But maybe for debugging I'd
   // want to know the particulars.  So, stash this at a very high log
   // level to imply its debugging nature.  Note that a Search operation
   // will cause tons of these.

   Log::Kind     oldStderrKind  = wr.stderrKind;
   unsigned int  oldStderrValue = wr.stderrValue;

   wr.stderrKind  = Log::NORMAL;
   wr.stderrValue = 10;

   PyObject*  exception;
   PyObject*  v;
   PyObject*  tb;
   PyErr_Fetch(&exception, &v, &tb);
   PyErr_NormalizeException(&exception, &v, &tb);
   PyErr_Display(exception, v, tb);
   PyErr_Restore(exception, v, tb);

   wr.stderrKind  = oldStderrKind;
   wr.stderrValue = oldStderrValue;

   Python::exceptionExpected();
#endif
}

/*
** Python type interfaces
*/

QString  YumIO::pkgtupFormat(PyObject* pkgtup)
{
   QString  name  = pyString(PyTuple_GetItem(pkgtup, 0));
   QString  arch  = pyString(PyTuple_GetItem(pkgtup, 1));
   QString  epoch = pyString(PyTuple_GetItem(pkgtup, 2));
   QString  ver   = pyString(PyTuple_GetItem(pkgtup, 3));
   QString  rel   = pyString(PyTuple_GetItem(pkgtup, 4));

   if (epoch == QString("0")) {
      return tr("%1 %3-%4 (%2)").arg(name).arg(arch).arg(ver).arg(rel);
   } else {
      return tr("%1 %3:%4-%5 (%2)")
                .arg(name).arg(arch).arg(epoch).arg(ver).arg(rel);
   }
}

QString  YumIO::pkgtupRepoFormat(PyObject*  pkgtup,
                                 QString    repoid)
{
   return pkgtupFormat(pkgtup).append(tr(" [%1]").arg(repoid));
}

int YumIO::displayTransactionPkgs(QString    action,
                                  PyObject*  txmbrs)
{
   PyObject*  it = PyObject_GetIter(txmbrs);
   for (;;) {
      PyObject*  txmbr = PyIter_Next(it);
      if (txmbr == NULL) break;
      PyObject*  pkgtup = PyObject_GetAttrString(txmbr, "pkgtup");
      if (txmbr == NULL) {
         error("Internal error: txmbr lacks pkgtup\n");
         Py_DECREF(txmbr);
         Py_DECREF(it);
         return 0;
      }
      ts.descr.append(tr("\t%1: %2\n").arg(action).arg(pkgtupFormat(pkgtup)));
      Py_DECREF(pkgtup);
      Py_DECREF(txmbr);
   }

   Py_DECREF(it);

   return 1;
}

int YumIO::displayTransactionPkgsRepo(QString    action,
                                      PyObject*  txmbrs)
{
   PyObject*  it = PyObject_GetIter(txmbrs);
   for (;;) {
      PyObject*  txmbr = PyIter_Next(it);
      if (txmbr == NULL) break;
      PyObject*  pkgtup = PyObject_GetAttrString(txmbr, "pkgtup");
      if (txmbr == NULL) {
         error("Internal error: txmbr lacks pkgtup\n");
         Py_DECREF(txmbr);
         Py_DECREF(it);
         return 0;
      }
      QString  repoid = pyStringDec(PyObject_GetAttrString(txmbr, "repoid"));
      ts.descr.append(tr("\t%1: %2\n").arg(action)
                                      .arg(pkgtupRepoFormat(pkgtup, repoid)));
      Py_DECREF(pkgtup);
      Py_DECREF(txmbr);
   }

   Py_DECREF(it);

   return 1;
}

PyObject*  YumIO::findRelation(PyObject*    relatedto,
                               const char*  relation)
{
   PyObject*  it = PyObject_GetIter(relatedto);
   for (;;) {
      PyObject*  tuple = PyIter_Next(it);
      if (tuple == NULL) break;

      if (strcmp(Python::pyString(PyTuple_GetItem(tuple, 1)), relation) == 0) {
         PyObject*  result = PyTuple_GetItem(tuple, 0);
         Py_INCREF(result);
         Py_DECREF(tuple);
         Py_DECREF(it);
         return result;
      }

      Py_DECREF(tuple);
   }

   Py_DECREF(it);

   return NULL;
}

int YumIO::displayTransactionPkgsObs(QString    action,
                                     PyObject*  txmbrs)
{
   PyObject*  it = PyObject_GetIter(txmbrs);
   for (;;) {
      PyObject*  txmbr = PyIter_Next(it);
      if (txmbr == NULL) break;
      PyObject*  pkgtup = PyObject_GetAttrString(txmbr, "pkgtup");
      if (txmbr == NULL) {
         error("Internal error: txmbr lacks pkgtup\n");
         return 0;
      }
      PyObject*  relatedto = PyObject_GetAttrString(txmbr, "relatedto");
      if (relatedto == NULL) {
         error("Internal error: package has no relatedto\n");
         Py_DECREF(pkgtup);
         Py_DECREF(txmbr);
         Py_DECREF(it);
         return 0;
      }
      PyObject*  obspkg = findRelation(relatedto, "obsoletedby");
      if (obspkg == NULL) {
         error("Internal error: couldn't find obsoletedby package\n");
         Py_DECREF(relatedto);
         Py_DECREF(pkgtup);
         Py_DECREF(txmbr);
         Py_DECREF(it);
         return 0;
      }
      if (yumVersion() >= "2.3.2") {
         PyObject*  obstup = PyObject_GetAttrString(obspkg, "pkgtup");
         if (obspkg == NULL) {
            error("Internal error: couldn't find obsoletedby package\n");
            Py_DECREF(obspkg);
            Py_DECREF(relatedto);
            Py_DECREF(pkgtup);
            Py_DECREF(txmbr);
            Py_DECREF(it);
            return 0;
         }
         ts.descr.append(tr("\t%1: %2 with %3\n").arg(action)
                                                 .arg(pkgtupFormat(pkgtup))
                                                 .arg(pkgtupFormat(obstup)));
         Py_DECREF(obstup);
      } else {
         ts.descr.append(tr("\t%1: %2 with %3\n").arg(action)
                                                 .arg(pkgtupFormat(pkgtup))
                                                 .arg(pkgtupFormat(obspkg)));
      }
      Py_DECREF(obspkg);
      Py_DECREF(relatedto);
      Py_DECREF(pkgtup);
      Py_DECREF(txmbr);
   }

   Py_DECREF(it);

   return 1;
}

void  YumIO::noteTransactionPkgs(Package::Action  action,
                                 PyObject*        txmbrs)
{
   PyObject*  it = PyObject_GetIter(txmbrs);

   for (;;) {
      PyObject*  txmbr = PyIter_Next(it);
      if (txmbr == NULL) break;
      PyObject*  pkgtup = PyObject_GetAttrString(txmbr, "pkgtup");
      if (txmbr == NULL) {
         error("Internal error: txmbr lacks pkgtup\n");
         Py_DECREF(txmbr);
         Py_DECREF(it);
         return;
      }

      QString          name  = pyString(PyTuple_GetItem(pkgtup, 0));
      Package::Entry*  entry = hash[EntryKey(false, name)];
      switch (action) {
      case Package::INSTALL:
         ts.run.nominalGrabSize += PG_R_GRAB(entry->size);
         if (!QFile::exists(getLocalPkg(entry->obj))) {
            ts.run.totalGrabSize += PG_R_GRAB(entry->size);
         }
         ts.run.totalTsSize += PG_R_INSTALL(entry->size);
         break;
      case Package::UPDATE:
         ts.run.nominalGrabSize += PG_R_GRAB(entry->size);
         if (!QFile::exists(getLocalPkg(entry->obj))) {
            ts.run.totalGrabSize += PG_R_GRAB(entry->size);
         }
         ts.run.totalTsSize += PG_R_UPDATE(entry->size);
         ts.run.totalTsSize += PG_R_COMPLETE_UPDATE(entry->next->size);
         ts.updates.insert(name);
         break;
      case Package::REMOVE:
         ts.run.totalTsSize += PG_R_REMOVE(entry->size);
         break;
      default:
         // DO NOTHING
         break;
      }

      Py_DECREF(pkgtup);
      Py_DECREF(txmbr);
   }

   Py_DECREF(it);
}

void  YumIO::error(QString  text)
{
   log(Log::ERROR, 0, text);
   dialogs->critical(tr("NUU Yum Error"), text);
}

void  YumIO::warning(QString  text)
{
   log(Log::WARNING, 0, text);
   dialogs->warning(tr("NUU Yum Warning"), text);
}

void  YumIO::error(const char*  text)
{
   error(QString(text));
}


void  YumIO::warning(const char*  text)
{
   warning(QString(text));
}
